# BCON-related package utilities
#
# - functions:
#
# get_const(package, instance, index, buffer): get BCON value/name, index = -1 for [temp],
#   buffer = BCON & return list of values starting at index/None = find and load one

import logging

import dz_pack
import dz_prefs
import dz_util
from dz_defs import *

def get_const(p, i, x, b):
    """Return the value/name of a particular BCON."""
    g2 = p.group
    ci = p.instance
    b2 = b              # b2 = supplied buffer

    if i < 0x1000 and (not b):  # global BCON reference
        k = i - 0x100
        if k < len(p.prefs.glob_bcon_names):
            n = p.prefs.glob_bcon_names[k]
            try:
                v = p.prefs.glob_bcon_values[k][x]
            except IndexError:
                v = 0
                logging.warning("index error in global BCON 0x%X:0x%X" % (i,x))
            return "(%s):0x%X=%s" % (n, x, dz_util.ext16(v))
        else:
            logging.warning("BCON 0x%X out of range" % i)
            return "***"        # out of range

    if i >= 0x2000 and i < 0x3000 and ci >= 0x1000 and ci < 0x2000:
        g2 = p.semi             # if private to sg, use the sg group
        if g2 == 0:
            logging.error("BCON 0x%X:0x%X refers to missing semi-global" % (i,x))
            return "***"        # bad reference

    if not b:
        r = p.find({'type': TYPE_BCON, 'group': g2, 'instance': i})
        if not r:                   # not found
            logging.warning("BCON 0x%X not found" % i)
            if x == -1:
                return ""
            else:
                return ":0x%X" % x      # just index (unknown value)

        s = "***"
        b = p.get(r['offset'], r['size'])   # b = buffer (if BCON found)

    if b:
        s = ""
        if not b2:
            # get the cstring, maxchars = 0x40 at index = 0 (title)
            j = 0x40
            if '\x00' in b[:0x40]:
                j = b.index('\x00')
            if x == -1:
                s = "(%s)" % b[:j]
            else:
                s = "(%s):0x%X=" % (b[:j], x)

        if x != -1:
            w1 = x
            wx = dz_util.s2i(b[0x40:0x42]) & 0x7FFF     # count

            while w1 < wx:
                if b2: s += "0x%X: " % w1   # index

                j = 0x42 + (w1 + w1)        # offset to value
                if j+1 < len(b):
                    w2 = dz_util.s2i(b[j:j+2])
                    s += dz_util.ext16(w2)
                else:
                    logging.warning("index error in BCON 0x%X:0x%X" % (i,x))
                    s += "***"              # range error
                if not b2:
                    w1 = wx
                else:
                    s += "\n"
                    w1 += 1

    return s
