# content handling

import gtk
import logging
import pango

import dz_edit
import dz_mpack
import dz_nav
import dz_pack
import dz_types
import dz_util
import dz_view
from dz_defs import *

def do_open(obj, filename):
    logging.info("open %s" % filename)

    b = (filename == obj.prefs.objects_name)

    obj.status_text = "Opening %s..." % filename
    obj.progress_show = True
    obj.progress_amount = 0
    obj.progress_divisor = 100

    obj.wait_gtk()

    # open specified pack
    if not b and filename != obj.prefs.holiday1 and filename != obj.prefs.holiday2:
        p = dz_mpack.MutablePackage(filename, obj.prefs, False)    # mutable package
    else:
        p = dz_pack.Package(filename, obj.prefs, b)         # read-only package

    if b:
        obj.pack1 = p
    else:
        obj.pack2 = p
        p.objects = obj.pack1

    obj.progress_amount = 25
    obj.wait_gtk()

    if p.toc_valid:
        name = obj.do_title(filename)[:-8] # name = everything but the trailing ".package"

        if b:   # objects
            s = obj.prefs.sg_ids[:]        # slice copy these
            sn = obj.prefs.sg_names[:]
            s.insert(0, GROUP_GLOBAL)
            sn.insert(0, "Global")
            s2 = obj.prefs.object_group
            s2n = obj.prefs.object_names
            for i in range(len(s2)):
                if s2[i] not in s:
                    s.append(s2[i])
                    sn.append(s2n[i])

        else:   # not objects
            # collect list of groups
            s = []
            r = p.find({})
            while r:
                g = r['group']
                if g not in s:
                    s.append(g)
                r = p.find({'index': r['index']})

            # get names of groups
            sn = []
            for i in range(len(s)):
                g = s[i]
                sn.append(dz_view.get_group_name(p, g))

        # find child groups
        scg, scs = [], []
        for i in range(len(s)):
            g = s[i]
            if g in obj.prefs.object_group:
                j = obj.prefs.object_group.index(g)
                g2 = obj.prefs.object_semi[j]
                if g2 != 0 and g2 in s:
                    scg.append(i)               # flag this group as a child
                    scs.append(s.index(g2))     # save index of its parent

        # disconnect the old store, remove old column, select tab
        nb = obj.wTree.get_widget("notebook1")
        if b:
            tv = obj.wTree.get_widget("obj_contentview")
            nb.set_current_page(0)
        else:
            tv = obj.wTree.get_widget("contentview")
            nb.set_current_page(1)
            # tab label = name
            lb = nb.get_tab_label(nb.get_nth_page(1))
            lb.set_text(name)
            if len(name) > 12:
                lb.set_max_width_chars(12)
            else:
                lb.set_max_width_chars(-1)

        obj.current_cv = tv
        tv.set_model()
        x = tv.get_column(0)
        if x: tv.remove_column(x)

        # setup tree store (name, group, type, subtype, instance, meta)
        ts = gtk.TreeStore(str, 'guint', 'guint', 'guint', 'guint', int)

        obj.progress_divisor = len(s) * 2

        # insert sort the data into the store
        for i in range(len(s)):
            obj.progress_amount = len(s) + i
            obj.wait_gtk()
            if i not in scg:
                ix = False
                j = ts.get_iter_root()
                while j:
                    if sn[i].lower() < ts.get_value(j, 0).lower():
                        tip = ts.insert_before(None, j, [sn[i], s[i], 0, 0, 0, 1])
                        ix = True
                        break   # terminate j loop
                    j = ts.iter_next(j)
                if not ix:
                    tip = ts.append(None, [sn[i], s[i], 0, 0, 0, 1])  # header (meta = 1)
                tic = ts.append(tip, ["...", s[i], 0, 0, 0, 2])       # group placeholder (2)
                x = 0
                while i in scs[x:]:         # add child groups
                    j = scs.index(i, x)
                    g = s[scg[j]]
                    xx = False
                    jj = ts.iter_children(tip)
                    while jj:
                        if sn[scg[j]].lower() < ts.get_value(jj, 0).lower():
                            tic = ts.insert_before(tip, jj, [sn[scg[j]], g, 0, 0, 0, 1])
                            xx = True
                            break   # terminate jj loop
                        jj = ts.iter_next(jj)
                    if not xx:
                        tic = ts.append(tip, [sn[scg[j]], g, 0, 0, 0, 1])   # header
                    ti2 = ts.append(tic, ["...", g, 0, 0, 0, 2])      # group placeholder
                    x = j + 1

        # setup tree view with my new data
        cell = gtk.CellRendererText()
        tvc = gtk.TreeViewColumn("Name", cell)
        tvc.set_cell_data_func(cell, obj.on_get_tv_name)

        tv.append_column(tvc)
        tv.set_search_column(0)     # make this searchable
        tv.set_model(ts)            # connect it up

        # set this navigation stack
        nb = obj.wTree.get_widget("notebook2")
        obj.nav_stacks[nb.get_current_page()] = []
        obj.nav_cursor[nb.get_current_page()] = 0

        # set the cursor
        tv.set_cursor(ts.get_path(ts.get_iter_root()))

    obj.progress_show = False
    obj.status_text = ""
    obj.wait_gtk()

    if filename != obj.prefs.objects_name and filename not in obj.prefs.recent_files:
        if filename != obj.prefs.holiday1 and filename != obj.prefs.holiday2:
            obj.prefs.recent_files.append(filename)
            if len(obj.prefs.recent_files) > obj.max_recents:
                obj.prefs.recent_files.pop(0)
            dz_nav.do_set_recents(obj)

def do_contentview_change(obj, page_num, modify=False):
    p = obj.get_current_pack()                      # pack = current 
    widget = obj.current_cv                         # widget = current contents view
    path, tvc = widget.get_cursor()                 # path = current (cursor)

    nb = obj.wTree.get_widget("notebook2")          # nb = result view notebook
    if page_num >= 0:
        sw = nb.get_nth_page(page_num)              # get the scrollwindow
        b = sw.get_child().get_buffer()             # get its textbuffer

        # check whether this is background change
        if b and nb.get_current_page() != page_num:
            f = b.get_data("file_name")                 # buffer's pack file
            path = b.get_data("current_pos")            # buffer's path

            if f and f != p.filename:                   # file is not the same as current?
                if f == obj.prefs.objects_name:
                    p = obj.pack1
                    widget = obj.wTree.get_widget("obj_contentview")
                else:
                    p = obj.pack2
                    widget = obj.wTree.get_widget("contentview")

    m = widget.get_model()                          # get treestore
    i = m.get_iter(path)                            # get data
    x = m.get_value(i, 5)                           # meta

    s = ""          # text result
    sx = []         # markup
    if x == 0:
        s, sx = dz_view.get_view(p, m.get_value(i, 1), m.get_value(i, 2), \
            m.get_value(i, 3), m.get_value(i, 4), obj.prefs.view_mode)
    else:
        s = dz_view.get_view2(p, m.get_value(i, 1), m.get_value(i, 2), \
            m.get_value(i, 3), m.get_value(i, 4), x)

    if nb.get_n_pages() == 0 or obj.cv_event == 2:  # middle mouse button
        page_num = dz_nav.do_new_tab(obj)

    if obj.cv_event == 3:                           # right mouse button
        pass

    logging.info("change content on page %d to %s" % (page_num, m.get_value(i, 0)))

    w2 = nb.get_nth_page(page_num)                  # get ScrolledWindow
    cb = nb.get_tab_label(w2)
    cb.label.set_text(m.get_value(i, 0))            # set the tab label
    s2 = dz_view.get_tips(p, m.get_value(i, 1), m.get_value(i, 2), \
        m.get_value(i, 3), m.get_value(i, 4), x)
    cb.ttips.set_tip(cb.ebox, s2)                   # set the tooltip for the label

    # get a new buffer
    b = gtk.TextBuffer()
    b.set_text(dz_util.get_good_string(s))

    # save file/position
    b.set_data("file_name", p.filename)
    b.set_data("current_pos", path)
    b.set_data("current_group", m.get_value(i, 1))
    b.set_data("current_type", m.get_value(i, 2))
    b.set_data("current_subtype", m.get_value(i, 3))
    b.set_data("current_instance", m.get_value(i, 4))
    b.set_data("current_meta", m.get_value(i, 5))

    # apply overall tag
    wm = gtk.WRAP_NONE
    if obj.prefs.wrap == 1:
        wm = gtk.WRAP_WORD
    t = b.create_tag(None, font=obj.prefs.font_name, wrap_mode=wm)
    b.apply_tag(t, b.get_start_iter(), b.get_end_iter())

    # create link tag
    t2 = b.create_tag("link", foreground='blue')

    # apply link tag to linkages in the buffer
    for j in range(len(sx)):
        mt, mr, ma, mb = sx[j]      # get markup
        if mt == 1 or mt == 2:      # linkage
            b.apply_tag(t2, b.get_iter_at_offset(ma), b.get_iter_at_offset(mb))
            k = "link_type_%d" % mb
            b.set_data(k, mt)       # type 1: behavior tree, 2: GUID
            k = "link_ref_%d" % mb
            b.set_data(k, mr)       # link_ref(end offset) = reference

    w = w2.get_child()                              # get the TextView
    w.set_buffer(b)
    w.set_editable(False)
    w.set_cursor_visible(False)

    if not modify:
        dz_edit.do_select_edit_page(obj)

def do_contentview_modify(obj, p):
    # save the scrolledwindow position
    nb = obj.wTree.get_widget("notebook2")
    sw = nb.get_nth_page(p)
    x = sw.get_hadjustment().get_value()
    y = sw.get_vadjustment().get_value()

    # update content
    obj.cv_event = 0
    do_contentview_change(obj, p, True)

    obj.wait_gtk()  # wait for buffer update

    # restore the scrolledwindow position
    sw.get_hadjustment().set_value(x)
    sw.get_vadjustment().set_value(y)

def do_button(obj, e, it, b2):
    t2 = b2.get_tag_table().lookup("link")

    # check if this is linkage markup
    if ord(it.get_char()) != 0 and t2 and it.has_tag(t2):

        # find the end of markup (or the end of buffer)
        while not it.ends_tag(t2):
            if not it.forward_char(): break

        # get the link type and reference of this markup
        k = "link_type_%d" % it.get_offset()
        lt = it.get_buffer().get_data(k)
        k = "link_ref_%d" % it.get_offset()
        r = it.get_buffer().get_data(k)
        if r:
            obj.cv_event = e.button
            if lt == 1:
                dz_nav.do_goto(obj, r, TYPE_BHAV)
            elif lt == 2:
                dz_nav.do_goto_value(obj, r, None, None, None, True)
            return True     # stop other handlers from using this event (i.e. drag n drop)

    # select line (for editor) when single-clicked
    if e.type == gtk.gdk.BUTTON_PRESS and ord(it.get_char()) != 0:
        t1 = b2.get_tag_table().lookup("edit")
        if not t1:
            t1 = b2.create_tag("edit", underline=pango.UNDERLINE_SINGLE)
        x = it.get_line()
        b2.remove_tag(t1, b2.get_start_iter(), b2.get_end_iter())
        b2.apply_tag(t1, b2.get_iter_at_line(x), b2.get_iter_at_line(x+1))

    word_chars = range(0x30, 0x3A) + range(0x41, 0x5B) + range(0x61, 0x7B)  # alphanumeric chars

    # select "word" when double-clicked
    if e.type == gtk.gdk._2BUTTON_PRESS and ord(it.get_char()) != 0:
        r = it.backward_char()
        while r and ord(it.get_char()) in word_chars:
            r = it.backward_char()
        if ord(it.get_char()) not in word_chars:
            r = it.forward_char()
        it2 = it.copy()             # word start
        r = it.forward_char()
        while r and ord(it.get_char()) in word_chars:
            r = it.forward_char()   # word end
        b2.select_range(it2, it)
        return True

def do_expander(obj, m, i2):
    g = m.get_value(i2, 1)              # group
    t = m.get_value(i2, 2)              # type
    x = m.get_value(i2, 5)              # meta
    p = obj.get_current_pack()

    # note: group placeholders are for children of the group (i.e. types)
    # and type placeholders are for children of types (i.e. instances)

    if x == 2:                          # group placeholder
        obj.status_text = "Expanding..."
        obj.wait_gtk()

        s = []
        sn = []
        for t2 in dz_types.known_values:
            r = p.find({'group': g, 'type': t2})
            if r:
                t2i = dz_types.known_values.index(t2)
                s.append(t2)
                sn.append("<" + dz_types.known_names[t2i] + ">")

        if obj.prefs.view_unknown:
            for t2 in dz_types.unknown_values:
                r = p.find({'group': g, 'type': t2})
                if r:
                    t2i = dz_types.unknown_values.index(t2)
                    s.append(t2)
                    sn.append("*" + dz_types.unknown_names[t2i] + "*")

            r = p.find({'group': g})
            while r:
                t2 = r['type']
                if t2 not in s:
                    s.append(t2)
                    sn.append("0x%X" % t2)
                r = p.find({'group': g, 'index': r['index']})

        if len(s) != 0:                 # replace group placeholder
            m.set(i2, 0, sn[0], 1, g, 2, s[0], 3, 0, 4, 0, 5, 3)  # header (meta = 3)
            i2c = m.append(i2, ["...", g, s[0], 0, 0, 4])         # type placeholder (4)
            s.pop(0)
            sn.pop(0)
            while len(s) != 0:          # insert other types
                i2 = m.insert_after(None, i2, [sn[0], g, s[0], 0, 0, 3])  # header
                i2c = m.append(i2, ["...", g, s[0], 0, 0, 4])             # placeholder
                s.pop(0)
                sn.pop(0)
        else:
            m.remove(i2)    # del the placeholder

    elif x == 4:                        # type placeholder
        obj.status_text = "Expanding..."
        obj.wait_gtk()

        # get instance list
        s = []
        sn = []
        sh = []
        r = p.find({'type': t, 'group': g})
        while r:
            x = r['instance']
            x2 = r['subtype']
            if t == TYPE_GLUA or t == TYPE_OLUA:
                s2 = p.get_name_lua(r['offset'], r['size'])
                sn.append(dz_util.get_good_string(s2))
            else:
                s2 = p.get_name(r['offset'], r['size'])
                s2 = dz_util.get_good_string(s2)
                sn.append("0x%X: %s" % (x, s2))
            s.append(x)
            sh.append(x2)
            r = p.find({'type': t, 'group': g, 'index': r['index']})

        # sort list
        s2 = []
        s2n = []
        s2h = []
        for x in range(len(s)):
            ix = False
            for y in range(len(s2)):
                if t == TYPE_GLUA or t == TYPE_OLUA:
                    xb = sn[x].lower() < s2n[y].lower()
                else:
                    xb = s[x] < s2[y]
                if xb:
                    s2.insert(y, s[x])
                    s2n.insert(y, sn[x])
                    s2h.insert(y, sh[x])
                    ix = True
                    break   # terminate y loop
            if not ix:
                s2.append(s[x])
                s2n.append(sn[x])
                s2h.append(sh[x])

        if len(s2) != 0:                # replace placeholder
            m.set(i2, 0, s2n[0], 1, g, 2, t, 3, s2h[0], 4, s2[0], 5, 0)
            s2.pop(0)
            s2n.pop(0)
            s2h.pop(0)
            while len(s2) != 0:         # insert other instances
                i2 = m.insert_after(None, i2, [s2n[0], g, t, s2h[0], s2[0], 0])
                s2.pop(0)
                s2n.pop(0)
                s2h.pop(0)
        else:
            m.remove(i2)    # del the placeholder

    obj.status_text = ""
    obj.wait_gtk()
