# editor handling

import gtk
import pango

import dz_cont
import dz_mpack
import dz_util
from dz_defs import *

def do_set_range(obj, r, d):
    nb = obj.wTree.get_widget("notebook2")          # nb = result view notebook
    sw = nb.get_nth_page(nb.get_current_page())     # get the scrollwindow
    if sw:
        b = sw.get_child().get_buffer()             # get its textbuffer
        if b:
            d2 = dict(type=b.get_data("current_type"), group=b.get_data("current_group"), \
                instance=b.get_data("current_instance"), subtype=b.get_data("current_subtype"), \
                data=d, range=r)
            obj.pack2.set(d2)       # modify this instance

            # update result view
            dz_cont.do_contentview_modify(obj, nb.get_current_page())

def do_set_byte(obj, x, v):
    do_set_range(obj, [x, x+1], chr(v))

def on_get_pos(column, cell, model, i, obj):
    d = model.get_value(i, 0)       # position
    cell.set_property('text', "%X:" % d)
    cell.set_property('font', obj.prefs.font_name)
    cell.set_property('sensitive', False)

def on_get_values(column, cell, model, i, data):
    # data[0] = obj, data[1] = column num
    s = model.get_value(i, 1)       # length
    p = model.get_data("cursor_path")
    # render the focus a little more visibly
    if p and p[0] == model.get_path(i) and p[1] == column:
        cell.set_property('underline', pango.UNDERLINE_SINGLE)
    else:
        cell.set_property('underline', pango.UNDERLINE_NONE)
    if data[1] < s:
        d = model.get_value(i, data[1]+2)
        cell.set_property('text', "%02X" % d)
        cell.set_property('sensitive', True)
        cell.set_property('editable', True)
    else:
        cell.set_property('text', "")
        cell.set_property('sensitive', False)
        cell.set_property('editable', False)
    cell.set_property('font', data[0].prefs.font_name)

def on_set_values(cell, path, new_text, obj, i, model):
    x = (int(path) << 4) + i            # offset in buffer

    # evaluate input, set visual
    v = dz_util.hex2int(new_text)
    if v < 0: v = 0
    if v > 255: v = 255
    cell.set_property('text', "%02X" % v)

    # compare to old value
    it = model.get_iter(path)
    d = model.get_value(it, i+2)
    if v != d:
        model.set_value(it, i+2, v)     # set ListStore
        do_set_byte(obj, x, v)          # store in mpack

def on_get_ascii(column, cell, model, i, obj):
    a = model.get_value(i, 1)           # length
    s = ""
    for x in range(a):
        c = model.get_value(i, x+2)
        if c >= 0x20 and c <= 0x7E:
            s += chr(c)
        else:
            s += "."
    cell.set_property('text', s)
    cell.set_property('font', obj.prefs.font_name)
    cell.set_property('sensitive', False)

def on_hex_cursor_changed(tv):
    p, c = tv.get_cursor()
    m = tv.get_model()
    m.set_data("cursor_path", (p, c))   # set cursor_path
    tv.queue_draw()                     # redraw listview

def do_update_hex_view(obj, b):
    # disconnect the old store, remove old columns
    tv = obj.wTree.get_widget("hextreeview")
    tv.set_model()
    while tv.get_columns():
        tv.remove_column(tv.get_column(0))

    # new list store (pos, length, values)
    ls = gtk.ListStore(int,int, int,int,int,int, int,int,int,int, int,int,int,int, int,int,int,int)
    cell = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn("Position", cell)
    tvc.set_cell_data_func(cell, on_get_pos, obj)
    tv.append_column(tvc)

    # add values columns
    for i in range(16):
        cell = gtk.CellRendererText()
        cell.connect("edited", on_set_values, obj, i, ls)
        tvc = gtk.TreeViewColumn("%X" % i, cell)
        tvc.set_cell_data_func(cell, on_get_values, (obj, i))
        tv.append_column(tvc)

    # add ascii column
    cell = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn("ASCII", cell)
    tvc.set_cell_data_func(cell, on_get_ascii, obj)
    tv.append_column(tvc)

    # connect to new store
    tv.set_model(ls)
    tv.connect("cursor-changed", on_hex_cursor_changed)

    # add data to store
    x = 0
    x2 = [0] * 16
    for i in range(len(b) >> 4):
        for i in range(16):
            x2[i] = ord(b[x+i])
        ls.append([x, 16] + x2)
        x += 16

    a = len(b) & 15
    if a != 0:
        for i in range(a):
            x2[i] = ord(b[x+i])
        ls.append([x, a] + x2)

def do_select_edit_page(obj, page=-1):
    x = 0
    p = page
    mb = False
    nb = obj.wTree.get_widget("notebook2")          # nb = result view notebook
    if page < 0:
        p = nb.get_current_page()

    sw = nb.get_nth_page(p)                         # get the scrollwindow
    if sw:
        b = sw.get_child().get_buffer()             # get its textbuffer
        if b:
            f = b.get_data("file_name")
            t = b.get_data("current_type")
            m = b.get_data("current_meta")

            # check if selected = instance, in non-objects mutable package file
            mb = m == 0 and f != obj.prefs.objects_name and obj.pack2.__class__ == dz_mpack.MutablePackage

            if mb:
                g = b.get_data("current_group")
                i = b.get_data("current_instance")
                s = b.get_data("current_subtype")
                r = obj.pack2.find({'type': t, 'group': g, 'subtype':s, 'instance': i})
                if r:
                    b = obj.pack2.get(r['offset'], r['size'])
                    if b:
                        do_update_hex_view(obj, b)

                if t == TYPE_BCON:
                    x = 1
                elif t == TYPE_STR or t == TYPE_CATS or t == TYPE_CTSS or t == TYPE_TTAs:
                    x = 2
                elif t == TYPE_BHAV:
                    x = 3

    obj.wTree.get_widget("edit_notebook").set_current_page(x)       # set edit page
    if mb:
        obj.wTree.get_widget("edit_notebook").set_sensitive(True)   # enable editor views
        obj.wTree.get_widget("hex_vbox").set_sensitive(True)
        obj.wTree.get_widget("resource_vbox").set_sensitive(True)
    else:
        obj.wTree.get_widget("edit_notebook").set_sensitive(False)  # disable editor views
        obj.wTree.get_widget("hex_vbox").set_sensitive(False)
        obj.wTree.get_widget("resource_vbox").set_sensitive(False)
