# navigation handling

import gtk
import logging
import os

import dz_cont
import dz_label
import dz_util
from dz_defs import *

def do_tree_style_next(tree, x):
    if tree.iter_has_child(x):          # if has child, child = next
        x = tree.iter_children(x)
    elif tree.iter_next(x) == None:     # if no child and no sibling
        x = tree.iter_parent(x)         # then back up to parent
        while x != None and tree.iter_next(x) == None:
            x = tree.iter_parent(x)     # find next parent that has a sibling
        if x != None:
            x = tree.iter_next(x)       # next = sibling
    else:
        x = tree.iter_next(x)           # next = sibling
    return x

def do_find_in_ct(obj, d):
    bt, bs, bg, bi, bm = False, False, False, False, False
    if 'type' in d:
        bt = True
        mType = d['type']
    if 'subtype' in d:
        bs = True
        mSubType = d['subtype']
    if 'group' in d:
        bg = True
        mGroup = d['group']
    if 'instance' in d:
        bi = True
        mInstance = d['instance']
    if 'meta' in d:
        bm = True
        mMeta = d['meta']

    tv = obj.current_cv
    m = tv.get_model()
    j = m.get_iter_root()

    while j:
        j2 = j              # save this iter in case this is a find
        eGroup = m.get_value(j, 1)
        eType = m.get_value(j, 2)
        eSubType = m.get_value(j, 3)
        eInstance = m.get_value(j, 4)
        eMeta = m.get_value(j, 5)

        j = do_tree_style_next(m, j)

        if bg and mGroup != eGroup:
            continue        # loop to while j if this test fails
        if bt and mType != eType:
            continue
        if bi and mInstance != eInstance:
            continue
        if bs and mSubType != eSubType:
            continue
        if bm and mMeta != eMeta:
            continue

        return j2           # return saved iter

    return None     # not found

def do_set_ct_cursor(obj, it, meta):
    if it:
        tv = obj.current_cv

        m = tv.get_model()
        if meta == 0:
            tv.set_cursor(m.get_path(it))
        else:
            if not tv.row_expanded(m.get_path(it)):
                tv.expand_row(m.get_path(it), False)
        obj.wait_gtk()

def do_goto(obj, instance, rtype):
    # note: don't forget to set cv_event first
    p = obj.get_current_pack()

    nb = obj.wTree.get_widget("notebook2")
    sw = nb.get_nth_page(nb.get_current_page())     # current scrollwindow
    b = sw.get_child().get_buffer()                 # textbuffer
    g = b.get_data("current_group")

    if instance < 0x1000:
        g = GROUP_GLOBAL
    elif instance >= 0x2000 and instance < 0x3000:
        g = dz_util.get_semi(p, g)

    # check if this reference is global + external
    if not p.is_objects and p.objects:
        r = p.find({'type': rtype, 'group': g, 'instance': instance})
        if not r:
            g2 = dz_util.get_semi(p, g)             # get this semi-global

            p = obj.pack1                          # switch over to global
            scv = obj.cv_event                     # save the trigger
            obj.set_content_page(0)
            obj.wait_gtk()
            obj.cv_event = scv

            if instance >= 0x1000 and instance < 0x2000 and g2:
                j = do_find_in_ct(obj, {'group': g2, 'meta': 1})    # find group header
                if j:
                    do_set_ct_cursor(obj, j, 1)

    # find the indicated group in the content tree
    j = do_find_in_ct(obj, {'group': g, 'meta': 1})     # find group header
    if j:
        # expand this group if it isn't already
        do_set_ct_cursor(obj, j, 1)

        # find specified type of the indicated group
        j = do_find_in_ct(obj, {'group': g, 'type': rtype, 'meta': 3})  # find type header
        if j:
            do_set_ct_cursor(obj, j, 1)

            # find specified instance, type, group
            j = do_find_in_ct(obj, {'group': g, 'type': rtype, 'instance': instance})  # find instance
            do_set_ct_cursor(obj, j, 0)

def do_goto_value(obj, group, rtype, subtype, instance, ref):
    p = obj.get_current_pack()

    # ref = check if reference is global + external, switch if necessary
    if ref and not p.is_objects and p.objects:
        r = p.find({'group': group})    # note: only check group for now
        if not r:
            p = obj.pack1                          # switch over to global
            scv = obj.cv_event                     # save the trigger
            obj.set_content_page(0)
            obj.wait_gtk()
            obj.cv_event = scv

    # expand semi-global parent if necessary
    g = dz_util.get_semi(p, group)
    if g and g != group:
        j = do_find_in_ct(obj, {'group': g, 'meta': 1})
        if j:
            do_set_ct_cursor(obj, j, 1)         # expand semi-global group

    j = do_find_in_ct(obj, {'group': group, 'meta': 1})
    if j:
        if rtype:
            do_set_ct_cursor(obj, j, 1)         # expand group
            j = do_find_in_ct(obj, {'group': group, 'type': rtype, 'meta': 3})
            if j and instance:
                do_set_ct_cursor(obj, j, 1)     # expand type
                j = do_find_in_ct(obj, {'group': group, 'type': rtype, 'instance': instance})

        do_set_ct_cursor(obj, j, 0)     # set_cursor (no expand)

def do_set_recents(obj):
    # verify recent files can open
    i = 0
    while i < len(obj.prefs.recent_files):
        try:
            f = open(obj.prefs.recent_files[i], 'rb')
        except:
            f = None
        if f:
            f.close()
            i += 1
        else:
            obj.prefs.recent_files.pop(i)

    # put holiday packs in recents (move to File menu?)
    s = obj.prefs.recent_files[:]
    if obj.prefs.holiday1:
        s.insert(0, obj.prefs.holiday1)
    if obj.prefs.holiday2:
        s.insert(0, obj.prefs.holiday2)

    # set menu
    x = len(s)
    if x > 0:
        w = obj.wTree.get_widget("openbutton")
        m = gtk.Menu()
        for i in range(x):
            h, t = os.path.split(s[i])
            e = gtk.MenuItem(t, use_underline=False)
            m.append(e)
            e.connect("activate", obj.on_openrecent_activate, i)
            e.show()
        w.set_menu(m)

def do_remove_nav_stack(obj, index):
    x = index
    while x+1 in obj.nav_stacks:   # move stacks starting from index down one
        obj.nav_stacks[x] = obj.nav_stacks[x+1]
        obj.nav_cursor[x] = obj.nav_cursor[x+1]
        x += 1
    obj.nav_stacks[x] = []         # clear last stack
    obj.nav_cursor[x] = 0

def do_insert_nav_stack(obj, index):
    x = index
    while x+1 in obj.nav_stacks:   # find last stack
        x += 1
    while x != index:               # move stacks starting from last up one
        obj.nav_stacks[x+1] = obj.nav_stacks[x]
        obj.nav_cursor[x+1] = obj.nav_cursor[x]
        x -= 1
    obj.nav_stacks[x] = []         # clear stack at index
    obj.nav_cursor[x] = 0

def get_nav_stack(obj, page_num):
    if page_num in obj.nav_stacks:
        return obj.nav_stacks[page_num]
    return None

def set_nav_buttons(obj, page_num):
    b1 = False
    b2 = False
    x = get_nav_stack(obj, page_num)
    if x:
        if obj.nav_cursor[page_num] != 0:
            b1 = True
        if obj.nav_cursor[page_num] != len(x) - 1:
            b2 = True
    obj.wTree.get_widget("backbutton").set_sensitive(b1)
    if b1:
        w = obj.wTree.get_widget("backbutton")
        m = gtk.Menu()
        for i in range(obj.nav_cursor[page_num] - 1, -1, -1):
            m2 = x[i][0].get_model()
            e = gtk.MenuItem(m2.get_value(m2.get_iter(x[i][1]), 0))
            m.append(e)
            e.connect("activate", obj.on_backmenu_activate, i)
            e.show()
        w.set_menu(m)
    obj.wTree.get_widget("forwardbutton").set_sensitive(b2)
    if b2:
        w = obj.wTree.get_widget("forwardbutton")
        m = gtk.Menu()
        for i in range(obj.nav_cursor[page_num] + 1, len(x)):
            m2 = x[i][0].get_model()
            e = gtk.MenuItem(m2.get_value(m2.get_iter(x[i][1]), 0))
            m.append(e)
            e.connect("activate", obj.on_forwardmenu_activate, i)
            e.show()
        w.set_menu(m)

def set_nav_stack(obj, page_num, path):
    x = get_nav_stack(obj, page_num)
    if x:
        if path in x:       # note: path = (treeview, path)
            x.remove(path)  # remove old reference
        if len(x) > obj.nav_size:
            x.pop(0)        # remove bottom entry if stack is full
    else:
        x = []
    x.append(path)
    obj.nav_stacks[page_num] = x
    obj.nav_cursor[page_num] = len(x) - 1
    set_nav_buttons(obj, page_num)

def do_nav_indexed(obj, c):
    nb = obj.wTree.get_widget("notebook2")
    p = nb.get_current_page()
    x = get_nav_stack(obj, p)
    if x:
        # check if selected treeview is current, switch if not
        if x[c][0] != obj.current_cv:
            if x[c][0] == obj.wTree.get_widget("obj_contentview"):
                obj.set_content_page(0)
            else:
                obj.set_content_page(1)
            obj.wait_gtk()

        # set treeview cursor
        obj.cv_event = -1                  # quiet mode
        x[c][0].set_cursor(x[c][1])         # treeview.set_cursor(path)

        obj.nav_cursor[p] = c              # set navigation cursor
        set_nav_buttons(obj, p)
        dz_cont.do_contentview_change(obj, p)

def do_new_tab(obj):
    logging.info("new tab")

    nb = obj.wTree.get_widget("notebook2")
    if nb.get_n_pages() == 0:
        obj.wTree.get_widget("close_tab1").set_sensitive(True)

    sw = gtk.ScrolledWindow()
    sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    xv = gtk.TextView()
    xv.connect("motion-notify-event", obj.on_motion_xv, sw)
    xv.connect("button-press-event", obj.on_button_xv, sw)
    sw.add(xv)                      # add text view to scroll window
    sw.show_all()
    xv.set_editable(False)
    xv.set_cursor_visible(False)

    lb = dz_label.label(obj.blank_tab)
    lb.show_all()
    lb.button.connect("clicked", obj.on_label_button_clicked, sw)

    nb.append_page(sw, lb)          # add a page
    nb.set_current_page(-1)         # select it

    # set this navigation stack
    obj.nav_stacks[nb.get_current_page()] = []
    obj.nav_cursor[nb.get_current_page()] = 0

    # default = no history/navigation cursor
    obj.wTree.get_widget("backbutton").set_sensitive(False)
    obj.wTree.get_widget("forwardbutton").set_sensitive(False)

    obj.wTree.get_widget("edit_notebook").set_current_page(0)  # Editor = None

    return nb.get_n_pages() - 1     # new = last page
