# preferences and dynamic definitions handler

import cPickle
import gtk
import logging
import os

import dz_pack
import dz_sys
import dz_util
from dz_defs import *

class Prefs:
    def __init__(self):
        # default values
        self.dump_one = True
        self.dump_name = "dump.txt"
        self.dump_mode = 0
        self.objects_name = "objects.package"
        self.wants_name = "Wants.package"
        self.maximize = False
        self.w_size = None
        self.ccode = 1                  # English US
        self.no_str_lookup = False
        self.wrap = 0                   # default = none
        self.view_mode = 0              # view code
        self.font_name = "monospace 10" # default font
        self.recent_files = []
        self.holiday1 = None            # one of the holiday packs
        self.holiday2 = None            # note: the order doesn't matter
        self.view_unknown = False       # view any type
        self.view_toolbar = True        # show/hide toolbar
        self.view_editbar = False       # show/hide editbar
        self.hpaned_pos = -1            # position of content/viewer divider
        self.vpaned_pos = -1            # position of viewer/editor divider
        self.tree_font = "sans 10"      # default tree font

        # load preferences
        try:
            logging.info("loading prefs")
            f = open("prefs", 'r')
        except:
            logging.info("prefs not found, using defaults")
            f = None
        if f:
            try:
                d = cPickle.load(f)     # load prefs dict
            except:
                logging.error("prefs failed to load, using defaults")
                d = None
            f.close()
            if d:
                try:
                    self.dump_one = d['dump_one']
                    self.dump_name = d['dump_name']
                    self.dump_mode = d['dump_mode']
                    self.objects_name = d['objects_name']
                    self.wants_name = d['wants_name']
                    self.maximize = d['maximize']
                    self.w_size = d['w_size']
                    self.ccode = d['ccode']
                    self.wrap = d['wrap']
                    self.view_mode = d['view_mode']
                    self.font_name = d['font_name']
                    self.recent_files = d['recent_files']
                    self.holiday1 = d['holiday1']
                    self.holiday2 = d['holiday2']
                    self.view_unknown = d['view_unknown']
                    self.view_toolbar = d['view_toolbar']
                    self.view_editbar = d['view_editbar']
                    self.hpaned_pos = d['hpaned_pos']
                    self.vpaned_pos = d['vpaned_pos']
                    self.tree_font = d['tree_font']
                except:
                    logging.warning("prefs outdated, some values may contain defaults")
        else:
            # get the default paths (and country code) from sys
            p, p2, cc = dz_sys.get_default_path()
            logging.info("sys defaults: path = %s, base = %s, ccode = %d" % (p, p2, cc))
            s = os.sep
            if p:
                self.ccode = cc
                p = p + "TSData" + s + "Res" + s
                self.objects_name = p + "Objects" + s + "objects.package"
                self.wants_name = p + "Wants" + s + "Wants.package"
                if p2:
                    p2 = p2 + "TSData" + s + "Res" + s + "Catalog" + s + "Bins" + s
                    self.holiday1 = p2 + "H05.bundle.package"
                    self.holiday2 = p2 + "H06.bundle.package"
                    try:
                        f = open(self.holiday1, 'rb')
                        f.close()
                        logging.info("holiday pack %s found" % self.holiday1)
                    except:
                        self.holiday1 = None
                    try:
                        f = open(self.holiday2, 'rb')
                        f.close()
                        logging.info("holiday pack %s found" % self.holiday2)
                    except:
                        self.holiday2 = None

    # save preferences
    def save(self):
        d = dict(dump_one=self.dump_one, dump_name=self.dump_name, \
            dump_mode=self.dump_mode, objects_name=self.objects_name, \
            wants_name=self.wants_name, maximize=self.maximize, \
            w_size=self.w_size, ccode=self.ccode, wrap=self.wrap, \
            view_mode=self.view_mode, font_name=self.font_name, \
            recent_files=self.recent_files, holiday1=self.holiday1, \
            holiday2=self.holiday2, view_unknown=self.view_unknown, \
            view_toolbar=self.view_toolbar, view_editbar=self.view_editbar, \
            hpaned_pos=self.hpaned_pos, vpaned_pos=self.vpaned_pos, \
            tree_font=self.tree_font)
        try:
            logging.info("saving prefs")
            f = open("prefs", 'w')
        except:
            logging.error("prefs failed to open for writing")
            f = None
        if f:
            try:
                cPickle.dump(d, f)       # save prefs dict
            except:
                logging.error("prefs failed to save")
            f.close()

    # load definitions
    def loaddef(self):
        # default = all results give IndexErrors
        self.glob_bhav_names = []
        self.glob_bhav_params = []
        self.glob_bcon_names = []
        self.glob_bcon_values = []
        #self.glob_named_tree = []
        self.sg_names = []
        self.sg_ids = []
        self.object_names = []
        self.object_group = []
        self.object_guids = []
        self.object_semi = []
        self.wants_guids = []
        self.wants_types = []
        self.wants_names = []
        self.holiday_names = []
        self.holiday_group = []
        self.holiday_guids = []
        self.holiday_semi = []

        try:
            logging.info("loading defs")
            f = open("defs", 'r')
        except:
            logging.warning("no defs found")
            f = None
        if f:
            try:
                d = cPickle.load(f)     # load defs dict
            except:
                logging.error("defs failed to load")
                d = None
            f.close()
            if d:
                try:
                    self.glob_bhav_names = d['glob_bhav_names']
                    self.glob_bhav_params = d['glob_bhav_params']
                    self.glob_bcon_names = d['glob_bcon_names']
                    self.glob_bcon_values = d['glob_bcon_values']
                    self.sg_names = d['sg_names']
                    self.sg_ids = d['sg_ids']
                    self.object_names = d['object_names']
                    self.object_group = d['object_group']
                    self.object_semi = d['object_semi']
                    self.object_guids = d['object_guids']
                    self.wants_guids = d['wants_guids']
                    self.wants_types = d['wants_types']
                    self.wants_names = d['wants_names']
                    self.holiday_names = d['holiday_names']
                    self.holiday_group = d['holiday_group']
                    self.holiday_guids = d['holiday_guids']
                    self.holiday_semi = d['holiday_semi']
                except:
                    logging.warning("defs invalidated")
                    self.glob_bhav_names = []

    # refresh and save definitions
    def reconfig(self, main):
        # default = all results give IndexErrors
        self.glob_bhav_names = []
        self.glob_bhav_params = []
        self.glob_bcon_names = []
        self.glob_bcon_values = []
        self.sg_names = []
        self.sg_ids = []
        self.object_names = []
        self.object_group = []
        self.object_guids = []
        self.object_semi = []
        self.wants_guids = []
        self.wants_types = []
        self.wants_names = []

        self.holiday_names = []
        self.holiday_group = []
        self.holiday_guids = []
        self.holiday_semi = []

        main.progress_show = True
        main.progress_amount = 0
        main.progress_divisor = 100

        main.wait_gtk()

        # extract definitions from objects
        p1 = dz_pack.Package(self.objects_name, self, True)
        if p1.toc_valid:

            main.status_text = "Reading global behavior names..."
            main.progress_amount = 0
            main.wait_gtk()

            # get global BHAV names, params
            r = p1.find({'type': TYPE_BHAV, 'group': GROUP_GLOBAL})
            if not r:
                logging.critical("global BHAVs not found")
            while r:    # note: this depends on global BHAVs starting at 0x100 and not containing holes
                b = p1.get(r['offset'], r['size'])
                if b:
                    j = 0x40
                    if '\x00' in b[:0x40]: j = b.index('\x00')
                    self.glob_bhav_names.append(b[:j])          # title
                    self.glob_bhav_params.append(ord(b[0x45]))  # params
                r = p1.find({'type': TYPE_BHAV, 'group': GROUP_GLOBAL, 'index': r['index']})

            # get a list of global BCON instances
            s1 = []
            r = p1.find({'type': TYPE_BCON, 'group': GROUP_GLOBAL})
            if not r:
                logging.critical("global BCONs not found")
            while r:
                s1.append(r['instance'])
                r = p1.find({'type': TYPE_BCON, 'group': GROUP_GLOBAL, 'index': r['index']})

            # get global BCON names, values
            i = 0x100
            while i <= max(s1):
                r = p1.find({'type': TYPE_BCON, 'group': GROUP_GLOBAL, 'instance': i})
                if r:
                    b = p1.get(r['offset'], r['size'])
                    if b:
                        j = 0x40
                        if '\x00' in b[:0x40]: j = b.index('\x00')
                        self.glob_bcon_names.append(b[:j])          # title
                        s = []
                        wx = dz_util.s2i(b[0x40:0x42]) & 0x7FFF     # count
                        for x in range(wx):
                            j = 0x42 + (x + x)
                            s.append(dz_util.s2i(b[j:j+2]))
                        self.glob_bcon_values.append(s)             # values
                else:
                    self.glob_bcon_names.append("")     # BCON hole
                    self.glob_bcon_values.append([])
                i += 1
                r = p1.find({'type': TYPE_BCON, 'group': GROUP_GLOBAL, 'instance': i})

            logging.info("last global BCON = 0x%X" % (i-1))

            main.status_text = "Reading object names..."
            main.progress_amount = 5
            main.wait_gtk()

            # get names, groups
            s1 = []
            r = p1.find({'type': TYPE_NREF})
            while r:
                b = p1.get(r['offset'], r['size'])
                if b:
                    self.object_names.append(b)             # object name
                    self.object_group.append(r['group'])
                    s1.append(r['instance'])
                r = p1.find({'type': TYPE_NREF, 'index': r['index']})

            main.status_text = "Reading object GUIDs..."
            main.progress_amount = 10
            main.wait_gtk()

            # get GUIDs
            r = p1.find({'type': TYPE_OBJD})
            s2, s2g, s2i = [], [], []
            while r:
                b = p1.get(r['offset'], r['size'])
                if b:
                    s2.append(dz_util.s2i(b[0x5C:0x60]))   # GUID
                    s2g.append(r['group'])
                    s2i.append(r['instance'])
                r = p1.find({'type': TYPE_OBJD, 'index': r['index']})

            main.status_text = "Sorting object GUIDs..."
            main.progress_amount = 20
            main.wait_gtk()

            # sort GUIDs by NREF
            for i in range(len(s1)):
                for j in range(len(s2)):
                    if s1[i] != s2i[j]:
                        continue    # next item in j loop
                    if self.object_group[i] != s2g[j]:
                        continue
                    self.object_guids.append(s2[j])
                    break           # terminate j loop

            main.status_text = "Reading object semi-globals..."
            main.progress_amount = 30
            main.wait_gtk()

            # get semi-global names of all the groups that use them
            s2, s2g = [], []
            r = p1.find({'type': TYPE_GLOB})
            while r:
                b = p1.get(r['offset'], r['size'])
                if b:
                    k = ord(b[0x40])        # length (of pstring)
                    if ord(b[len(b)-1]) == 0:
                        s2.append(b[0x40:len(b)-1])  # cstring, actually
                    else:
                        s2.append(b[0x41:0x41+k])
                    s2g.append(r['group'])
                r = p1.find({'type': TYPE_GLOB, 'index': r['index']})

            main.status_text = "Sorting object semi-globals..."
            main.progress_amount = 40
            main.wait_gtk()

            # find semi-global values for each object that uses them
            for i in range(len(s1)):
                x = False
                for j in range(len(s2)):
                    if self.object_group[i] == s2g[j]:
                        self.object_semi.append(dz_util.calc_crc24(s2[j]) + 0x7F000000)
                        x = True
                        break       # terminate the j loop
                if not x:
                    self.object_semi.append(0)  # 0 = no semi for this object

            # get list of unique semi-global names
            for i in range(len(s2)):
                if s2[i] not in self.sg_names:
                    self.sg_names.append(s2[i])

        # get holiday data

        ph1 = dz_pack.Package(self.holiday1, self, False)
        ph2 = dz_pack.Package(self.holiday2, self, False)
        if ph1.toc_valid or ph2.toc_valid:

            main.status_text = "Reading holiday object names..."
            main.progress_amount = 50
            main.wait_gtk()

            # get names, groups
            s1 = []
            if ph1.toc_valid:
                r = ph1.find({'type': TYPE_NREF})
                while r:
                    b = ph1.get(r['offset'], r['size'])
                    if b:
                        self.holiday_names.append(b)             # object name
                        self.holiday_group.append(r['group'])
                        s1.append(r['instance'])
                    r = ph1.find({'type': TYPE_NREF, 'index': r['index']})
            if ph2.toc_valid:
                r = ph2.find({'type': TYPE_NREF})
                while r:
                    b = ph2.get(r['offset'], r['size'])
                    if b:
                        self.holiday_names.append(b)             # object name
                        self.holiday_group.append(r['group'])
                        s1.append(r['instance'])
                    r = ph2.find({'type': TYPE_NREF, 'index': r['index']})

            main.status_text = "Reading holiday object GUIDs..."
            main.progress_amount = 60
            main.wait_gtk()

            # get GUIDs
            s2, s2g, s2i = [], [], []
            if ph1.toc_valid:
                r = ph1.find({'type': TYPE_OBJD})
                while r:
                    b = ph1.get(r['offset'], r['size'])
                    if b:
                        s2.append(dz_util.s2i(b[0x5C:0x60]))   # GUID
                        s2g.append(r['group'])
                        s2i.append(r['instance'])
                    r = ph1.find({'type': TYPE_OBJD, 'index': r['index']})
            if ph2.toc_valid:
                r = ph2.find({'type': TYPE_OBJD})
                while r:
                    b = ph2.get(r['offset'], r['size'])
                    if b:
                        s2.append(dz_util.s2i(b[0x5C:0x60]))   # GUID
                        s2g.append(r['group'])
                        s2i.append(r['instance'])
                    r = ph2.find({'type': TYPE_OBJD, 'index': r['index']})

            main.status_text = "Sorting holiday object GUIDs..."
            main.progress_amount = 70
            main.wait_gtk()

            # sort GUIDs by NREF
            for i in range(len(s1)):
                for j in range(len(s2)):
                    if s1[i] != s2i[j]:
                        continue    # next item in j loop
                    if self.holiday_group[i] != s2g[j]:
                        continue
                    self.holiday_guids.append(s2[j])
                    break           # terminate j loop

            main.status_text = "Reading holiday object semi-globals..."
            main.progress_amount = 80
            main.wait_gtk()

            # get semi-global names of all the groups that use them
            s2, s2g = [], []
            if ph1.toc_valid:
                r = ph1.find({'type': TYPE_GLOB})
                while r:
                    b = ph1.get(r['offset'], r['size'])
                    if b:
                        k = ord(b[0x40])        # length (of pstring)
                        if ord(b[len(b)-1]) == 0:
                            s2.append(b[0x40:len(b)-1])  # cstring, actually
                        else:
                            s2.append(b[0x41:0x41+k])
                        s2g.append(r['group'])
                    r = ph1.find({'type': TYPE_GLOB, 'index': r['index']})
            if ph2.toc_valid:
                r = ph2.find({'type': TYPE_GLOB})
                while r:
                    b = ph2.get(r['offset'], r['size'])
                    if b:
                        k = ord(b[0x40])        # length (of pstring)
                        if ord(b[len(b)-1]) == 0:
                            s2.append(b[0x40:len(b)-1])  # cstring, actually
                        else:
                            s2.append(b[0x41:0x41+k])
                        s2g.append(r['group'])
                    r = ph2.find({'type': TYPE_GLOB, 'index': r['index']})

            main.status_text = "Sorting holiday object semi-globals..."
            main.progress_amount = 90
            main.wait_gtk()

            # find semi-global values for each object that uses them
            for i in range(len(s1)):
                x = False
                for j in range(len(s2)):
                    if self.holiday_group[i] == s2g[j]:
                        self.holiday_semi.append(dz_util.calc_crc24(s2[j]) + 0x7F000000)
                        x = True
                        break       # terminate the j loop
                if not x:
                    self.holiday_semi.append(0)  # 0 = no semi for this object

            # get list of unique semi-global names
            for i in range(len(s2)):
                if s2[i] not in self.sg_names:
                    self.sg_names.append(s2[i])

        # calculate semi-global ids
        for i in range(len(self.sg_names)):
            self.sg_ids.append(dz_util.calc_crc24(self.sg_names[i]) + 0x7F000000)

        # get definitions from Wants
        p2 = dz_pack.Package(self.wants_name, self, False)
        if p2.toc_valid:

            main.status_text = "Reading wants..."
            main.progress_amount = 95
            main.wait_gtk()

            r = p2.find({'type': TYPE_WXML})
            while r:
                b = p2.get(r['offset'], r['size'])
                if b:
                    self.wants_guids.append(r['instance'])   # GUID
                    if '"nodeText"' in b:
                        j = b.index('"nodeText"')
                        x = b.index('>', j)
                        y = b.index('<', j)
                        self.wants_names.append(b[x+1:y])         # name
                    if '"objectType"' in b:
                        j = b.index('"objectType"')
                        x = b.index('>', j)
                        y = b.index('<', j)
                        s = b[x+1:y]
                        if s == "None":
                            i = Want_None
                        elif s == "Sim":
                            i = Want_Sim
                        elif s == "Guid":
                            i = Want_Guid
                        elif s == "Category":
                            i = Want_Category
                        elif s == "Skill":
                            i = Want_Skill
                        elif s == "Career":
                            i = Want_Career
                        elif s == "Badge":
                            i = Want_Badge
                        else:
                            raise ValueError, "bad want type"
                        self.wants_types.append(i)                  # type
                r = p2.find({'type': TYPE_WXML, 'index': r['index']})

        main.status_text = ""
        main.progress_show = False
        main.wait_gtk()

        # save definitions
        d = dict(glob_bhav_names=self.glob_bhav_names, glob_bhav_params=self.glob_bhav_params, \
            glob_bcon_names=self.glob_bcon_names, glob_bcon_values=self.glob_bcon_values, \
            sg_names=self.sg_names, sg_ids=self.sg_ids,object_names=self.object_names, \
            object_group=self.object_group,object_semi=self.object_semi, \
            object_guids=self.object_guids,wants_guids=self.wants_guids, \
            wants_types=self.wants_types,wants_names=self.wants_names, \
            holiday_names=self.holiday_names,holiday_group=self.holiday_group, \
            holiday_guids=self.holiday_guids,holiday_semi=self.holiday_semi)
        try:
            f = open("defs", 'w')
        except:
            f = None
            logging.error("defs failed to open for writing")
        if f:
            try:
                cPickle.dump(d, f)       # save defs dict
            except:
                logging.error("defs failed to save")
            f.close()
