# primitive BHAV operations
#
# get_data(package, owner, value): get the name of a data variable
# get_guid(package, value): get the value/name of an object by GUID
# get_title(package, instance): return the BHAV's title, etc. (None = not found)
# get_title2(package, instance): return the BHAV's title only
# get_etree(package, instance, type): return the title of an event tree
# get_prim(package, opcode, buffer, index, format, node_ver): get output for a primitive
# prim_00 .. prim_7E(package, buffer, index, format, node_ver): get operand output

import dz_bcon
import dz_prefs
import dz_str
import dz_util
from dz_defs import *
from dz_glob import *

def get_data(p, o, x):
    """Return the name of a specific variable."""
    if o == 0 or o == 1:        # My Attribute or Stack Obj's Attribute
        return dz_util.get_html(p,TYPE_DATA,"%s 0x%X" % (gString84[o], x))
    elif o == 2 or o == 5:      # My Semi Attribute or Stack Objects Semi Attribute
        return dz_util.get_html(p,TYPE_DATA,"%s 0x%X" % (gString84[o], x))
    elif o == 3 or o == 4:      # My or Stack Object's
        if x == 0x0B:           # object's object ID (itself)
            if o == 3: return dz_util.get_html(p,TYPE_DATA,"Me")
            elif o == 4: return dz_util.get_html(p,TYPE_DATA,"Stack Object")
        else:
            return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gString8D, x), x))
    elif o == 6:                # Global (from Simulation)
        return dz_util.get_html(p,TYPE_DATA,"Global(%s)" % dz_util.chk_range(gString81, x))
    elif o == 7:                # Literal Value
        return dz_util.get_html(p,TYPE_DATA,dz_util.ext16(x))
    elif o == 8 or o == 9 or o == 0x19:     # Temp or Param or Local
        return dz_util.get_html(p,TYPE_DATA,"%s %d" % (gString84[o], x))
    elif o == 0x0A:             # Stack Object
        if x == 0:
            return dz_util.get_html(p,TYPE_DATA,"Stack Object")
        else:
            return dz_util.get_html(p,TYPE_DATA,"invalid")
    elif o == 0x0B or o == 0x11 or o == 0x1E or o == 0x1F or o == 0x30 or o == 0x31:
        # data owner [temp x] (temp x = data index)
        return dz_util.get_html(p,TYPE_DATA,"%s [Temp %d]" % (gString84[o], x))
    elif o == 0x0C or o == 0x0E or o == 0x0F or o == 0x1C or o == 0x1D:
        # motives/check tree ad
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gString86, x), x))
    elif o == 0x12 or o == 0x13 or o == 0x20:   # my/stack object's/neighbor's person data
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringC8, x), x))
    elif o == 0x15 or o == 0x26 or o == 0x33:   # Object Definition
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringCC, x), x))
    elif o == 0x17:             # room [temp 0]
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringDB, x), x))
    elif o == 0x18:             # neighbor in stack object
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringDD, x), x))
    elif o == 0x1A:             # Constant Value
        a = x >> 13             # x = aaabbbbb bccccccc
        b = (x >> 7) & 0x3F
        c = x & 0x7F
        if (a & 4) != 0:        # extended
            b += 0x40
            a &= 3
        if a == 0:              # private
            b += 0x1000
        elif a == 1:            # semi-global
            b += 0x2000
        elif a == 2:            # global
            b += 0x100
        elif a == 3:            # FUBAR
            b == 0xF5BA
        return dz_util.get_html(p,TYPE_DATA,"%s 0x%X" % (gString84[o], b) + dz_bcon.get_const(p, b, c, None))
    elif o == 0x2F:             # Constants [temp]
        a = x >> 13             # x = aaabbbbb bbbbbccc
        b = (x >> 3) & 0x3FF
        c = x & 7
        if (a & 4) != 0:        # extended
            b += 0x40
            a &= 3
        if a == 0:              # private
            b += 0x1000
        elif a == 1:            # semi-global
            b += 0x2000
        elif a == 2:            # global
            b += 0x100
        elif a == 3:            # FUBAR
            b == 0xF5BA
        return dz_util.get_html(p,TYPE_DATA,"%s 0x%X" % (gString84[o], b) + dz_bcon.get_const(p, b, -1, None) + \
            ":[Temp %d]" % c)
    elif o == 0x21:             # job data [temp 0,1]
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringF3, x), x))
    elif o == 0x22:             # neighborhood data
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringF9, x), x))
    elif o == 0x23:             # stack object's function
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringF5, x), x))
    elif o == 0x27 or o == 0x28:    # Temp or Stack Object's Temp Token
        return dz_util.get_html(p,TYPE_DATA,"%s %s (0x%X)" % (gString84[o], dz_util.chk_range(gStringFC, x), x))
    elif o < 0x33:              # other owners
        # 0x0D, 0x10, 0x14, 0x16, 0x1B, 0x24, 0x25, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x32
        return dz_util.get_html(p,TYPE_DATA,"%s %s" % (gString84[o], dz_util.ext16(x)))
    else:
        return dz_util.get_html(p,TYPE_DATA,"unknown 0x%X %d (0x%X)" % (o, x, x))

def get_guid(p, x):
    """Return the GUID number, group value and name of an object (if available)."""
    s = "GUID 0x%08X" % x
    pp = p.prefs
    g_names = None

    if x in pp.object_guids:
        i = pp.object_guids.index(x)
        g = pp.object_group[i]
        g_names = pp.object_names

    if x in pp.holiday_guids:
        i = pp.holiday_guids.index(x)
        g = pp.holiday_group[i]
        g_names = pp.holiday_names

    if g_names:
        s += " "

        p.m_offset += len(s)
        s2 = dz_util.get_html(p,TYPE_LINK_GROUP,"(0x%X: %s)" % (g, g_names[i]), g)
        p.markup.append( (2, g, p.m_offset, p.m_offset + len(s2)) )
        s += s2
        p.m_offset += len(s2)

    return s

def get_title(p, i):
    """Return the BHAV's name, number of params, and target group."""
    k = len(p.prefs.glob_bhav_names)    # if global, check the global name cache
    if i < (k + 0x100):
        i -= 0x100
        d = dict(name=p.prefs.glob_bhav_names[i], params=p.prefs.glob_bhav_params[i],\
            group=GROUP_GLOBAL)
        return d

    g2 = p.group
    ci = p.instance
    if i >= 0x2000 and i < 0x3000 and ci >= 0x1000 and ci < 0x2000:
        g2 = p.semi                 # if private to sg call, use the sg group

    r = p.find({'type': TYPE_BHAV, 'group': g2, 'instance': i})
    if not r: return r              # not found

    x = 256
    if x > r['size']: x = r['size'] # check against actual raw size

    b = p.get2(r['offset'], x)      # get data out of a truncated buffer
    if not b: return b              # inexplicably missing (!)

    j = 0x40                        # get title (cstring, maxchars = 0x40 at index = 0)
    if '\x00' in b[:0x40]: j = b.index('\x00')

    s = "." + b[:j]

    if p.group != g2:               # sg call
        s = "0x%X" % g2 + s

    return dict(name=s, params=ord(b[0x45]), group=g2)

def get_title2(p, i):
    """Return the BHAV's name or coded string if not found."""
    d = get_title(p, i)
    if d:
        return d['name']
    else:
        if not p.is_objects and p.objects:
            p2 = p.objects
            p2.group = p.group
            p2.instance = p.instance
            p2.semi = p.semi
            d = get_title(p2, i)
            if d:
                return dz_util.get_html(p,TYPE_LINK_FUNCTION,d['name'],(d['group'], i))
        return "[BHAV 0x%X]" % i

def get_etree(p, i, t):
    """Return the name of an event behavior tree."""
    if i == 0:
        s = "No Event Tree"
        s2 = ""
    else:
        s = "Using "
        if t == 0:
            s += "Private"
        elif t == 1:
            s += "SemiGlobal"
        else:
            s += "Global"
        s += " Event Tree: "

        p.m_offset += len(s)
        s2 = get_title2(p, i)
        p.markup.append( (1, i, p.m_offset, p.m_offset + len(s2)) )
        p.m_offset += len(s2)
    return s+s2

# evaluate primitive

def get_prim(p, opcode, b, x, f, node_version):
    """Return a name and operand decoding for a specific primitive."""

    # check whether opcode is not an "implied" function
    if opcode != 2 and opcode != 8 and opcode != 0xB and opcode != 0xC \
        and opcode != 0x16 and opcode != 0x1E:
        s = dz_util.get_html(p,TYPE_FUNCTION,dz_util.chk_range(gString8B, opcode) + ": ")
    else:
        s = ""

    p.m_offset += len(s)

    # lateral this off to yet another function
    return s + prim_func[opcode](p, b, x, f, node_version)

# evaluate operands of specific primitives

def prim_00(p, b, x, f, node_version):
    """Sleep"""
    return "for %s ticks" % get_data(p, 9, ord(b[x]))   # Param

def prim_01(p, b, x, f, node_version):
    """Generic Sims Call"""
    c1 = ord(b[x])

    s = dz_util.get_html(p,TYPE_FUNCTION,dz_util.chk_range(gStringDC, c1))
    if c1 == 0:
        s += " Temp 0:neighborhood, Temp 1:evict, Temp 2:save lot, Temp 3:reset tutorial"
    elif c1 == 4:
        s += " Stack Obj:nID, Temp 0:familyID"
    elif c1 == 5:
        s += " Temp 0:familyID"
    elif c1 == 6:
        s += " Stack Obj:nID"
    elif c1 == 0x0D:
        s += " Stack Obj"
    elif c1 == 0x11:
        s += " Temp 0:lotID"
    elif c1 == 0x12:
        s += " Temp 0:speed"
    elif c1 == 0x15:
        s += " Temp 0:child nID, Temp 1:parent nID"
    elif c1 == 0x16:
        s += " Temp 0:new spouse nID, Temp 1:initial spouse nID"
    elif c1 == 0x17:
        s += " Temp 0:remove nID, Temp 1:relative nID"
    elif c1 == 0x18:
        s += " Temp 0:age"
    elif c1 == 0x19 or c1 == 0x28 or c1 == 0x29 or c1 == 0x2C or c1 == 0x31 or c1 == 0x33:
        s += " Temp 0"
    elif c1 == 0x1C:
        s += " Temp 0:amount, Temp 2:multiplier"
    elif c1 == 0x1D or c1 == 0x26:
        s += " Temp 0:nID"
    elif c1 == 0x1E:
        s += " Temp 0:outfit"
    elif c1 == 0x1F:
        s += " Temp 0,1:GUID"
    elif c1 == 0x20:
        s += " Temp 0:outfit, Temp 1:result value"
    elif c1 == 0x22:
        s += " Temp 0:source, Temp 1:destination, Temp 2:result value"
    elif c1 == 0x23 or c1 == 0x24:
        s += " Temp 0:lotID, Temp 1:result value"
    elif c1 == 0x2D:
        s += " Temp 0:source, Stack Obj:destination"
    elif c1 == 0x2E:
        s += " Temp 0:tableID, Temp 1:index, Temp 3:fallback"
    elif c1 == 0x30:
        s += " Temp 0:nID, Temp 1:result value"
    elif c1 == 0x32:
        s += " Temp 0:take, Temp 1:target nID, Temp 2:percent, Temp 3,4:amount, Temp 5:from assets"
    elif c1 == 0x34:
        s += " Temp 0:direction, Temp 1:result wall obj"
    elif c1 == 0x35:
        s += " Temp 0:tableID, Temp 1:index, Temp 2:state, Temp 3:fallback"
    return s

def prim_02(p, b, x, f, node_version):
    """(Evaluate) Expression"""
    w1 = dz_util.s2i(b[x:x+2])      # LHS value
    w2 = dz_util.s2i(b[x+2:x+4])    # RHS value
    c0 = ord(b[x+5])                # operator
    c1 = ord(b[x+6])                # LHS owner
    c2 = ord(b[x+7])                # RHS owner

    s = get_data(p, c1, w1)         # left hand side (result variable)

    if c0 == 0x14:                  # abs(rhs) operator
        s += dz_util.get_html(p,TYPE_OPERATOR," := abs(")
    elif c0 == 0x15:                # Assign 32bit Value
        s += dz_util.get_html(p,TYPE_OPERATOR," := int32(")
    else:
        s += dz_util.get_html(p,TYPE_OPERATOR," %s " % gString88[c0])     # operator name

    if c2 == 7 and c0 > 7 and c0 < 11 and w2 > 0:       # literal (non-0), flag operator
        if c1 == 3 or c1 == 4:          # My or Stack Object's
            if w1 == 5:
                s += dz_util.chk_range(gStringD0, w2 - 1)  # Wall adj. flags
            elif w1 == 8:
                s += dz_util.chk_range(gString8E, w2 - 1)  # Flags 1
            elif w1 == 0x0D:
                s += dz_util.chk_range(gStringE5, w2 - 1)  # Wall placement flags
            elif w1 == 0x22:
                s += dz_util.chk_range(gString200, w2 - 1) # Hidden Flags
            elif w1 == 0x28:
                s += dz_util.chk_range(gStringD6, w2 - 1)  # Flags 2
            elif w1 == 0x2A:
                s += dz_util.chk_range(gStringCA, w2 - 1)  # Placement flags
            elif w1 == 0x2B:
                s += dz_util.chk_range(gStringCB, w2 - 1)  # Movement flags
            elif w1 == 0x3F:
                s += dz_util.chk_range(gStringFB, w2 - 1)  # Exclusive placement flags
            elif w1 == 0x45:
                s += dz_util.chk_range(gStringFD, w2 - 1)  # Wall cutout flags
            else:
                s += get_data(p, c2, w2)    # unknown flag

        elif c1 == 0x12 or c1 == 0x13 or c1 == 0x20:    # my/stack object's/neighbor's person data
            if w1 == 0x1E:
                s += dz_util.chk_range(gStringB2, w2 - 1)  # Censorship flags
            elif w1 == 0x44:
                s += dz_util.chk_range(gString201, w2 - 1) # Ghost flags
            elif w1 == 0x51:
                s += dz_util.chk_range(gString8F, w2 - 1)  # Body flags
            elif w1 == 0x9E:
                s += dz_util.chk_range(gString202, w2 - 1) # Selection flags
            elif w1 == 0x9F:
                s += dz_util.chk_range(gString204, w2 - 1) # Person flags
            else:
                s += get_data(p, c2, w2)    # unknown flag

        elif c1 == 0x15 or c1 == 0x26 or c1 == 0x33:    # Object Definition
            if w1 == 0x27:
                s += dz_util.chk_range(gStringCD, w2 - 1)  # Room sort flags
            elif w1 == 0x28:
                s += dz_util.chk_range(gStringCE, w2 - 1)  # Function sort flags
            else:
                s += get_data(p, c2, w2)    # unknown flag

        else:
            s += get_data(p, c2, w2)    # unknown flag
    else:
        s += get_data(p, c2, w2)        # not a literal (non-0) flag
        if c0 == 0x14 or c0 == 0x15:    # abs(rhs) or Assign 32bit Value
            s += dz_util.get_html(p,TYPE_OPERATOR,")")
    return s

def prim_03(p, b, x, f, node_version):
    """Find Best Interaction (FBA)"""
    w1 = dz_util.s2i(b[x:x+2])          # motive
    w2 = dz_util.s2i(b[x+2:x+4])        # flags

    if w2 == 0:
        s = "for the N current worst motives."
    else:
        w3 = 1
        s = ""
        for c1 in range(16):
            if w1 & w3:
                s += " for %s." % gString86[c1]     # motives
            w3 = w3 << 1

        if w2 & 1:
            s += " Choose remaining motives from the worst."
        if w2 & 0x200:
            s + " Only test objects in room held in temp 0"
            if (w2 & 0x400) != 0:
                s += " including Out Of World Objects"
            s += "."
        if w2 & 0x800:
            s += " Find Nested Interactions Only"
            if w2 & 0x1000:
                s += " and only on current Interaction object"
            s += "."
    return s

def prim_04(p, b, x, f, node_version):
    """~(old)Grab"""
    return ""

def prim_05(p, b, x, f, node_version):
    """~(old)Drop"""
    return ""

def prim_06(p, b, x, f, node_version):
    """~(old)Change Suit"""
    return ""

def prim_07(p, b, x, f, node_version):
    """Refresh"""
    c1 = ord(b[x])      # target
    c2 = ord(b[x+2])    # property

    if c1 == 0:
        s = "My "
    else:
        s = "Stack Object's "

    if c2 == 0:
        s += "graphic"
    elif c2 == 1:
        s += "lighting contribution"
    else:
        s += "room score contribution"
    return s

def prim_08(p, b, x, f, node_version):
    """Random Number"""
    w1 = dz_util.s2i(b[x:x+2])          # result
    c1 = ord(b[x+2])
    w2 = dz_util.s2i(b[x+4:x+6])        # source
    c2 = ord(b[x+6])
    s = get_data(p, c1, w1) + " "
    s += dz_util.get_html(p,TYPE_OPERATOR,":=")
    s += " random from 0 to < %s" % get_data(p, c2, w2)
    return s

def prim_09(p, b, x, f, node_version):
    """~(old)Burn"""
    return ""

def prim_0A(p, b, x, f, node_version):
    """Tutorial (unused)"""
    return ""

def prim_0B(p, b, x, f, node_version):
    """Get Distance To"""
    w1 = dz_util.s2i(b[x:x+2])          # result (temp)
    sx = ord(b[x+2])                    # source flags
    c2 = ord(b[x+3])                    # source
    w2 = dz_util.s2i(b[x+4:x+6])
    ux = ord(b[x+6])                    # unit flags

    s = get_data(p, 8, w1) + " "
    s += dz_util.get_html(p,TYPE_OPERATOR,":=")
    s += " distance from "
    if sx & 1:
        s += get_data(p, c2, w2)
    else:
        s += "Me"
    s += " to Stack Object"

    if ux & 2:
        s += " in 1/100ths of a tile"
    return s

def prim_0C(p, b, x, f, node_version):
    """Get Direction To"""
    w1 = dz_util.s2i(b[x:x+2])          # result
    c1 = ord(b[x+2])
    sx = ord(b[x+4])                    # source flags
    c2 = ord(b[x+5])                    # source
    w2 = dz_util.s2i(b[x+6:x+8])
    ux = ord(b[x+8])                    # unit flags

    s = get_data(p, c1, w1) + " "
    s += dz_util.get_html(p,TYPE_OPERATOR,":=")
    s += " direction from "
    if sx & 1:
        s += get_data(p, c2, w2)
    else:
        s += "Me"
    s += " to Stack Object"

    if (ux & 2) == 0:
        s += " in degrees"
    return s

def prim_0D(p, b, x, f, node_version):
    """Push Interaction"""
    ix = ord(b[x])                      # literal #
    sx = ord(b[x+1])                    # object (local/param)
    px = ord(b[x+2])                    # priorities
    fx = ord(b[x+3])                    # flags
    ic = ord(b[x+4])                    # icon (local)
    c1 = ord(b[x+5])                    # interaction #
    w1 = dz_util.s2i(b[x+6:x+8])
    c2 = ord(b[x+8])                    # link
    w2 = dz_util.s2i(b[x+9:x+11])
    c3 = ord(b[x+11])                   # result
    w3 = dz_util.s2i(b[x+12:x+14])
    gx = ord(b[x+14])                   # flags 2
    ii = ord(b[x+15])                   # icon index

    if fx & 0x10:
        s = "getting interaction # from %s of " % get_data(p, c1, w1)
    elif gx & 2:
        s = "getting interaction # from results of last FBA call "
    else:
        s = "#0x%X of " % ix

    if fx & 2:
        s += get_data(p, 0x19, sx)      # local
    else:
        s += get_data(p, 9, sx)         # param
    s += " onto the stack object's queue, "

    s += dz_util.chk_range(gStringE0, px)              # Priorities

    if fx & 1:
        s += ", use icon from " + get_data(p, 0x19, ic)
    elif gx & 4:
        s += " use Icon from selector GUID in temp4/5"

    if gx & 8:
        s += ", Getting Icon Index from Temp6"
    else:
        s += ", Icon Index is 0x%X" % ii

    if fx & 8: s += ", use name"
    if fx & 0x20: s += ", Force run check tree"
    if gx & 1: s += ", Passing first 4 params in"
    if fx & 0x40:
        s += ", Linking to interaction with ID in " + get_data(p, c2, w2)
    if fx & 0x80:
        s += ", Returing ID in " + get_data(p, c3, w3)
    return s

def prim_0E(p, b, x, f, node_version):
    """Find Best Object for Function"""
    nx = ord(b[x])                      # function number
    fx = ord(b[x+2])                    # flags
    c1 = ord(b[x+3])                    # relative to object
    w1 = dz_util.s2i(b[x+4:x+6])

    s = dz_util.chk_range(gStringC9, nx)                   # Function table

    if fx & 1:
        s += ", Outside Only"
    elif fx & 2:
        s += ", Inside Only"
    elif fx & 4:
        s += ", In Room Only"

    if fx & 0x10:
        s += ", In Line Of Site"        # [sic]
    if fx & 0x40:
        s += ", Ignoring Lockout"
    if fx & 8:
        s += ", Relative to object in " + get_data(p, c1, w1)
    return s

def prim_0F(p, b, x, f, node_version):
    """Break Point"""
    w1 = dz_util.s2i(b[x:x+2])
    c1 = ord(b[x+2])
    fx = ord(b[x+4])

    if fx & 1:
        s = "(IGNORE IN RELEASE)"
    else:
        s = "if (%s " % get_data(p, c1, w1)
        s += dz_util.get_html(p,TYPE_OPERATOR,"!=")
        s += " 0)"
    return s

def prim_10(p, b, x, f, node_version):
    """Find Location For"""
    directions = ["N", "NE", "E", "SE", "S", "SW", "W", "NW"]
    gx = ord(b[x])                      # mode
    sx = ord(b[x+1])                    # start at (local)
    fx = ord(b[x+2])                    # flags
    dx = ord(b[x+3])                    # direction
    c1 = ord(b[x+4])                    # source
    w1 = dz_util.s2i(b[x+5:x+7])
    c2 = ord(b[x+7])                    # reference
    w2 = dz_util.s2i(b[x+8:x+10])
    vx = ord(b[x+10])                   # stack/local/slot var

    if fx & 8:
        s = "%s relative to %s, " % (get_data(p, c1, w1), get_data(p, c2, w2))
        if gx:
            s += gStringEF[gx]          # Find GLB
            if gx >= 5 and gx <= 8:
                s += " %d" % vx

        s += ", facing"
        k = 1
        for j in range(8):              # all tolerable directions
            if (dx & k) == 0:
                s += " " + directions[j]
            k = k << 1
    elif fx & 1:
        s = "Stack Object, start at " + get_data(p, 0x19, sx)
    else:
        s = "Stack Object"

    if fx & 4: s += ", user editable"
    if (fx & 2) == 0: s += ", prefer empty"
    if fx & 0x10: s += ", on level ground"
    if fx & 0x20: s += ", with empty border"
    if fx & 0x40: s += ", begin in front of refobj"
    if fx & 0x80: s += ", with line of site to center"  # [sic]
    return s

def prim_11(p, b, x, f, node_version):
    """Idle for Input"""
    dx = ord(b[x])                      # duration (param)
    w1 = dz_util.s2i(b[x+2:x+4])        # flags 1
    fx = ord(b[x+4])                    # flags 2

    if fx & 1:
        s = "Handle Sub Queue Interactions"
    else:
        s = "for %s ticks, " % get_data(p, 9, dx)
        if w1 == 0: s += "do not "
        s += "allow push"
    return s

def prim_12(p, b, x, f, node_version):
    """Remove Object Instance"""
    tx = ord(b[x])                      # target
    fx = ord(b[x+2])                    # flags

    if tx == 0:
        s = "Me"
    else:
        s = "Stack Object"

    if fx & 1: s += ", return immediately"
    if (fx & 2) == 0: s += ", cleanup all"
    return s

def prim_13(p, b, x, f, node_version):
    """Make New Character"""
    sx = ord(b[x])                      # skin (local)
    ax = ord(b[x+1])                    # age (local)
    gx = ord(b[x+2])                    # gender (local)
    c1 = ord(b[x+3])                    # parent 2
    w1 = dz_util.s2i(b[x+4:x+6])
    c2 = ord(b[x+6])                    # parent 1
    w2 = dz_util.s2i(b[x+7:x+9])
    fx = ord(b[x+9])                    # flags
    c3 = ord(b[x+10])                   # thumbnail outfit
    w3 = dz_util.s2i(b[x+11:x+13])

    s = ""      # in Soviet Russia, Character Makes New You!
    if fx & 8:
        s += "getting data from object with GUID in temp 0/1 "
    elif fx & 0x10:
        s += "getting data from object with GUID in temp Token "
    elif fx & 1:
        if fx & 2:
            s += "using Neighbor ID in "
        else:
            s += "using ID in "
        s += get_data(p, c2, w2) + " as parent 1 and "
        if fx & 4:
            s += "Neighbor ID in "
        else:
            s += "ID in "
        s += get_data(p, c1, w1) + " as parent 2 "

    if sx != 0 and sx != 0xFF:      # (this condition is weird)
        s += "age in " + get_data(p, 0x19, ax)
        s += ", gender in " + get_data(p, 0x19, gx)
        s += ", skin color in " + get_data(p, 0x19, sx)
    if fx & 0x20:
        s += ", Getting character from Bin" # (?)
    if fx & 0x40:
        s += ", Using external guid for Thumbnail Outfit from "
        if fx & 0x80:
            s += "GUID in temp 2/3"
        else:
            s += get_data(p, c3, w3)
    return s

def prim_14(p, b, x, f, node_version):
    """Run Functional Tree"""
    nx = ord(b[x])                      # function number
    fx = ord(b[x+2])                    # flags

    s = dz_util.get_html(p,TYPE_FUNCTION,dz_util.chk_range(gStringC9, nx))   # Functional table

    if fx & 1: s += ", change icon"
    if fx & 2: s += ", passing parameters from calling tree"
    if fx & 4: s += ", running check tree only"
    return s

def prim_15(p, b, x, f, node_version):
    """~Show String ( UNUSED )"""
    return ""

def prim_16(p, b, x, f, node_version):
    """Turn Body Towards"""
    c1 = ord(b[x])                      # value
    return gStringD8[c1]

def prim_17(p, b, x, f, node_version):
    """Play / Stop Sound Event"""
    w1 = dz_util.s2i(b[x:x+2])          # id
    w2 = dz_util.s2i(b[x+2:x+4])        # sample rate (?)
    fx = ord(b[x+4])                    # flags
    vx = ord(b[x+5])                    # volume (?)

    if fx & 0x40:
        s = "Stop Sound "
    else:
        s = "Play Sound "

    if w1 < 10000:
        s += "Private id = 0x%X" % w1
    elif w1 < 20000:
        s += "Global id = 0x%X" % (w1-10000)
    else:
        s += "SemiGlobal id = 0x%X" % (w1-20000)

    if fx & 2:
        s += ", with stack obj as source"
    else:
        if fx & 0x10:
            s += ", auto vary"
        if w2 != 0:
            s += ", sampled at " + dz_util.ext16(w2)
        if vx != 0:
            s += ", volume %d" % vx
    return s

def prim_18(p, b, x, f, node_version):
    """~UNUSED (was old relationship)"""
    return ""

def prim_19(p, b, x, f, node_version):
    """Alter Budget"""
    tx = ord(b[x])                      # type
    c1 = ord(b[x+1])                    # amount
    w1 = dz_util.s2i(b[x+2:x+4])
    fx = ord(b[x+4])                    # flags
    ex = ord(b[x+6])                    # expenses
    w2 = dz_util.s2i(b[x+7:x+9])        # multiplier

    if tx == 0:
        c1 = 7      # literal
    elif tx == 1:
        c1 = 9      # param
    elif tx == 2:
        c1 = 0x19   # local

    if fx & 3 == 0:
        s = "subtract "
    elif fx & 1:
        s = "test if "
    else:
        s = "add "

    if fx & 8 != 0 and node_version != 0:
        s += "amount in temp 2 and 3"
    else:
        s += get_data(p, c1, w1)

    if fx & 4:
        s += " Multiplied by value in Temp2"
    elif w2 != 0:
        s += " Multiplied by " + dz_util.ext16(w2)

    if fx & 3 == 1:
        s += " may be subtracted"
    if fx & 3 == 3:
        s += " may be added"
    s += " as " + dz_util.chk_range(gStringF0, ex)         # Expenses
    return s

def prim_1A(p, b, x, f, node_version):
    """Relationship"""
    vx = ord(b[x])                      # var
    fx = ord(b[x+1])                    # flags
    w2 = dz_util.s2i(b[x+6:x+8])        # source/dest or object 2 value

    if node_version == 0:               # old-style parameter usage
        gx = ord(b[x+2])                # flags 2
        ox = ord(b[x+3])                # rel object (local)
        c2 = ord(b[x+4])

        s = ""
        if fx & 4 == 0:
            s += get_data(p, c2, w2) + " := "   # dest (load rel)
        s += "var %d of " % vx
        if fx & 3 == 0:
            s += "Me to Stack Object"
        elif fx & 3 == 1:
            s += "Stack Object to Me"
        elif fx & 3 == 2:
            s += "Stack Object to " + get_data(p, 0x19, ox)
        else:
            s += get_data(p, 0x19, ox) + " to Stack Object"

        if fx & 4:
            s += " := " + get_data(p, c2, w2)   # source (store rel)
        if gx & 1: s += ", fail if too small"
        if gx & 2: s += ", use neighbor IDs"

    else:                               # new-style parameter usage
        c3 = ord(b[x+2])                # rel object 1
        w3 = dz_util.s2i(b[x+3:x+5])
        c2 = ord(b[x+5])                # rel object 2
        c1 = ord(b[x+8])                # result/source value
        w1 = dz_util.s2i(b[x+9:x+11])

        s = "Access var %d " % vx
        if fx & 2:
            if vx < len(gRelLabels):
                s += "(%s) " % gRelLabels[vx]   # nID/global rel
            else:
                s += "(unknown) "
        s += "of " + get_data(p, c3, w3)
        s += " to " + get_data(p, c2, w2) + ". "

        if fx & 4:
            s += "Get value from "      # (get = store value to rel)
        else:
            s += "Put value into "
        s += get_data(p, c1, w1)

        if fx & 1: s += ", fail if too small"
        if fx & 2:
            s += ", use neighbor IDs"
        elif fx & 8:
            s += ", don't check presence of second object" # "object to sim" relationship
    return s

def prim_1B(p, b, x, f, node_version):
    """Go To Relative Position"""
    px = (ord(b[x+2]) + 2) & 0xFF       # location
    qx = (ord(b[x+3]) + 2) & 0xFF       # direction
    fx = ord(b[x+6])                    # flags

    s = "Location = %s, Direction = %s" % (gString82[px], gString83[qx])

    if fx & 2: s += ", no failure trees"
    if fx & 4: s += ", allow different altitudes"
    return s

def prim_1C(p, b, x, f, node_version):
    """Run Tree by Name"""
    if (x + 15) >= len(b):
        return "ancient unknown format"

    fx = ord(b[x+2])                    # flags
    ix = ord(b[x+4]) - 1                # string id
    mx = ord(b[x+5])                    # mode
    c1 = ord(b[x+6])                    # param 0
    w1 = dz_util.s2i(b[x+7:x+9])
    c2 = ord(b[x+9])                    # param 1
    w2 = dz_util.s2i(b[x+10:x+12])
    c3 = ord(b[x+12])                   # param 2
    w3 = dz_util.s2i(b[x+13:x+15])

    if fx & 1:                          # global named tree
        s = dz_util.get_html(p,TYPE_FUNCTION,dz_str.read_str2(p, 0, 0x12F, ix))
    elif fx & 2:
        s = dz_util.get_html(p,TYPE_FUNCTION,dz_str.read_str2(p, 1, 0x12F, ix))
    else:
        s = dz_util.get_html(p,TYPE_FUNCTION,dz_str.read_str2(p, 2, 0x12F, ix))

    # How to look for tree with given name:
    # private -> semiglobal -> global -> "false" if not found
    if fx & 8: s += ", ignore semiglobal trees"
    if fx & 4: s += ", ignore global trees"
    if fx & 0x20:
        s += ", Passing in params from the current tree"
    elif fx & 0x10:
        s += ", Passing in params where param 0 = " + get_data(p, c1, w1)
        s += ", param 1 = " + get_data(p, c2, w2)
        s += ", param 2 = " + get_data(p, c3, w3)

    # confusing and badly worded (orly? yarly)
    if mx == 0:
        s += ", run in my stack"
    elif mx == 1:
        s += ", run in Stack Object's stack"
    elif mx == 2:
        s += ", push onto my stack"
    return s

def prim_1D(p, b, x, f, node_version):
    """Set Motive Change"""
    c1 = ord(b[x])                      # rate
    c2 = ord(b[x+1])                    # limit
    mx = ord(b[x+2])                    # motive
    fx = ord(b[x+3])                    # flags
    w1 = dz_util.s2i(b[x+4:x+6])
    w2 = dz_util.s2i(b[x+6:x+8])

    if fx & 1:
        s = "clear all"
    else:
        s = get_data(p, 0x0E, mx)       # motive
        s += " += " + get_data(p, c1, w1)
        s += " per hour, stop at " + get_data(p, c2, w2)

    if fx & 2:
        s += ", Auto Clearing the Person Data Motive Decay value"
    return s

def prim_1E(p, b, x, f, node_version):
    """Gosub Found Action"""
    c1 = ord(b[x])                      # value
    return dz_util.get_html(p,TYPE_FUNCTION,dz_util.chk_range(gString1FE, c1))

def prim_1F(p, b, x, f, node_version):
    """Set to Next"""
    gx = dz_util.s2i(b[x:x+4])          # guid
    mx = ord(b[x+4])                    # mode/var flag
    cx = mx & 0x7F
    c0 = ord(b[x+5])                    # result
    lx = ord(b[x+6])                    # local
    v0 = ord(b[x+7])
    fx = ord(b[x+8])                    # flags
    w1 = dz_util.s2i(b[x+9:x+11])       # data index
    w2 = dz_util.s2i(b[x+11:x+13])      # data value

    s = ""
    if c0 != 0x0A and (mx & 0x80) != 0:
        s += get_data(p, c0, v0) + " "
        s += dz_util.get_html(p,TYPE_OPERATOR,":=")
        s += " next "

    s += dz_util.chk_range(gStringA4, cx)              # Next object
    if cx == 4 or cx == 7:              # type == guid
        s += " "
        p.m_offset += len(s)
        s += get_guid(p, gx)
    if cx == 9 or cx == 0x22:           # object in local
        s += " %d" % lx

    if node_version:
        if fx & 2:
            s += " where %s = %s" % (dz_util.chk_range(gString8D, w1), dz_util.ext16(w2))   # ":="?
        if fx & 1:
            s += " [including disabled objects]"
    return s

def prim_20(p, b, x, f, node_version):
    """Test Object Type"""
    gx = dz_util.s2i(b[x:x+4])          # guid
    w1 = dz_util.s2i(b[x+4:x+6])        # type var
    c1 = ord(b[x+6])
    fx = ord(b[x+7])                    # flags

    if fx & 4:
        s = "Return GUID selected in Temp0/1 "
    else:
        s = "type of " + get_data(p, c1, w1)

    if gx == 0x4C7CAB2B:
        s += " the GUID of the temporary inventory token"
    else:
        s += " "
        p.m_offset += len(s)
        s += get_guid(p, gx)

    if (fx & 5) == 1: s += " Checking against original, not current GUID"
    if (fx & 6) == 2: s += ", incoming ID is a neighbor ID"
    return s

def prim_21(p, b, x, f, node_version):
    """Find 5 Worst Motives"""
    w1 = dz_util.s2i(b[x+4:x+6])        # short owner
    w2 = dz_util.s2i(b[x+6:x+8])        # motive type

    s = gString99[w1] + " lowest "
    s += gStringA5[w2] + "into temps 0-4"
    return s

def prim_22(p, b, x, f, node_version):
    """UI Effect"""
    sig_name = [ "Press Control", "Disable Control", "Enable Control", \
        "Play Effect on Control", "Set Event tree on Control", \
        "Reset State on All Controls", "Disable All Controls", \
        "Reset event trees on All Controls", "Set Event Tree on TNS node", \
        "Set Control Visible", "Set Control Hidden" ]

    sx = ord(b[x])                      # signal
    w1 = dz_util.s2i(b[x+1:x+3])        # window
    w2 = dz_util.s2i(b[x+3:x+5])        # control
    fx = ord(b[x+5])                    # flags
    w3 = dz_util.s2i(b[x+6:x+8])        # start/stop
    w4 = dz_util.s2i(b[x+8:x+10])       # event tree
    tx = ord(b[x+10])                   # tree type
    w5 = dz_util.s2i(b[x+11:x+13])      # TNS ID
    c5 = ord(b[x+13])

    if sx >= len(sig_name):
        s = "unknown"
    else:
        s = sig_name[sx]

    if sx < 5 or sx > 8:
        s += " "
        if fx & 4:
            s += dz_str.read_str2(p, 0, 0x96, w2)
        elif fx & 8:
            s += dz_str.read_str2(p, 1, 0x96, w2)
        else:
            s += dz_str.read_str2(p, 2, 0x96, w2)

    if sx != 8:
        s += " in window "
        if fx & 1:
            s += dz_str.read_str2(p, 0, 0x96, w1)
        elif fx & 2:
            s += dz_str.read_str2(p, 1, 0x96, w1)
        else:
            s += dz_str.read_str2(p, 2, 0x96, w1)

    if sx == 3:
        if w3 != 0:
            s += " Starting Effect"
        else:
            s += " Stopping Effect"

    if sx == 4 or sx == 8:
        s += " "
        p.m_offset += len(s)
        s += get_etree(p, w4, tx)
    if sx == 8:
        s += ", Getting TNS ID from " + get_data(p, c5, w5)
    return s

def prim_23(p, b, x, f, node_version):
    """Camera Control"""
    w1 = dz_util.s2i(b[x:x+2])          # timeout
    dx = ord(b[x+3])                    # distance
    fx = ord(b[x+4])                    # flags

    if fx & 1:
        s = "show stack obj"

        if dx == 1: s += ", far"
        elif dx == 2: s += ", mid"
        elif dx == 3: s += ", near"
        s += " zoom"

        if (fx & 0x40) == 0: s += ", slow down"
        if fx & 8: s += ", center"
        if fx & 0x20:
            s += ", using timeout in temp 0"
        else:
            s += ", using timeout of " + dz_util.ext16(w1)
    else:
        s = "un-show stack obj"
    return s

def prim_24(p, b, x, f, node_version):
    """Dialog"""
    tns_type = [ "Sim Type", "System Type", "System Dialog Type", \
        "Birthday Type", "Sim (about Object ID in Temp 1) Type" ]

    if (x + 15) >= len(b):
        return "ancient unknown format"

    ic = ord(b[x+1])                    # icon
    yb = ord(b[x+3]) - 1                # yes button
    nb = ord(b[x+4]) - 1                # no button
    dp = ord(b[x+5])                    # dialog primitive
    tx = ord(b[x+6]) - 1                # title
    tv = (ord(b[x+7]) >> 4) & 7         # temp
    it = (ord(b[x+7]) >> 1) & 7         # icon type
    fx = ord(b[x+8])                    # flags
    px = ord(b[x+9]) + 1                # priority
    to = ord(b[x+10])                   # timeout
    lx = ord(b[x+11])                   # local data pass through
    tt = ord(b[x+12])                   # tns type

    if node_version:
        mx = dz_util.s2i(b[x+13:x+15]) - 1      # message
        cx = (ord(b[x+2]) << 8) + ord(b[x]) - 1 # cancel
    else:
        mx = ord(b[x+2]) - 1                # message
        cx = ord(b[x]) - 1                  # cancel

    # in, you must go
    if dp == 0xFF: return "unknown"     # test object dialog (?)

    if dp == 2:     # yes-no-cancel
        s = get_data(p, 8, tv) + " := tri choice: "
    else:
        s = dz_util.chk_range(gStringD9, dp) + ": "        # Dialog primitives
    if dp == 9:     # Stop Text Notification
        s += ", Getting Text Notification ID from temp %d, TNS " % tv
        if tt == 0: s += tns_type[0]
        elif tt == 1: s += tns_type[1]
    if dp == 0x0E:  # Food Rack
        s += "passing through data starting with local var %d" % lx

    # non-Rack/STN type dialogs
    if dp < 9 or dp == 0x0A or dp == 0x0F or dp == 0x13 or dp > 0x14:
        if fx & 2:  # ?
            s += ", expecting message Index in temp 0"
        elif fx & 0x40:
            s += "[Global STR# 0x12D:0x%X]" % mx
        elif fx & 1:
            s += "[SemiGlobal STR# 0x12D:0x%X]" % mx
        else:
            s += "[STR# 0x12D:0x%X]" % mx
        if dp == 8 or dp == 0x0A:
            if tt < len(tns_type):
                s += ", TNS " + tns_type[tt]
            s += ", priority %d" % px
            s += ", timeout %d" % to
            if dp == 0x0A:
                s += ", getting"
            else:
                s += ", putting"
            s += " Text ID in temp %d" % tv
        else:
            if fx & 4:      # ?
                s += ", expecting Yes button Index in temp 0"
            elif (fx & 4) == 0 and yb >= 0:
                s += ", Yes: [STR# 0x12D:0x%X]" % yb
            if fx & 8:      # ?
                s += ", expecting No button Index in temp 0"
            elif (fx & 8) == 0 and nb >= 0:
                s += ", No: [STR# 0x12D:0x%X]" % nb
            if fx & 0x10:   # ?
                s += ", expecting Title Index in temp 0"
            elif (fx & 0x10) == 0 and tx >= 0:
                s += ", Title: [STR# 0x12D:0x%X]" % tx
            if fx & 0x20:   # ?
                s += ", expecting Cancel Index in temp 0"
            elif (fx & 0x20) == 0 and cx >= 0:
                s += ", Cancel: [STR# 0x12D:0x%X]" % cx

        if dp != 0x16 and dp != 0x19:   # not TutorialNext nor TutorialNextModal
            s += ", icon: "
            if it == 3:     # indexed
                s += "BMP = 0x%X" % (ic + 5000)
            elif it == 4:   # named
                s += "[STR# 0x12D:0x%X]" % (ic - 1)
            else:
                s += dz_util.chk_range(gStringF4, it)
    return s

def prim_25(p, b, x, f, node_version):
    """Test Sim Interacting With"""
    return "me with stack obj."

def prim_26(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_27(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_28(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_29(p, b, x, f, node_version):
    """~(old)Set Balloon/Headline"""
    return ""

def prim_2A(p, b, x, f, node_version):
    """Create New Object Instance"""
    gx = dz_util.s2i(b[x:x+4])          # guid
    px = ord(b[x+4])                    # place
    fx = ord(b[x+5])                    # flags
    lx = ord(b[x+6])                    # local reference
    sx = ord(b[x+9])                    # slot
    hx = ord(b[x+10])                   # flags 2

    if fx & 4:
        s = "neighbor in stack obj"
    elif fx & 0x40:
        s = "from GUID in temp Inventory Token"
    elif fx & 0x80:
        s = "from GUID in temp0/1"
    else:
        s = get_guid(p, gx)

    s += ", place "
    if px == 4:
        s += "in stack objs slot %d" % sx
    elif px == 0x0A:
        s += "in obj in temp 0's slot %d" % sx
    else:
        s += dz_util.chk_range(gStringA7, px)  # Object place

    if px == 8 or px == 9:      # in/of local
        s += " %d" % lx

    if fx & 2:
        s += ", pass object ids to main"
    elif fx & 0x10:
        s += ", pass temp 0 to main"

    if fx & 1: s += ", do not duplicate"
    if fx & 8: s += ", fail if tile is non-empty"
    if hx & 1: s += ", Moving in a new Sim"
    if hx & 2: s += ", copying design mode materials from object in temp5"
    return s

def prim_2B(p, b, x, f, node_version):
    """~(old)Drop Onto"""
    return ""

def prim_2C(p, b, x, f, node_version):
    """~(old)Animate Sim [old]"""
    return ""

def prim_2D(p, b, x, f, node_version):
    """Go To Routing Slot"""
    w1 = dz_util.s2i(b[x:x+2])          # param/slot/local
    w2 = dz_util.s2i(b[x+2:x+4])        # mode
    fx = ord(b[x+4])                    # flags

    if w2 == 0:
        s = "in " + get_data(p, 9, w1)  # param
    elif w2 == 1:
        if fx & 2:
            s = "From Slot index in Temp 0"
        else:
            s = "0x%X" % w1
    elif w2 == 2:
        s = "global 0x%X" % w1
    elif w2 == 3:
        s = "in " + get_data(p, 0x19, w1)   # local
    else:
        s = "unknown (0x%X) 0x%X" % (w2, w1)

    if fx & 1: s += ", no failure trees"
    if fx & 4: s += ", ignoring dest obj footprint"
    if fx & 8: s += ", allow different altitudes"
    return s

def prim_2E(p, b, x, f, node_version):
    """Snap"""
    w1 = dz_util.s2i(b[x:x+2])          # slot
    w2 = dz_util.s2i(b[x+2:x+4])        # mode
    fx = ord(b[x+4])                    # flags
    w3 = dz_util.s2i(b[x+8:x+10])       # flags 2

    s = dz_util.chk_range(gStringCF, w2)               # Snap mode

    if w2 == 0 or w2 == 3 or w2 == 4:
        if fx & 8:
            s += " to slot in temp 0"
        else:
            s += " 0x%X" % w1

    if w3 == 1: s += ", from slot in temp 1"
    if fx & 2: s += ", ask person to move"
    if fx & 0x10: s += ", TEST ONLY"
    return s

def prim_2F(p, b, x, f, node_version):
    """~(old)Reach"""
    return ""

def prim_30(p, b, x, f, node_version):
    """Stop ALL Sounds"""
    w1 = dz_util.s2i(b[x:x+2])          # owner
    if w1 == 0:
        s = "of Me"
    else:
        s = "of Stack Object"
    return s

def prim_31(p, b, x, f, node_version):
    """Notify the Stack Object out of Idle"""
    return ""   # no operands

def prim_32(p, b, x, f, node_version):
    """Add/Change the Action String"""
    fx = ord(b[x+2])                    # flags
    hx = ord(b[x+3])                    # flags 2
    gx = dz_util.s2i(b[x+5:x+9])        # guid
    mx = ord(b[x+9])                    # mode
    ix = ord(b[x+10])                   # icon index
    c1 = ord(b[x+11])                   # object
    w1 = dz_util.s2i(b[x+12:x+14])

    if mx == 0:
        if node_version < 2:
            sx = ord(b[x+4]) - 1        # string index
        else:
            sx = ord(b[x+14]) - 1
        s = "Add / Change Interaction string Mode, "
        if (hx & 1) != 0 and node_version:
            s += "Disabled (Propagating), "
        if hx & 2:
            s += "Disabled (NonPropagating), "
        if fx & 0x10:
            s += "Getting index from temp 0, getting string from "
            if fx & 4:
                s += "globals"
            elif fx & 8:
                s += "semiglobals"
            else:
                s += "privates"
        elif fx & 4:
            s += dz_str.read_str2(p, 0, 0x12E, sx)
        elif fx & 8:
            s += dz_str.read_str2(p, 1, 0x12E, sx)
        else:
            s += dz_str.read_str2(p, 2, 0x12E, sx)

    else:
        s = "Interaction Icon Change Mode"
        if fx & 0x20:
            s += ", Using external guid for Thumbnail Outfit from "
            if fx & 0x40:
                s += "GUID in temp2/3"
            else:
                p.m_offset += len(s)
                s += get_guid(p, gx)
        else:
            s += ", Using object ID in " + get_data(p, c1, w1)
        if fx & 0x80:
            s += ", Getting icon index into model table from temp 1"
        else:
            s += ", Using index of %d" % ix
    return s

def prim_33(p, b, x, f, node_version):
    """Manage Inventory (Tokens)"""
    token_type = [ "Global", "Lot", "Family", "Neighbor", "Game-Wide" ]

    # note: deep hurting ahead

    fx = ord(b[x])                      # flags/token type
    c3 = ord(b[x+1])                    # var 3
    w3 = dz_util.s2i(b[x+2:x+4])
    ax = ord(b[x+4])                    # action
    gx = dz_util.s2i(b[x+5:x+9])        # GUID/var 4
    c4 = ord(b[x+6])                    # var 4/copy token type
    w4 = dz_util.s2i(b[x+7:x+9])
    cx = ord(b[x+9])                    # category
    c2 = ord(b[x+10])                   # var 2
    w2 = dz_util.s2i(b[x+11:x+13])
    c1 = ord(b[x+13])                   # var 1
    w1 = dz_util.s2i(b[x+14:x+16])

    if node_version == 0:
        fx = ((fx & 0x3C) << 1) | (fx & 0x83)   # bump

    xx = fx & 7         # token type
    x4 = c4 & 7         # source type for copy token

    # Robot Chicken
    s = "Access the "
    if xx < len(token_type):
        s += token_type[xx] + " "
    else:
        s += "unknown "
    if fx & 8:    
        s += "Counted"
    else:
        s += "Singular"
    s += " Inventory"
    if xx != 0:
        s += " from ID " + get_data(p, c3, w3)  # var 3
    s += ". with category %d" % cx

    # if not "Copy token at index" nor some Token Property Assignment
    if ( (fx & 8) and (ax != 0xB) ) or \
        ( ((fx & 8) == 0) and (ax != 0xE) and (ax != 0xF) and (ax != 0x12) ):
        if gx == 0:
            s += " of GUID from Stack Object. "
        else:
            s += " "
            p.m_offset += len(s)
            s += get_guid(p, gx) + ". "
    else:
        s += ". "

    if fx & 8:  # Counted type actions
        if ax == 0:
            s += "Add token. Pull count from " + get_data(p, c1, w1)
        elif ax == 1:
            s += "Add to token at index from " + get_data(p, c2, w2)
            s += ". Pull count from " + get_data(p, c1, w1)
        elif ax == 2:
            s += "Remove token. Pull count from " + get_data(p, c1, w1)
        elif ax == 3:
            s += "Remove to token at index from" + get_data(p, c2, w2)
            s += ". Pull count from " + get_data(p, c1, w1)
        elif ax == 4:
            s += "Remove all tokens."
        elif ax == 5:
            s += "Remove all tokens from token at index from" + get_data(p, c2, w2) + "."
        elif ax == 6:
            s += "Find the token. Put count into " + get_data(p, c1, w1) + "."
        elif ax == 7:
            s += "Read token into My Temp Token."
        elif ax == 8:
            s += "Read token at index from " + get_data(p, c2, w2)
            s += " into My Temp Token."
        elif ax == 9:
            s += "Set To Next token. Starting at index from " + get_data(p, c2, w2)
        elif ax == 0x0A:
            s += "Store the count of the tokens in this inventory into "
            s += get_data(p, c1, w1) + "."
        elif ax == 0x0B:
            s += "Copy token at index " + get_data(p, c2, w2)
            s += " from the "
            if x4 < len(token_type):
                s += token_type[x4] + " "
            else:
                s += "unknown "
            s += "inventory"
            if x4 > 0 and x4 <= 3:  # not global/game-wide
                s += " with id " + get_data(p, c1, w1)
            s += "."

    else:   # Singular type actions
        if node_version < 2:
            cx = 0x0C

        if ax == 0:
            s += "Add token."
        if ax == 1:
            s += "Remove token at index from " + get_data(p, c2, w2) + "."
        if ax == 2:
            s += "Remove "
            if cx & 4:
                if fx & 0x10: s += "visible "
                else: s += "hidden "
            if cx & 8:
                if fx & 0x10: s += "memory "
                else: s += "non-memory "
            if cx & 0x20:
                if cx & 1: s += "shopping "
                else: s += "non-shopping "
            s += "tokens."
        if ax == 3:
            s += "Set To Next "
            if cx & 4:
                if fx & 0x10: s += "visible "
                else: s += "hidden "
            if cx & 8:
                if fx & 0x10: s += "memory "
                else: s += "non-memory "
            if cx & 0x20:
                if cx & 1: s += "shopping "
                else: s += "non-shopping "
            s += "token. Starting at index from " + get_data(p, c2, w2)
            if fx & 0x80:
                s += ", Reversed."
            else:
                s += "."
        if ax == 4:
            s += "Push property on token at index from " + get_data(p, c2, w2)
            s += ". Get property value from " + get_data(p, c1, w1)
        if ax == 5:
            s += "Pop property off token at index from " + get_data(p, c2, w2)
            s += ". Put property value into " + get_data(p, c1, w1) + "."
        if ax == 6:
            s += "Read token into My Temp Token at index from " + get_data(p, c2, w2) + "."
        if ax == 7:
            s += "Get property from token in My Temp Token at index from "
            s += get_data(p, c2, w2)
            s += ". Put property value into " + get_data(p, c1, w1) + "."
        if ax == 9:
            s += "Save My Temp Token back to the location it was loaded from."
        if ax == 0x0A:
            s += "Store the count of the tokens in this inventory into "
            s += get_data(p, c1, w1) + "."
        if ax == 0x0C:
            s += "Set To Next "
            if cx & 4:
                if fx & 0x10: s += "visible "
                else: s += "hidden "
            if cx & 8:
                if fx & 0x10: s += "memory "
                else: s += "non-memory "
            if cx & 0x20:
                if cx & 1: s += "shopping "
                else: s += "non-shopping "
            s += "token. Starting at index from " + get_data(p, c2, w2)
            if fx & 0x80:
                s += ", Reversed."
            else:
                s += "."
        if ax == 0x0D:
            s += "Store the count of the "
            if cx & 4:
                if fx & 0x10: s += "visible "
                else: s += "hidden "
            if cx & 8:
                if fx & 0x10: s += "memory "
                else: s += "non-memory "
            if cx & 0x20:
                if cx & 1: s += "shopping "
                else: s += "non-shopping "
            s += "tokens in this inventory into " + get_data(p, c1, w1) + "."
        if ax == 0x0E:
            s += "Token Index " + get_data(p, c4, w4)
            s += ", Property " + get_data(p, c2, w2)
            s += " Assign to: " + get_data(p, c1, w1)
        if ax == 0x0F:
            s += get_data(p, c1, w1) + " Assign to: Token Index "
            s += get_data(p, c4, w4) + ", Property " + get_data(p, c2, w2)
        if ax == 0x10:
            s += "Add Token And Instance Info of Stack Object."
        if ax == 0x11:
            s += "Create Object from Token at Index."
        if ax == 0x12:
            s += "Copy token at index " + get_data(p, c2, w2)
            s += " from the "
            if x4 < len(token_type):
                s += token_type[x4] + " "
            else:
                s += "unknown "
            s += "inventory"
            if x4 > 0 and x4 <= 3:  # not global/game-wide
                s += " with id " + get_data(p, c1, w1)
            s += "."
        if ax == 0x13:
            s += "Add Token And Instance Info of Stack Object excluding contained objects."
    return s

def prim_34(p, b, x, f, node_version):
    """~unused (TSO)"""
    return ""

def prim_35(p, b, x, f, node_version):
    """~unused (TSO)"""
    return ""

def prim_36(p, b, x, f, node_version):
    """~unused (TSO)"""
    return ""

def prim_37(p, b, x, f, node_version):
    """~unused (TSO)"""
    return ""

def prim_38(p, b, x, f, node_version):
    """~unused (TSO)"""
    return ""

def prim_39(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_3A(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_3B(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_3C(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_3D(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_3E(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_3F(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_40(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_41(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_42(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_43(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_44(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_45(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_46(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_47(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_48(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_49(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_4A(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_4B(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_4C(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_4D(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_4E(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_4F(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_50(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_51(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_52(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_53(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_54(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_55(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_56(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_57(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_58(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_59(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_5A(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_5B(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_5C(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_5D(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_5E(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_5F(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_60(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_61(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_62(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_63(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_64(p, b, x, f, node_version):
    """~reserved"""
    return ""

def prim_65(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_66(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_67(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_68(p, b, x, f, node_version):
    """~unused"""
    return ""

def prim_69(p, b, x, f, node_version):
    """Animate Object"""
    w2 = dz_util.s2i(b[x:x+2])          # anim index
    fx = ord(b[x+2])                    # flags
    w3 = dz_util.s2i(b[x+4:x+6])        # event tree
    c1 = ord(b[x+6])                    # target object
    w1 = dz_util.s2i(b[x+7:x+9])
    tt = ord(b[x+9])                    # tree type
    gx = ord(b[x+10])                   # flags 2

    if fx & 4:
        s = "anim id in STR# 0x86:[Param %d]" % w2
    else:
        s = "anim = " + dz_str.read_str2(p, 2, 0x86, w2) + ","    # private

    s += " affecting " + get_data(p, c1, w1)
    s += " "
    p.m_offset += len(s)
    s += get_etree(p, w3, tt)

    if fx & 8: s += ", Interruptible"
    if fx & 1: s += ", Flipped"
    if fx & 0x10: s += ", Start at tag in Temp 0"
    if fx & 0x20: s += ", Loop Count in Temp 1"
    if fx & 0x80: s += ", No Blend in"
    if fx & 0x40: s += ", No Blend out"
    if fx & 2: s += ", Anim Speed in Temp2"
    if gx & 1: s += ", Flip Flag in Temp 3"
    if gx & 4: s += ", Sync to calling objects Anim"
    if gx & 8: s += ", Align blend out with to calling objects Anim"
    if gx & 0x80: s += ", Not Hurryable"
    return s

def prim_6A(p, b, x, f, node_version):
    """Animate Sim"""
    w1 = dz_util.s2i(b[x:x+2])          # anim index
    fx = ord(b[x+2])                    # flags
    w2 = dz_util.s2i(b[x+4:x+6])        # event tree
    lt = ord(b[x+6])                    # list type
    tt = ord(b[x+7])                    # tree type
    gx = ord(b[x+8])                    # flags 2
    c3 = ord(b[x+9])                    # IK object
    w3 = dz_util.s2i(b[x+10:x+12])
    px = ord(b[x+12])                   # priority

    if fx & 4:
        s = "anim id in Param %d" % w1
    else:
        s = "anim id 0x%X" % w1
    s += " from "
    if lt in anim_list_type:
        s += anim_list_type[lt]
    else:
        s += "Object"
    s += " list"

    if (fx & 4) == 0:
        s += " ("
        if lt == 0x80:
            s += dz_str.read_str2(p, 0, 0x81, w1)    # global (adult)
        elif lt in anim_list_type:
            s += dz_str.read_str2(p, 2, lt, w1)      # private
        else:
            s += dz_str.read_str2(p, 2, 0x86, w1)
        s += ")"

    s += ", "
    p.m_offset += len(s)
    s += get_etree(p, w2, tt)

    if fx & 8: s += ", Interruptible"
    if fx & 1: s += ", Flipped"
    if fx & 0x20: s += ", Trans to Idle"
    if fx & 0x80: s += ", No Blend in"
    if fx & 0x40: s += ", No Blend out"
    if fx & 0x10: s += ", Start at tag in Temp 0"
    if fx & 2: s += ", Anim Speed in Temp2"
    if gx & 1: s += ", Flip Flag in Temp 3"
    if gx & 2: s += ", Synching Anim to last Anim Run"
    if gx & 0x10: s += ", Using Controlling Object as Anim Source"
    if gx & 0x20: s += ", Not Hurryable"

    s += ", IK Object " + get_data(p, c3, w3)

    s += ", "
    if px == 0:
        s += "low"
    elif px == 1:
        s += "medium"
    elif px == 2:
        s += "high"
    s += " priority"
    return s

def prim_6B(p, b, x, f, node_version):
    """Animate Overlay"""
    w1 = dz_util.s2i(b[x:x+2])          # anim index
    fx = ord(b[x+2])                    # flags
    w2 = dz_util.s2i(b[x+4:x+6])        # event tree
    c3 = ord(b[x+6])                    # target object
    w3 = dz_util.s2i(b[x+7:x+9])
    lt = ord(b[x+9])                    # list type
    px = ord(b[x+12])                   # priority
    tt = ord(b[x+14])                   # tree type
    gx = ord(b[x+15])                   # flags 2

    nh = 0                              # not hurryable
    if node_version:
        px = ord(b[x+11])
        nh = ord(b[x+12]) & 1

    if fx & 4:
        s = "anim id in Param %d" % w1
    else:
        s = "anim id 0x%X" % w1
    s += " from "
    if lt in anim_list_type:
        s += anim_list_type[lt]
    else:
        s += "Object"
    s += " list"

    if (fx & 4) == 0:
        s += " ("
        if lt == 0x80:
            s += dz_str.read_str2(p, 0, 0x81, w1)    # global (adult)
        elif lt in anim_list_type:
            s += dz_str.read_str2(p, 2, lt, w1)      # private
        else:
            s += dz_str.read_str2(p, 2, 0x86, w1)
        s += ")"

    s += " affecting " + get_data(p, c3, w3)

    s += ", "
    p.m_offset += len(s)
    s += get_etree(p, w2, tt)

    if fx & 8: s += ", Interruptible"
    if fx & 1: s += ", Flipped"
    if fx & 0x80: s += ", No Blend in"
    if fx & 0x40: s += ", No Blend out"
    if fx & 0x10: s += ", Start at tag in Temp 0"
    if fx & 0x20: s += ", Loop Count in Temp 1"
    if fx & 2: s += ", Anim Speed in Temp2"
    if gx & 1: s += ", Flip Flag in Temp 3"
    if gx & 0x10: s += ", Sync to calling objects Anim"
    if gx & 0x20: s += ", Align blend out with to calling objects Anim"
    if gx & 0x40: s += ", Using tree owners anim list"
    if nh: s += ", Not Hurryable"

    s += ", "
    if px == 0:
        s += "low"
    elif px == 1:
        s += "medium"
    elif px == 2:
        s += "high"
    s += " priority"
    return s

def prim_6C(p, b, x, f, node_version):
    """Animate Stop"""
    anim_type = [ "All Overlay animations", "All Full Body Animations", "All Animations", \
        "Carry Poses", "Idle Animations", "Gesture Animations", "Reaction Animations", \
        "Normal Animations", "Facial Animations", "Facial Idle Animations", \
        "Receptivity Animations" ]

    w1 = dz_util.s2i(b[x:x+2])          # anim index
    fx = ord(b[x+2])                    # flags
    c2 = ord(b[x+3])                    # target object
    w2 = dz_util.s2i(b[x+4:x+6])
    lt = ord(b[x+6])                    # list type
    at = ord(b[x+7])                    # anim type
    px = ord(b[x+8])                    # priority

    if at == 0:
        if fx & 4:
            s = "anim id in Param %d" % w1
        else:
            s = "anim id 0x%X" % w1
        s += " from "
        if lt in anim_list_type:
            s += anim_list_type[lt]
        else:
            s += "Object"
        s += " list"

        if (fx & 4) == 0:
            s += " ("
            if lt == 0x80:
                s += dz_str.read_str2(p, 0, 0x81, w1)    # global (adult)
            elif lt in anim_list_type:
                s += dz_str.read_str2(p, 2, lt, w1)      # private
            else:
                s += dz_str.read_str2(p, 2, 0x86, w1)
            s += ")"

    else:
        if at > len(anim_type):
            s = "Unknown Type"
        else:
            s = anim_type[at - 1]

    s += " affecting " + get_data(p, c2, w2)

    if fx & 1: s += ", Flipped"
    if fx & 2: s += ", blended Out"
    if fx & 0x20: s += ", short blended Out"
    if fx & 8: s += ", Flip Flag in Temp 3"
    if fx & 0x40: s += ", Stopping normal AND flipped anims"

    s += ", Priority "
    if px == 0:
        s += "Any"
    elif px == 1:
        s += "Low"
    elif px == 2:
        s += "Medium"
    else:
        s += "High"
    return s

def prim_6D(p, b, x, f, node_version):
    """Change Material"""
    tx = dz_util.s2i(b[x:x+2])          # texture index
    fx = ord(b[x+2])                    # flags
    mx = dz_util.s2i(b[x+3:x+5])        # mesh index
    c1 = ord(b[x+5])                    # target object
    w1 = dz_util.s2i(b[x+6:x+8])
    c2 = ord(b[x+8])                    # source object
    w2 = dz_util.s2i(b[x+9:x+11])
    gx = ord(b[x+13])                   # flags 2

    if gx & 2:
        s = "using snap shot generated material"
    elif fx & 8:
        s = "found in obj in " + get_data(p, c2, w2)
        if fx & 0x10:
            if gx & 1:
                s += ", using Moving Texture Name index from temp 0"
            else:
                s += ", using material index from temp 0"
        else:
            if gx & 1:
                s += ", using Moving Texture Name index %d" % tx
            else:
                s += ", using material index %d" % tx
    else:
        s = "found in me, to: "
        if fx & 0x10:
            if gx & 1:
                s += ", using Moving Texture Name index from temp 0"
            else:
                s += ", using material index from temp 0"
        else:
            if fx & 2:
                s += dz_str.read_str2(p, 0, 0x88, tx)   # global
            elif fx & 4:
                s += dz_str.read_str2(p, 1, 0x88, tx)   # semiglobal
            else:
                s += dz_str.read_str2(p, 2, 0x88, tx)   # private

    if fx & 1:
        s += ", mesh found in obj in" + get_data(p, c2, w2)
        if fx & 0x20:
            s += ", using mesh group index from temp 1"
        else:
            s += ", using mesh group index %d" % mx
    elif fx & 0x20:
        s += ", using mesh group index from temp 1"
    elif mx & 0x8000:       # mx < 0
        s += ", over all model"
    else:
        s += ", On Mesh Group: "
        if fx & 0x40:
            s += dz_str.read_str2(p, 0, 0x87, mx)   # global
        elif fx & 0x80:
            s += dz_str.read_str2(p, 1, 0x87, mx)   # semiglobal
        else:
            s += dz_str.read_str2(p, 2, 0x87, mx)   # private

    s += ", affecting ID in " + get_data(p, c1, w1)
    return s

def prim_6E(p, b, x, f, node_version):
    """Look At"""
    fx = ord(b[x])                      # flags
    c1 = ord(b[x+1])                    # target
    w1 = dz_util.s2i(b[x+2:x+4])
    sx = ord(b[x+4])                    # speed
    xx = ord(b[x+5])                    # speed 2 (?)
    sn = ord(b[x+8])                    # slot number
    w2 = dz_util.s2i(b[x+9:x+11])       # event tree
    tt = ord(b[x+11])                   # tree type
    st = ord(b[x+14])                   # slot type
    gx = ord(b[x+15])                   # flags 2

    if (fx & 1) == 0:
        s = get_etree(p, w2, tt)

        if (fx & 0x80) == 0:
            s += ", Targeting ID in " + get_data(p, c1, w1)
        else:
            s += ", Targeting Camera"

        s += " using "
        if st == 0:
            s += "default 3/4 height"
        elif st == 1:
            s += "targeting slot number %d" % sn
        elif st == 2:
            s += "routing slot number %d" % sn
        else:
            s += "containment slot number %d" % sn

        if fx & 4: s += ", including Spine"
        if fx & 0x10: s += ", using duration in temp 0"
        if fx & 2: s += ", no early exit"

    else:
        s = "STOP"

    if fx & 8:
        s += ", using turnTowardsSpeed in temp 1"
    elif node_version:
        s += ", turn towards speed of %d degrees per second" % (sx * 2)
    else:
        s += ", turn towards speed of %d degrees per second" % sx

    if (gx & 2) and node_version or (c1 & 8) and (node_version == 0):
        s += ", using turnAwaySpeed in temp 1"
    elif gx & 1:
        s += ", using turnAwaySpeed in temp 2"
    elif node_version:
        s += ", turn away speed of %d degrees per second" % (xx * 2)
    else:
        s += ", turn away speed of %d degrees per second" % sx

    if fx & 0x20: s += ", ignoring Room"
    if fx & 0x40: s += ", ignoring Frustrum"
    if gx & 4: s += ", Not Hurryable"
    return s

def prim_6F(p, b, x, f, node_version):
    """Change Light"""
    fx = ord(b[x+1])                    # flags
    c1 = ord(b[x+2])                    # target object
    w1 = dz_util.s2i(b[x+3:x+5])
    w2 = dz_util.s2i(b[x+5:x+7])        # duration
    ix = ord(b[x+7])                    # intensity
    tx = ord(b[x+8])                    # target light

    s = "on object in " + get_data(p, c1, w1)
    if tx != 0xFF:
        s += ", Targeting light " + dz_str.read_str2(p, 2, 0x8E, tx)   # private
    else:
        s += ", Targeting all lights on Object"

    if fx & 1:
        s += ", Fade duration in Temp 1,"
    else:
        s += ", Fading over a duration of %d ticks," % w2
    if fx & 2:
        s += " Intensity in Temp 0"
    else:
        s += " Intensity of %d percent" % ix
    return s

def prim_70(p, b, x, f, node_version):
    """Effect Stop/Start"""
    fx_action = [ "Soft Start Effect", "Hard start effect", "Soft stop effect", \
        "Hard stop effect", "Soft stop all effects", "Hard stop all effects", \
        "Fire and Forget Effect", "Interrogate Bone for effects", \
        "Hard stop all effects and clear Queue", "Hard stop ALL effects", \
        "Set State 1 for all effects", "Set State 2 for all effects", \
        "Set State 3 for all effects", "Set State 4 for all effects", \
        "Soft stop ALL effects" ]

    ax = ord(b[x])                      # action
    c1 = ord(b[x+1])                    # target object
    w1 = dz_util.s2i(b[x+2:x+4])
    ei = ord(b[x+4])                    # effect index
    sn = ord(b[x+6])                    # slot number
    st = ord(b[x+9])                    # slot type
    fx = ord(b[x+10])                   # flags
    gx = ord(b[x+11])                   # flags 2
    c2 = ord(b[x+12])                   # icon
    w2 = dz_util.s2i(b[x+13:x+15])
    si = ord(b[x+15])                   # sheet index

    if ax < len(fx_action):
        s = fx_action[ax]
    else:
        s = "unknown effect"        
    s += " on object in " + get_data(p, c1, w1)

    if ax == 4 or ax == 5:
        if fx & 0x40:
            s += ", Passing in effect ID in temp 1"
    elif ax < 7 or ax == 0x0E:
        if ei != 0xFF:
            s += ", "
            if fx & 1:
                s += dz_str.read_str2(p, 0, 0x8F, ei)   # !!!
            elif fx & 2:
                s += dz_str.read_str2(p, 1, 0x8F, ei)   # semiglobal
            else:
                s += dz_str.read_str2(p, 2, 0x8F, ei)   # private
        else:
            s += ", Affecting default effect"

    if ax != 9:
        s += ", of Slot Type "
        if st == 0:
            s += "Target"
        elif st == 1:
            s += "Routing"
        else:
            s += "Containment"
        s += ", using slot number %d" % sn

    if gx & 0x10:
        s += ", getting icon value from Temp6"
    elif gx & 4:
        s += ", putting in Icon with GUID in temp4/5"
        if gx & 8:
            s += " getting model name index from Temp6"
        else:
            s += " using default object model"
    elif fx & 0x10:
        s += ", putting in Icon from neighbor ID in " + get_data(p, c2, w2)
    elif fx & 0x20:
        s += ", putting in Conversation Icon index found in " + get_data(p, c2, w2)
        s += " using sheet " + dz_str.read_str2(p, 2, 0x95, si)     # ??
    elif fx & 4:
        s += ", putting in Icon from object in " + get_data(p, c2, w2)
        if gx & 8:
            s += " getting model name index from Temp6"
        else:
            s += " using default object model"

    if fx & 0x80:
        s += ", putting effect in priority Queue"
    return s

def prim_71(p, b, x, f, node_version):
    """Snap Into"""
    c1 = ord(b[x])                      # source
    w1 = dz_util.s2i(b[x+1:x+3])
    c2 = ord(b[x+3])                    # target
    w2 = dz_util.s2i(b[x+4:x+6])
    sn = ord(b[x+6])                    # slot number
    fx = ord(b[x+9])                    # flags

    s = "Snap Object in " + get_data(p, c1, w1)
    s += " into Object in " + get_data(p, c2, w2)

    if fx & 1:
        s += ", using slot # in temp 0"
    else:
        s += ", using slot number %d" % sn

    if fx & 2: s += ", TEST ONLY"
    if fx & 4: s += ", Reset root bones"
    return s

def prim_72(p, b, x, f, node_version):
    """Assign Locomotion Animations"""
    w1 = dz_util.s2i(b[x:x+2])          # group
    fx = ord(b[x+2])                    # flags

    if w1 == 0:
        s = "Pop all entries from Stack."
    elif w1 == 1:
        s = "Pop last entry from Stack."
    else:
        s = "Push group #%d onto Stack" % (w1 - 2)
        s += ", getting group from "
        if fx & 4:
            s += "Global."
        elif fx & 2:
            s += "SemiGlobal."
        else:
            s += "Private."
    return s

def prim_73(p, b, x, f, node_version):
    """Debug"""
    debug_action = [ "Toggle Window Open/Close", "Open Animation Ticker", \
        "Show Slots", "Show Bones", "Toggle Anim Info to Debug Window", \
        "Perform Cheat", "Dump Happy Log" ]

    c1 = ord(b[x])
    w1 = dz_util.s2i(b[x+1:x+3])
    c2 = ord(b[x+3])
    w2 = dz_util.s2i(b[x+4:x+6])
    c3 = ord(b[x+6])
    w3 = dz_util.s2i(b[x+7:x+9])
    c4 = ord(b[x+9])
    w4 = dz_util.s2i(b[x+10:x+12])
    sx = ord(b[x+12])                   # index
    ax = ord(b[x+13])                   # action
    fx = ord(b[x+14])                   # flags

    if ax == 0:
        if fx & 4:
            s = dz_str.read_str2(p, 0, 0x132, sx)   # global
        elif fx & 2:
            s = dz_str.read_str2(p, 1, 0x132, sx)   # semiglobal
        else:
            s = dz_str.read_str2(p, 2, 0x132, sx)   # private
        s += ", param 0: " + get_data(p, c1, w1)
        s += ", param 1: " + get_data(p, c2, w2)
        s += ", param 2: " + get_data(p, c3, w3)
        s += ", param 3: " + get_data(p, c4, w4)
    else:
        if ax > len(debug_action):
            s = "unknown"
        else:
            s = debug_action[ax - 1]
        if ax == 6:
            s += " "
            if fx & 4:
                s += dz_str.read_str2(p, 0, 0x132, sx)   # global
            elif fx & 2:
                s += dz_str.read_str2(p, 1, 0x132, sx)   # semiglobal
            else:
                s += dz_str.read_str2(p, 2, 0x132, sx)   # private
    return s

def prim_74(p, b, x, f, node_version):
    """Reach/Put"""
    c1 = ord(b[x+3])                    # target
    w1 = dz_util.s2i(b[x+4:x+6])
    sn = ord(b[x+6])                    # slot number
    fx = ord(b[x+9])                    # flags
    mx = ord(b[x+10])                   # mode
    sx = dz_util.s2i(b[x+11:x+13])      # sim anim
    ox = dz_util.s2i(b[x+13:x+15])      # object anim

    if mx == 0:
        s = "Pick up Object in " + get_data(p, c1, w1)
    elif mx == 1:
        s = "Drop object onto Floor"
    else:
        s = "Drop onto Object in " + get_data(p, c1, w1)

    if mx != 1 and mx != 2:
        if fx & 1:
            s += ", using slot # in temp 0"
        else:
            s += ", using slot number %d" % sn

    s += ", object anim: "
    if ox != 0xFFFF:
        s += dz_str.read_str2(p, 2, 0x86, ox)   # ?
    else:
        s += "none"
    if fx & 4:
        s += ", using Sim age as filter on object Anim"

    s += ", grasp anim: "
    if sx != 0xFFFF:
        s += dz_str.read_str2(p, 2, 0x81, sx)   # ?
    else:
        s += "none"
    if fx & 2:
        s += ", expecting handedness in temp 3"
    return s

def prim_75(p, b, x, f, node_version):
    """Age"""
    ax = ord(b[x])                      # age type
    fx = ord(b[x+1])                    # flags

    if fx & 1:
        s = "Expect Age in Temp 0"
    else:
        s = "Set Age to "
        if ax == 1:
            s += "Child"
        elif ax == 2:
            s += "Toddler"
        elif ax == 3:
            s += "Teen"
        elif ax == 4:
            s += "Elder"
        elif ax == 7:
            s += "Baby"
        elif ax == 9:
            s += "Young Adult"
        else:
            s += "Adult"
    return s

def prim_76(p, b, x, f, node_version):
    """Array Operation"""
    ax = ord(b[x+1])                    # action (operation)
    ox = ord(b[x+2])                    # owner
    ai = dz_util.s2i(b[x+3:x+5])        # array index
    c1 = ord(b[x+5])                    # iter A
    w1 = dz_util.s2i(b[x+6:x+8])
    c2 = ord(b[x+8])                    # iter B
    w2 = dz_util.s2i(b[x+9:x+11])

    if ox == 0:
        s = "My Object Array: "
    else:
        s = "Stack Object's Object Array: "
    s += dz_str.read_str2(p, 2, 0x118, ai)

    s += ". Operation: "
    if ax == 0:
        s += "Clear contents of array."
    elif ax == 1:
        s += "Return size of array into " + get_data(p, c1, w1) + "."
    elif ax == 2:
        s += "Resize array to size from " + get_data(p, c1, w1) + "."
    elif ax == 3:
        s += "Init all elements of the array with value from " + get_data(p, c1, w1) + "."
    elif ax == 4:
        s += "Insert element to array from " + get_data(p, c1, w1) + " to the front of the array."
    elif ax == 5:
        s += "Insert element to array from " + get_data(p, c1, w1) + " to the back of the array."
    elif ax == 6:
        s += "Insert element to array from " + get_data(p, c1, w1)
        s += " to the position indicated by iterator in " + get_data(p, c2, w2) + "."
    elif ax == 7:
        s += "Remove front element from array."
    elif ax == 8:
        s += "Remove back element from array."
    elif ax == 9:
        s += "Remove element from array at position given by iterator in " + get_data(p, c2, w2) + "."
    elif ax == 0x0A:
        s += "Advance iterator in " + get_data(p, c2, w2)
        s += " to next occurance of element from " + get_data(p, c1, w1) + "."
    elif ax == 0x0B:
        s += "Swap elements of array at iterator " + get_data(p, c1, w1)
        s += " and iterator " + get_data(p, c2, w2)
    elif ax == 0x0C:
        s += "Sort array into highest to lowest order."
    elif ax == 0x0D:
        s += "Sort array into lowest to highest order."
    else:
        s += "Alex Fennell's magic array operation. If you see this string....be afraid."
    return s

def prim_77(p, b, x, f, node_version):
    """Message"""
    target_list = [ "Selectable Sims", "Selectable Sims + Neighbors", \
        "Selectable Sims + NPCS", "Neighbors Only", "NPCS Only", "All Sims", \
        "Objects", "Everything" ]

    ax = ord(b[x])                      # area (filter)
    w1 = dz_util.s2i(b[x+1:x+3])        # message
    tx = ord(b[x+3])                    # target
    fx = ord(b[x+4])                    # flags
    c2 = ord(b[x+5])                    # target/room
    w2 = dz_util.s2i(b[x+6:x+8])
    px = ord(b[x+8])                    # priority
    c3 = ord(b[x+9])                    # user 1
    w3 = dz_util.s2i(b[x+10:x+12])
    c4 = ord(b[x+12])                   # user 2
    w4 = dz_util.s2i(b[x+13:x+15])
    c1 = ord(b[x+15])

    s = "Sending Message ID inside " + get_data(p, c1, w1)
    s += ", priority %d " % px
    if fx & 4:
        s += "Targetting specific target in " + get_data(p, c2, w2)
    else:
        s += "Targetting "
        if tx < len(target_list):
            s += target_list[tx]
        else:
            s += "unknown"

        s += " using filter "
        if ax == 0:
            s += "In Same Room Only"
            if fx & 1:
                s += " targeting room in " + get_data(p, c2, w2)
        elif ax == 1:
            s += "On Same Level Only"
        elif ax == 2:
            s += "On Lot"
        elif ax == 3:
            s += "Inside building Only"
        elif ax == 4:
            s += "OutSide Building Only"

        s += ", getting user data from " + get_data(p, c3, w3)
        s += ", and " + get_data(p, c4, w4)
    return s

def prim_78(p, b, x, f, node_version):
    """RayTrace"""
    c1 = ord(b[x+1])                    # source
    w1 = dz_util.s2i(b[x+2:x+4])
    s1 = ord(b[x+4])                    # slot type 1
    sa = ord(b[x+5])                    # slot number 1
    c2 = ord(b[x+8])                    # target
    w2 = dz_util.s2i(b[x+9:x+11])
    s2 = ord(b[x+11])                   # slot type 2
    sb = ord(b[x+12])                   # slot number 2
    fx = ord(b[x+15])                   # flags

    s = "From Object in " + get_data(p, c1, w1)

    s += " using "
    if s1 == 0:
        s += "point at 3/4 height of object."
    elif s1 == 1:
        s += "targeting slot"
    elif s1 == 2:
        s += "routing slot"
    else:
        s += "containment slot"
    if s1:
        s += " number %d into the world. Object hit stored in Temp 0." % sa
    else:
        s += " To Object in " + get_data(p, c2, w2)
        s += " using "
        if s2 == 0:
            s += "point at 3/4 height of object."
        elif s2 == 1:
            s += "targeting slot"
        elif s2 == 2:
            s += "routing slot"
        else:
            s += "containment slot"
        if s2:
            s += " number %d." % sb

    if fx & 1: s += " Windows Ignored."
    return s

def prim_79(p, b, x, f, node_version):
    """Change Outfit"""
    fx = ord(b[x])                      # flags
    c2 = ord(b[x+1])                    # index
    w2 = dz_util.s2i(b[x+2:x+4])
    gx = dz_util.s2i(b[x+4:x+8])        # GUID
    ix = ord(b[x+8])                    # literal index
    c1 = ord(b[x+9])                    # sim
    w1 = dz_util.s2i(b[x+10:x+12])

    if fx & 0x10:
        s = "Rebuild current outfit on sim in "
    else:
        s = "Change Outfit on sim in "
    s += get_data(p, c1, w1)

    if fx & 1:
        s += " using Stack Object"
    elif fx & 2:
        s += " using "
        p.m_offset += len(s)
        s += get_guid(p, gx)
    elif fx & 0x40:
        s += ", using GUID in Temp 0/1"
    else:
        s += " using that sims outfits"

    s += " as source, using outfit "
    if fx & 4:
        s += "index from " + get_data(p, c2, w2)
    else:
        s += dz_util.chk_range(gStringFA, ix)  # Person outfits

    if fx & 0x20:
        s += ", clearing GUID pointers in person data fields"
    if fx & 8:
        s += ", writing changes to the .iff"    # something amiss here
    return s

def prim_7A(p, b, x, f, node_version):
    """On Timer"""
    tx = dz_util.s2i(b[x:x+2])          # ticks
    tt = ord(b[x+2])                    # tree type
    ex = dz_util.s2i(b[x+3:x+5])        # event tree
    fx = ord(b[x+5])                    # flags
    c1 = ord(b[x+6])                    # param 0
    w1 = dz_util.s2i(b[x+7:x+9])
    c2 = ord(b[x+9])                    # param 1
    w2 = dz_util.s2i(b[x+10:x+12])
    c3 = ord(b[x+12])                   # param 2
    w3 = dz_util.s2i(b[x+13:x+15])
    mx = ord(b[x+15])                   # mode

    if mx == 0:
        s = "Start Timer"
    elif mx == 1:
        s = "Modify Timer"
    elif mx == 2:
        s = "Delete Timer"
    else:
        s = "unknown"

    if mx != 2:
        if fx & 8:
            s += ", Getting ticks till fire from temp 1"
        else:
            s += ", Ticks till fire %d" % tx
        p.m_offset += len(s)
        s += get_etree(p, ex, tt)
        if fx & 2:
            s += ", Looping"
        if fx & 1:
            s += ", using current params as event tree params"
        else:
            s += ", passing in params found in " + get_data(p, c1, w1)
            s += ", " + get_data(p, c2, w2)
            s += ", " + get_data(p, c3, w3)
        if mx == 1 and (fx & 4) != 0:
            s += ", Reset Timer Ticks"
    return s

def prim_7B(p, b, x, f, node_version):
    """Cinematic"""
    sx = dz_util.s2i(b[x:x+2])          # scene index
    fx = ord(b[x+5])                    # flags
    c1 = ord(b[x+6])                    # scene var
    w1 = dz_util.s2i(b[x+7:x+9])
    c2 = ord(b[x+9])                    # array
    w2 = dz_util.s2i(b[x+10:x+12])

    if fx & 0x10:
        s = "Using scene name index found in " + get_data(p, c1, w1)
    else:
        s = "Using Scene "
        if fx & 0x20:
            s += dz_str.read_str2(p, 0, 0x97, sx)   # global
        elif fx & 0x40:
            s += dz_str.read_str2(p, 1, 0x97, sx)   # semiglobal
        else:
            s += dz_str.read_str2(p, 2, 0x97, sx)   # private

    s += ", Using array found in objID in " + get_data(p, c2, w2)

    if fx & 1: s += ", Flip Cinematic Anims Horizontally"
    if fx & 2: s += ", Pass in Flip Cinematic Anims Horizontally value in temp 0"
    if fx & 4: s += ", Start Animations Now"
    if fx & 8: s += ", Show Entire House"
    return s

def prim_7C(p, b, x, f, node_version):
    """Want Satisfy"""
    w5 = dz_util.s2i(b[x+1:x+3])        # ?
    gx = dz_util.s2i(b[x+3:x+7])        # GUID
    c3 = ord(b[x+7])                    # type var
    w3 = dz_util.s2i(b[x+8:x+10])
    c4 = ord(b[x+10])                   # level var
    w4 = dz_util.s2i(b[x+11:x+13])

    s = "GUID 0x%08X " % gx
    if gx in p.prefs.wants_guids:
        w2 = p.prefs.wants_guids.index(gx)
        s += "(" + p.prefs.wants_names[w2] + ") "

        wt = p.prefs.wants_types[w2]
        if wt == Want_None:
            s += "with value "
        elif wt == Want_Sim:
            s += "with person in "
        elif wt == Want_Guid or wt == Want_Category:
            s += "with object in "
        elif wt == Want_Skill:
            s += "with skill "
        elif wt == Want_Career:
            s += "with career "
        elif wt == Want_Badge:
            s += "with badge "
        s += get_data(p, c3, w3)

        if wt >= 4 and wt <= 6:
            s += " and level " + get_data(p, c4, w4)
    return s

def prim_7D(p, b, x, f, node_version):
    """Influence"""
    c1 = ord(b[x])                      # sim
    w1 = dz_util.s2i(b[x+1:x+3])
    ox = ord(b[x+5])                    # array owner
    w2 = dz_util.s2i(b[x+6:x+8])        # array value

    s = "Follow Sim in " + get_data(p, c1, w1)
    if ox:
        s += " stack object's object array: "
    else:
        s += " my object array: "
    s += dz_str.read_str2(p, 2, 0x118, w2)
    return s

def prim_7E(p, b, x, f, node_version):
    """Lua"""
    w1 = dz_util.s2i(b[x:x+2])          # STR# instance
    w2 = dz_util.s2i(b[x+2:x+4])        # index + 1
    w3 = dz_util.s2i(b[x+4:x+6])        # flags
    c4 = ord(b[x+6])                    # param 0
    w4 = dz_util.s2i(b[x+7:x+9])
    c5 = ord(b[x+9])                    # param 1
    w5 = dz_util.s2i(b[x+10:x+12])
    c6 = ord(b[x+12])                   # param 2
    w6 = dz_util.s2i(b[x+13:x+15])

    s = ""
    if w2 != 0:
        w2 -= 1
        if w3 & 2:
            s += "Private - \"" + dz_util.get_html(p,TYPE_FUNCTION,dz_str.read_str2(p, 2, w1, w2))
        elif w3 & 4:
            s += "Semi-Global - \"" + dz_util.get_html(p,TYPE_FUNCTION,dz_str.read_str2(p, 1, w1, w2))
        else:
            s += "Global - \"" + dz_util.get_html(p,TYPE_FUNCTION,dz_str.read_str2(p, 0, w1, w2))
        s += "\""

        if w3 & 1:
            s += ", defined in objLua file"
        else:
            s += ", defined in description"

        if w3 & 8:
            s += ", Passing in params where param 0 = " + get_data(p, c4, w4)
            s += ", param 1 = " + get_data(p, c5, w5)
            s += ", param 2 = " + get_data(p, c6, w6)
    return s

# function pointer table

prim_func = [ prim_00, prim_01, prim_02, prim_03, prim_04, prim_05, prim_06, prim_07, \
    prim_08, prim_09, prim_0A, prim_0B, prim_0C, prim_0D, prim_0E, prim_0F, \
    prim_10, prim_11, prim_12, prim_13, prim_14, prim_15, prim_16, prim_17, \
    prim_18, prim_19, prim_1A, prim_1B, prim_1C, prim_1D, prim_1E, prim_1F, \
    prim_20, prim_21, prim_22, prim_23, prim_24, prim_25, prim_26, prim_27, \
    prim_28, prim_29, prim_2A, prim_2B, prim_2C, prim_2D, prim_2E, prim_2F, \
    prim_30, prim_31, prim_32, prim_33, prim_34, prim_35, prim_36, prim_37, \
    prim_38, prim_39, prim_3A, prim_3B, prim_3C, prim_3D, prim_3E, prim_3F, \
    prim_40, prim_41, prim_42, prim_43, prim_44, prim_45, prim_46, prim_47, \
    prim_48, prim_49, prim_4A, prim_4B, prim_4C, prim_4D, prim_4E, prim_4F, \
    prim_50, prim_51, prim_52, prim_53, prim_54, prim_55, prim_56, prim_57, \
    prim_58, prim_59, prim_5A, prim_5B, prim_5C, prim_5D, prim_5E, prim_5F, \
    prim_60, prim_61, prim_62, prim_63, prim_64, prim_65, prim_66, prim_67, \
    prim_68, prim_69, prim_6A, prim_6B, prim_6C, prim_6D, prim_6E, prim_6F, \
    prim_70, prim_71, prim_72, prim_73, prim_74, prim_75, prim_76, prim_77, \
    prim_78, prim_79, prim_7A, prim_7B, prim_7C, prim_7D, prim_7E ]
