# STR#-related package utilities
#
# - functions:
#
# get_str(data, dict): return a specific string or list (None = not found)
# read_str(package, group, instance, id): return a specific string (None = not found)
# read_str2(package, scope, instance, id): return a specific string or coded string

import dz_util
import dz_pack
from dz_defs import *

def get_str(a, d):
    """Return a specific string or list of strings from raw data."""
    l = 'list' in d
    if not l:
        i = d['id']

    ccode = 1                       # English US (default)
    ccdict = {}
    if 'ccode' in d:
        ccode = d['ccode']

    desc = []
    bDoDesc = False

    if 'desc' in d:
        bDoDesc = True

    bPString = False
    bLocale = True
    bDesc = True
    bPad = False

    w2 = dz_util.s2i(a[0x40:0x42])  # string data type/num of pstrings
    w1 = 0
    c2 = 0
    r = []
    
    if w2 == 0xFFFD:                # -3 = fully localized and described
        w1 = dz_util.s2i(a[0x42:0x44])
        x = 0x44
    elif w2 == 0xFFFE:              # -2 = no-frills cstrings
        w1 = ord(a[0x42])
        x = 0x43
        bLocale = False
        bDesc = False
        bPad = True
    elif w2 < 0x8000:               # other = old-style pstrings
        w1 = w2
        x = 0x42
        bLocale = False
        bPString = True
        bDesc = False

    if l:
        # get the cstring, maxchars = 0x40 at index = 0 (title)
        j = 0x40
        if '\x00' in a[:0x40]: j = a.index('\x00')
        if j:
            r.append( ', Title = %s\n\n' % a[:j] )
        else:
            r.append( '\n\n')
        if bDoDesc: desc.append("")

    if w1 > 0:
        w3 = 0
        while w1 > 0:
            if bLocale or bPad:
                c2 = ord(a[x])      # get the ccode/pad byte of this string
                x += 1

            if bLocale:
                if c2 in ccdict:
                    w3 = ccdict[c2]
                else:
                    w3 = 0          # reset id to 0 when a new ccode appears

                ccdict.update({c2:w3+1})

            if bPString:
                c1 = a[x]           # length
                x += 1
                s = a[x:x+c1]
                x += c1
            else:                   # cstring
                j = a[x:].index('\x00')
                s = a[x:x+j]
                x += (j+1)

            # found a string?
            if (l or w3 == i) and (not bLocale or (c2 == ccode)):
                if l:
                    r.append( "0x%X: %s\n" % (w3, s) )  # add to list
                else:
                    return s            # string found, return it

            if bDesc:   # description
                j = a[x:].index('\x00')
                if bDoDesc and (not bLocale or (c2 == ccode)):
                    desc.append( "%s\n\n" % a[x:x+j] )    # add to list of descriptions
                x += (j+1)

            if x >= len(a):             # do a bit of range checking here
                w1 = 0

            w1 -= 1     # decrement total string count
            w3 += 1     # increment string id value

    if l:
        if bDoDesc:
            return r, desc
        else:
            return r    # list of strings
    else:
        return None     # id not found in STR# data

def read_str(p, g, i, x):
    """Return a specific string."""
    r = p.find({'type': TYPE_STR, 'group': g, 'instance': i})
    if not r:                   # not found
        return r

    b = p.get(r['offset'], r['size'])
    if not b:                   # missing (!)
        return b

    return get_str(b, {'id': x, 'ccode': p.prefs.ccode})

def read_str2(p, g, i, x):
    """Return a specific string or coded string if not found.
    
    Note that g is the scope of the group - not the actual group ID.
    0 = global, 1 = semi-global, other = current group
    """
    if g == 0:
        g2 = GROUP_GLOBAL
    elif g == 1:
        g2 = p.semi
    else:
        g2 = p.group

    if not p.prefs.no_str_lookup:
        s = read_str(p, g2, i, x)
        if s: return s

    if g == 0:
        return "[Global STR# 0x%X:0x%X]" % (i, x)
    elif g == 1:
        return "[SemiGlobal STR# 0x%X:0x%X]" % (i, x)
    else:
        return "[STR# 0x%X:0x%X]" % (i, x)
