# pointy-clicky-launchy stuff

import logging
import os
import pango
import sys

try:
    import pygtk
except:
    sys.exit(1)

import gobject
import gtk
import gtk.gdk
import gtk.glade

import dz_cont
import dz_edit
import dz_label
import dz_mpack
import dz_nav
import dz_pack
import dz_prefs
import dz_types
import dz_util
import dz_view
import dz_xpm
from dz_defs import *

# main window

class Window1:
    def __init__(self):
        logging.info("main window init")

        # load widgets
        self.wTree = gtk.glade.XML("dizziness.glade", "window1")

        # connect signals to my callbacks
        s = {}
        for a in dir(self):
            s[a] = getattr(self, a)
        self.wTree.signal_autoconnect(s)

        # set icon
        self.icon = gtk.gdk.pixbuf_new_from_file("dizziness.png")
        self.wTree.get_widget("window1").set_icon(self.icon)

        # default variables
        self.w2 = None      # prefs
        self.w3 = None      # about
        self.w4 = None      # font
        self.pack1 = None
        self.pack2 = None
        self.pack_name = ""
        self.cv_event = None
        self.cv_time = 0
        self.current_cv = None
        self.stop_find = False
        self.titles_only = False
        self.title = "dizziness 1.1a"
        self.blank_tab = "New"
        self.cursor_ptr = gtk.gdk.Cursor(gtk.gdk.LEFT_PTR)
        self.cursor_hand = gtk.gdk.Cursor(gtk.gdk.HAND2)
        self.cursor_text = gtk.gdk.Cursor(gtk.gdk.XTERM)
#        self.cursor_ptr = dz_xpm.arrow_cursor
#        self.cursor_hand = dz_xpm.hand_cursor
#        self.cursor_text = dz_xpm.xterm_cursor
        self.max_recents = 9    # move to prefs?
        self.case_sensitive = False
        self.nav_size = 25      # prefs?
        self.nav_stacks = {}
        self.nav_cursor = {}

        # load some other stuff
        self.prefs = dz_prefs.Prefs()
        self.prefs.loaddef()

        # remember window state, size
        if self.prefs.maximize:
            self.wTree.get_widget("window1").maximize()
        elif self.prefs.w_size:     # don't maximize + resize
            x, y = self.prefs.w_size
            self.wTree.get_widget("window1").resize(x, y)

        self.set_paned_flag = True   # set paned when ready

        # show
        self.wTree.get_widget("window1").show_all()

        # set up result notebook
        nb = self.wTree.get_widget("notebook2")
        nb.set_show_tabs(True)
        nb.set_scrollable(True)
        lb = dz_label.label(self.blank_tab)
        lb.show_all()
        sw = nb.get_nth_page(0)         # get ScrolledWindow
        lb.button.connect("clicked", self.on_label_button_clicked, sw)
        nb.set_tab_label(sw, lb)
        xv = self.wTree.get_widget("resultview")
        xv.connect("motion-notify-event", self.on_motion_xv, sw)
        xv.connect("button-press-event", self.on_button_xv, sw)

        # check menu items
        w = self.wTree.get_widget("word_wrap1")
        w.set_active(self.prefs.wrap != 0)

        # set view toolbar
        self.wTree.get_widget("view_toolbar1").set_active(self.prefs.view_toolbar)

        # set view_mode radio menu items
        if self.prefs.view_mode == 0:
            self.wTree.get_widget("view_code1").set_active(True)
            self.wTree.get_widget("view_data1").set_active(False)
            self.wTree.get_widget("databutton").set_active(False)
        else:
            self.wTree.get_widget("view_code1").set_active(False)
            self.wTree.get_widget("view_data1").set_active(True)
            self.wTree.get_widget("databutton").set_active(True)

        # stop defaults to disabled
        self.wTree.get_widget("stopbutton").set_sensitive(False)

        # recent files menu
        dz_nav.do_set_recents(self)

        # default = no history/navigation cursor
        self.wTree.get_widget("backbutton").set_sensitive(False)
        self.wTree.get_widget("forwardbutton").set_sensitive(False)

        # editor bar on/off
        if self.prefs.view_editbar:
            self.wTree.get_widget("view_editor1").set_active(False)
            self.wTree.get_widget("view_editor1").set_active(True)      # strobe effect
        else:
            self.wTree.get_widget("view_editor1").set_active(True)
            self.wTree.get_widget("view_editor1").set_active(False)     # disco time :D

        # setup edit_notebook
        nb = self.wTree.get_widget("edit_notebook")
        nb.set_show_tabs(False)     # these are meant for glade hacking only
        nb.set_show_border(False)   # as the pages are selected by context
        nb.set_scrollable(False)

        # setup status/progress vars
        self.status_text = ""
        self.old_status_text = ""
        self.progress_show = False
        self.old_progress_show = False
        self.progress_amount = 0
        self.old_progress_amount = 0
        self.progress_divisor = 100     # default = percent

        self.wTree.get_widget("progressbar1").set_size_request(100, 10)
        self.wTree.get_widget("progressbar1").hide()
        
        # note: unused items/buttons disabled here:
        self.wTree.get_widget("new1").set_sensitive(False)
        self.wTree.get_widget("save1").set_sensitive(False)
        self.wTree.get_widget("save_as1").set_sensitive(False)
        self.wTree.get_widget("undo1").set_sensitive(False)
        self.wTree.get_widget("redo1").set_sensitive(False)
        self.wTree.get_widget("cut1").set_sensitive(False)
        self.wTree.get_widget("paste1").set_sensitive(False)
        self.wTree.get_widget("delete1").set_sensitive(False)
        self.wTree.get_widget("newbutton").set_sensitive(False)
        self.wTree.get_widget("savebutton").set_sensitive(False)

    # main UI functions

    def wait_gtk(self):
        widget = self.wTree.get_widget("statusbar1")

        if self.status_text != self.old_status_text:
            cid = widget.get_context_id("timed")
            if self.old_status_text:
                widget.pop(cid)             # pop old text
            if self.status_text:
                widget.push(cid, self.status_text)
            self.old_status_text = self.status_text

        widget = self.wTree.get_widget("progressbar1")

        if self.progress_show != self.old_progress_show:
            if self.progress_show:
                widget.show()
            else:
                widget.hide()
            self.old_progress_show = self.progress_show

        if self.progress_amount != self.old_progress_amount:
            widget.set_fraction(float(self.progress_amount) / float(self.progress_divisor))
            self.old_progress_amount = self.progress_amount

        while gtk.events_pending(): gtk.main_iteration(False)   # idle until events have cleared

    def reconfig(self):
        logging.info("reconfigure definitions")

        self.wTree.get_widget("menubar1").set_sensitive(False)
        self.wTree.get_widget("toolbar1").set_sensitive(False)
        self.wait_gtk()
        self.prefs.reconfig(self)
        self.wTree.get_widget("menubar1").set_sensitive(True)
        self.wTree.get_widget("toolbar1").set_sensitive(True)

        dz_cont.do_open(self, self.prefs.objects_name)

    def set_paned(self):
        # paned positions
        if self.prefs.hpaned_pos < 0:
            self.wTree.get_widget("hpaned1").set_position(200)
        else:
            self.wTree.get_widget("hpaned1").set_position(self.prefs.hpaned_pos)
        if self.prefs.vpaned_pos < 0:
            self.wTree.get_widget("vpaned1").set_position(300)
        else:
            self.wTree.get_widget("vpaned1").set_position(self.prefs.vpaned_pos)

    def run(self):
        self.wait_gtk()
        self.set_paned()

        # reconfig defs if defs file not found
        if len(self.prefs.glob_bhav_names) == 0 and self.can_reconfig():
            self.reconfig()
        elif self.can_open_global():
            dz_cont.do_open(self, self.prefs.objects_name)
        dz_util.do_audit(self)

        logging.info("call gtk.main")
        gtk.main()

    def quit(self):
        # get paned positions
        self.prefs.hpaned_pos = self.wTree.get_widget("hpaned1").get_position()
        self.prefs.vpaned_pos = self.wTree.get_widget("vpaned1").get_position()

        # save preferences
        self.prefs.save()

        # close child windows
        if self.w4:
            self.w4.wTree.get_widget("window4").destroy()
        if self.w3:
            self.w3.wTree.get_widget("window3").destroy()
        if self.w2:
            self.w2.wTree.get_widget("window2").destroy()

        # done
        logging.info("call gtk.main_quit")
        gtk.main_quit()

    def can_open_global(self):                  # check for objects.package
        try:
            f = open(self.prefs.objects_name, 'rb')
        except:
            f = None
        if not f:
            return False
        else:
            f.close()
            return True

    def can_reconfig(self):                     # check for Wants.package
        try:
            f = open(self.prefs.wants_name, 'rb')
        except:
            f = None
        if not f:
            return False
        else:
            f.close()
            return self.can_open_global()

    def do_title(self, filename):
        h, t = os.path.split(filename)
        self.wTree.get_widget("window1").set_title("%s - %s" % (t, self.title))
        return t

    def on_get_tv_name(self, column, cell, model, i):
        s = model.get_value(i, 0)       # name
        cell.set_property('text', s)
        cell.set_property('font', self.prefs.tree_font)

    def get_current_pack(self):
        # get currently *selected* tab's pack
        if self.current_cv == self.wTree.get_widget("obj_contentview"):
            return self.pack1
        else:
            return self.pack2

    def set_content_page(self, page):
        # note: this invokes a switch_page which will change self.current_cv
        nb = self.wTree.get_widget("notebook1")
        nb.set_current_page(page)

    def do_cv_refresh(self):
        self.cv_event = 0
        p = self.get_current_pack()
        if p:
            nb = self.wTree.get_widget("notebook2")
            for i in range(nb.get_n_pages()):
                dz_cont.do_contentview_change(self, i)    # update all the pages

    def do_tv_refresh(self):
        self.wTree.get_widget("obj_contentview").queue_draw()
        self.wTree.get_widget("contentview").queue_draw()
        self.wTree.get_widget("searchview").queue_draw()

    # event handlers

    def on_window1_destroy(self, widget):       # close main = quit
        self.quit()

    def on_window1_state_event(self, widget, event):    # track maximize
        self.prefs.maximize = \
            (widget.window.get_state() & gtk.gdk.WINDOW_STATE_MAXIMIZED) != 0

    def on_window1_configure_event(self, widget, event):
        self.prefs.w_size = widget.window.get_size()    # track size

    # control handlers

    def on_new1_activate(self, widget):         # todo: new
        pass

    def on_new_tab1_activate(self, widget):     # create a new tab
        dz_nav.do_new_tab(self)

    def on_open1_activate(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(buttons=b)
        d.set_title("Open package file...")     # todo: move this string
        f = gtk.FileFilter()
        f.add_pattern("*.package")
        d.set_filter(f)
        d.set_filename(dz_util.get_good_filename(self.pack_name))
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.pack_name = d.get_filename()
            d.destroy()
            dz_cont.do_open(self, self.pack_name)
        else:
            d.destroy()
    def on_save1_activate(self, widget):        # todo: save
        pass
    def on_save_as1_activate(self, widget):     # todo: save as
        pass

    def on_close_tab1_activate(self, widget):
        nb = self.wTree.get_widget("notebook2") # close current tab
        if nb.get_n_pages() > 0:
            logging.info("close tab %d" % nb.get_current_page())
            dz_nav.do_remove_nav_stack(self, nb.get_current_page())
            nb.remove_page(nb.get_current_page())
        if nb.get_n_pages() == 0:
            self.wTree.get_widget("close_tab1").set_sensitive(False)
            self.wTree.get_widget("backbutton").set_sensitive(False)
            self.wTree.get_widget("forwardbutton").set_sensitive(False)
        dz_edit.do_select_edit_page(self)

    def on_label_button_clicked(self, widget, sw):
        nb = self.wTree.get_widget("notebook2")
        logging.info("close tab %d" % nb.page_num(sw))
        dz_nav.do_remove_nav_stack(self, nb.page_num(sw))
        nb.remove_page(nb.page_num(sw))
        if nb.get_n_pages() == 0:
            self.wTree.get_widget("close_tab1").set_sensitive(False)
            self.wTree.get_widget("backbutton").set_sensitive(False)
            self.wTree.get_widget("forwardbutton").set_sensitive(False)
        dz_edit.do_select_edit_page(self)

    def on_quit1_activate(self, widget):
        self.quit()

    def on_dump_text1_activate(self, widget):
        sm = self.prefs.dump_mode   # save mode
        self.prefs.dump_mode = 0
        dz_view.do_dump(self, 0)
        self.prefs.dump_mode = sm   # restore mode

    def on_dump_html1_activate(self, widget):
        sm = self.prefs.dump_mode   # save mode
        self.prefs.dump_mode = 1
        dz_view.do_dump(self, 1)
        self.prefs.dump_mode = sm   # restore mode

    def on_dump_data1_activate(self, widget):
        dz_view.do_dump(self, 2)

    def on_dump_group_tree1_activate(self, widget):
        dn = self.prefs.dump_name
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(action=gtk.FILE_CHOOSER_ACTION_SAVE, buttons=b)
        d.set_title("Dump Group Tree As...")
        d.set_default_response(gtk.RESPONSE_ACCEPT) # type new name, hit return = done
        d.set_filename(dz_util.get_good_filename(dn))
        d.set_do_overwrite_confirmation(True)       # don't overwrite by default
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.prefs.dump_name = d.get_filename()
            d.destroy()
            dz_view.do_dump(self, 3)
        else:
            d.destroy()
        self.prefs.dump_name = dn   # restore old dump name

    def on_extract_type1_activate(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.Dialog(title="Extract Type...", flags=gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT, buttons=b)
        cb = gtk.combo_box_new_text()               # cb = combobox
        cbv = []
        for i in range(len(dz_types.known_names)):
            cb.append_text("%s (0x%X)" % (dz_types.known_names[i], dz_types.known_values[i]))
            cbv.append(dz_types.known_values[i])
        for i in range(len(dz_types.unknown_names)):
            cb.append_text("%s (0x%X)" % (dz_types.unknown_names[i], dz_types.unknown_values[i]))
            cbv.append(dz_types.unknown_values[i])
        d.vbox.pack_start(cb, False, True, 0)
        cb.show()
        e = gtk.FileChooserWidget(action=gtk.FILE_CHOOSER_ACTION_SAVE)  # e = file chooser
        d.vbox.pack_start(e, True, True, 0)
        e.show()
        e.set_do_overwrite_confirmation(True)       # don't overwrite by default
        d.set_default_response(gtk.RESPONSE_ACCEPT)
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            x = cb.get_active()
            n = e.get_filename()
            d.destroy()
            if x != -1:
                if n[-8:].lower() != ".package":    # add extension if not already there
                    n += ".package"
                # get selected type value
                v = cbv[x]

                self.status_text = "Extracting..."
                self.wait_gtk()

                logging.info("extract type 0x%X" % v)

                # create new package file
                p2 = dz_mpack.MutablePackage(n, self.prefs, True)
                p2.build_start()

                # copy types to new package
                p = self.get_current_pack()
                g = 0
                r = p.find({'type': v})
                while r:
                    if g != r['group']:
                        g = r['group']
                        s = dz_view.get_group_name(p, g)
                        self.status_text = "Extracting %s..." % s
                        self.wait_gtk()

                    b = p.get(r['offset'], r['size'])
                    if b:
                        d = dict(group=r['group'], type=v, instance=r['instance'], subtype=r['subtype'], \
                            data=b)
                        p2.build_append(d)
                    r = p.find({'type': v, 'index': r['index']})

                # save
                p2.build_end()
                self.status_text = ""
                self.wait_gtk()
        else:
            d.destroy()

    def on_dump_selection1_activate(self, widget):
        dn = self.prefs.dump_name
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(action=gtk.FILE_CHOOSER_ACTION_SAVE, buttons=b)
        d.set_title("Dump Group As...")
        d.set_default_response(gtk.RESPONSE_ACCEPT) # type new name, hit return = done
        d.set_filename(dz_util.get_good_filename(dn))
        d.set_do_overwrite_confirmation(True)       # don't overwrite by default
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.prefs.dump_name = d.get_filename()
            d.destroy()
            dz_view.do_dump(self, 4)
        else:
            d.destroy()
        self.prefs.dump_name = dn   # restore old dump name

    def on_extract_selection1_activate(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(action=gtk.FILE_CHOOSER_ACTION_SAVE, buttons=b)
        d.set_title("Extract As...")
        d.set_default_response(gtk.RESPONSE_ACCEPT) # type new name, hit return = done
        d.set_do_overwrite_confirmation(True)       # don't overwrite by default
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            n = d.get_filename()
            d.destroy()
            if n[-8:].lower() != ".package":        # add extension if not already there
                n += ".package"

            # get current group
            p = self.get_current_pack()
            nb = self.wTree.get_widget("notebook2")
            sw = nb.get_nth_page(nb.get_current_page())     # current scrollwindow
            b = sw.get_child().get_buffer()                 # textbuffer
            g = b.get_data("current_group")

            logging.info("extract group 0x%X" % g)

            # create new package file
            p2 = dz_mpack.MutablePackage(n, self.prefs, True)
            p2.build_start()

            # copy group to new package
            r = p.find({'group': g})
            while r:
                b = p.get(r['offset'], r['size'])
                if b:
                    d = dict(group=g, type=r['type'], instance=r['instance'], subtype=r['subtype'], \
                        data=b)
                    p2.build_append(d)
                r = p.find({'group': g, 'index': r['index']})

            # save
            p2.build_end()

        else:
            d.destroy()

    def on_view_toolbar1_activate(self, widget):
        if widget.get_active():
            self.wTree.get_widget("toolbar1").show()
            self.prefs.view_toolbar = True
        else:
            self.wTree.get_widget("toolbar1").hide()
            self.prefs.view_toolbar = False

    def on_view_editor1_activate(self, widget):
        if widget.get_active():
            self.wTree.get_widget("edit_hbox").show()
            self.prefs.view_editbar = True
            if not self.wTree.get_widget("editorbutton").get_active():
                self.wTree.get_widget("editorbutton").set_active(True)
        else:
            self.wTree.get_widget("edit_hbox").hide()
            self.prefs.view_editbar = False
            if self.wTree.get_widget("editorbutton").get_active():
                self.wTree.get_widget("editorbutton").set_active(False)

    def on_view_code1_activate(self, widget):
        if widget.get_active():
            self.prefs.view_mode = 0
            self.wTree.get_widget("databutton").set_active(False)
            self.do_cv_refresh()

    def on_view_data1_activate(self, widget):
        if widget.get_active():
            self.prefs.view_mode = 1
            self.wTree.get_widget("databutton").set_active(True)
            self.do_cv_refresh()

    def on_word_wrap1_activate(self, widget):
        if widget.get_active():
            self.prefs.wrap = 1
        else:
            self.prefs.wrap = 0
        self.do_cv_refresh()

    def on_font1_activate(self, widget):
        if not self.w4:
            self.wTree.get_widget("tree_font1").set_sensitive(False)
            widget.set_sensitive(False)         # disable this widget
            self.w4 = Window4(self, 1)

    def on_tree_font1_activate(self, widget):
        if not self.w4:
            self.wTree.get_widget("font1").set_sensitive(False)
            widget.set_sensitive(False)         # disable this widget
            self.w4 = Window4(self, 2)

    def on_preferences1_activate(self, widget):
        if not self.w2:
            self.w2 = Window2(self)
            widget.set_sensitive(False)         # disable this widget
            self.wTree.get_widget("prefsbutton").set_sensitive(False)

    def on_openrecent_activate(self, widget, index):
        s = self.prefs.recent_files[:]
        if self.prefs.holiday1:
            s.insert(0, self.prefs.holiday1)
        if self.prefs.holiday2:
            s.insert(0, self.prefs.holiday2)
        dz_cont.do_open(self, s[index])

    def on_undo1_activate(self, widget):        # todo: undo
        pass
    def on_redo1_activate(self, widget):        # todo: redo
        pass
    def on_cut1_activate(self, widget):         # todo: cut
        pass

    def on_copy1_activate(self, widget):
        nb = self.wTree.get_widget("notebook2")
        p = nb.get_current_page()
        if p >= 0:
            sw = nb.get_nth_page(p)             # current scrollwindow
            v = sw.get_child()                  # textview
            v.emit("copy-clipboard")            # pass this off to the handler

    def on_paste1_activate(self, widget):       # todo: paste
        pass
    def on_delete1_activate(self, widget):      # todo: del
        pass

    def on_about1_activate(self, widget):
        if not self.w3:
            self.w3 = Window3(self)
            widget.set_sensitive(False)
            self.wTree.get_widget("aboutbutton").set_sensitive(False)
    def on_aboutbutton_clicked(self, widget):
        if not self.w3:
            self.w3 = Window3(self)
            widget.set_sensitive(False)
            self.wTree.get_widget("about1").set_sensitive(False)

    def on_newbutton_clicked(self, widget):
        self.on_new1_activate(widget)           # pass off to menu
    def on_openbutton_clicked(self, widget):
        self.on_open1_activate(widget)
    def on_savebutton_clicked(self, widget):
        self.on_save1_activate(widget)

    def on_undobutton_clicked(self, widget):
        self.on_undo1_activate(widget)          # pass off to menu
    def on_redobutton_clicked(self, widget):
        self.on_redo1_activate(widget)
    def on_cutbutton_clicked(self, widget):
        self.on_cut1_activate(widget)
    def on_copybutton_clicked(self, widget):
        self.on_copy1_activate(widget)
    def on_pastebutton_clicked(self, widget):
        self.on_paste1_activate(widget)

    def on_backbutton_clicked(self, widget):
        nb = self.wTree.get_widget("notebook2")
        p = nb.get_current_page()
        dz_nav.do_nav_indexed(self, self.nav_cursor[p] - 1)

    def on_forwardbutton_clicked(self, widget):
        nb = self.wTree.get_widget("notebook2")
        p = nb.get_current_page()
        dz_nav.do_nav_indexed(self, self.nav_cursor[p] + 1)

    def on_backmenu_activate(self, widget, index):
        dz_nav.do_nav_indexed(self, index)

    def on_forwardmenu_activate(self, widget, index):
        dz_nav.do_nav_indexed(self, index)

    def on_jumpbutton_clicked(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.Dialog(title="Jump to...", flags=gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT, buttons=b)
        e = gtk.Entry()                         # e = my text entry
        d.vbox.pack_start(e, True, True, 0)
        e.show()
        e.set_activates_default(True)           # Enter = OK
        d.set_default_response(gtk.RESPONSE_ACCEPT)
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            x = e.get_text()
            p = self.get_current_pack()
            try:
                x2 = eval(x)    # try to get some value
                if x2 in self.prefs.object_guids:
                    i = self.prefs.object_guids.index(x2)
                    x2 = self.prefs.object_group[i]
                elif x2 in self.prefs.holiday_guids:
                    i = self.prefs.holiday_guids.index(x2)
                    x2 = self.prefs.holiday_group[i]
                self.cv_event = 0
                dz_nav.do_goto_value(self, x2, None, None, None, True)  # goto group hash
            except:
                pass

        d.destroy()

    def on_editorbutton_toggled(self, widget):
        if widget.get_active():
            self.wTree.get_widget("view_editor1").set_active(True)
        else:
            self.wTree.get_widget("view_editor1").set_active(False)

    def on_prefsbutton_clicked(self, widget):
        if not self.w2:
            self.w2 = Window2(self)
            widget.set_sensitive(False)         # disable this widget
            self.wTree.get_widget("preferences1").set_sensitive(False)

    def on_databutton_toggled(self, widget):
        self.prefs.view_mode = widget.get_active()
        if self.prefs.view_mode == 0:
            self.wTree.get_widget("view_code1").set_active(True)
        else:
            self.wTree.get_widget("view_data1").set_active(True)
        self.do_cv_refresh()

    def on_dumpbutton_clicked(self, widget):
        self.on_dump_selection1_activate(widget)

    def on_motion_xv(self, widget, event, sw):
        if (event.is_hint):
            x, y, s = event.window.get_pointer()
        else:
            x, y, s = event.x, event.y, event.state
        x, y = widget.window_to_buffer_coords(gtk.TEXT_WINDOW_WIDGET, int(x), int(y))
        it = widget.get_iter_at_location(x, y)

        b = widget.get_buffer()

        # for more colors see /etc/X11/rgb.txt

        if 1 == 0:  # todo: add to prefs
            t1 = b.get_tag_table().lookup("hover")
            if not t1:
                t1 = b.create_tag("hover", background='light gray')
            x = it.get_line()
            b.remove_tag(t1, b.get_start_iter(), b.get_end_iter())
            b.apply_tag(t1, b.get_iter_at_line(x), b.get_iter_at_line(x+1))

        t2 = widget.get_buffer().get_tag_table().lookup("link")

        # change pointer behavior to be more browser-like
        if ord(it.get_char()) == 0:
            event.window.set_cursor(self.cursor_ptr)
        elif t2 and it.has_tag(t2):
            event.window.set_cursor(self.cursor_hand)
        else:
            event.window.set_cursor(self.cursor_text)

    def on_button_xv(self, widget, event, sw):
        x, y = widget.window_to_buffer_coords(gtk.TEXT_WINDOW_WIDGET, int(event.x), int(event.y))
        it = widget.get_iter_at_location(x, y)
        b2 = widget.get_buffer()
        return dz_cont.do_button(self, event, it, b2)

    def on_contentview_button_press_event(self, widget, event):
        self.cv_event = event.button
        self.cv_time = event.time

    def on_contentview_cursor_changed(self, widget):
        if self.cv_event == -1:             # quiet mode
            return

        nb = self.wTree.get_widget("notebook2")
        dz_cont.do_contentview_change(self, nb.get_current_page())

        # push path onto the navigation stack
        path, tvc = self.current_cv.get_cursor()
        dz_nav.set_nav_stack(self, nb.get_current_page(), (self.current_cv, path))  # path = (treeview, path)

    def on_contentview_row_expanded(self, widget, i, p):
        # get first child of expander
        m = widget.get_model()
        i2 = m.iter_children(i)
        dz_cont.do_expander(self, m, i2)

    def on_searchentry_activate(self, widget):  # widget.activate() or enter key
        self.on_findbutton_clicked(widget)

    def on_findbutton_clicked(self, widget):
        self.stop_find = False
        self.wTree.get_widget("findbutton").set_sensitive(False)
        self.wTree.get_widget("stopbutton").set_sensitive(True)

        # disconnect the old store, remove old column
        tv = self.wTree.get_widget("searchview")
        tv.set_model()
        x = tv.get_column(0)
        if x: tv.remove_column(x)

        # new list store (name, group, type, subtype, instance, meta)
        # note: this store needs the same fields as the content tree
        # because of the cursor_changed routine (see below)
        ls = gtk.ListStore(str, 'guint', 'guint', 'guint', 'guint', int)
        cell = gtk.CellRendererText()
        tvc = gtk.TreeViewColumn("Results", cell)
        tvc.set_cell_data_func(cell, self.on_get_tv_name)

        # connect it up
        tv.append_column(tvc)
        tv.set_model(ls)

        self.status_text = "Finding..."
        self.wait_gtk()

        # do find
        fs = self.wTree.get_widget("searchentry").get_text()
        if not self.case_sensitive:
            fs = fs.lower()
        tv = self.current_cv
        m = tv.get_model()
        j = m.get_iter_root()
        p = self.get_current_pack()

        # search content tree
        while not self.stop_find and (j != None):
            n = m.get_value(j, 0)
            if not self.case_sensitive:
                n2 = n.lower()
            else:
                n2 = n
            x = m.get_value(j, 5)

            if n2.find(fs) != -1:            # found in name?
                ls.append([n, m.get_value(j, 1), m.get_value(j, 2), \
                    m.get_value(j, 3), m.get_value(j, 4), m.get_value(j, 5)])
                self.wait_gtk()

            elif x == 0 and not self.titles_only:
                s, sx = dz_view.get_view(p, m.get_value(j, 1), m.get_value(j, 2), \
                    m.get_value(j, 3), m.get_value(j, 4), 0)
                if not self.case_sensitive:
                    s2 = s.lower()
                else:
                    s2 = s
                if s2.find(fs) != -1:       # found in information?
                    ls.append([n, m.get_value(j, 1), m.get_value(j, 2), \
                        m.get_value(j, 3), m.get_value(j, 4), m.get_value(j, 5)])
                    self.wait_gtk()

            # tree-style next
            j = dz_nav.do_tree_style_next(m, j)

        self.status_text = ""
        self.wait_gtk()

        self.wTree.get_widget("findbutton").set_sensitive(True)
        self.wTree.get_widget("stopbutton").set_sensitive(False)

    def on_stopbutton_clicked(self, widget):
        self.stop_find = True

    def on_searchview_cursor_changed(self, widget):
        # change the content browser to this selection
        p, tvc = widget.get_cursor()
        m = widget.get_model()
        i = m.get_iter(p)
        dz_nav.do_goto_value(self, m.get_value(i, 1), m.get_value(i, 2), m.get_value(i, 3), \
            m.get_value(i, 4), False)

    def on_searchview_row_activated(self, widget, path, view_col):  # double-click (or space/enter)
        # make the content browser visible
        if self.current_cv == self.wTree.get_widget("obj_contentview"):
            self.set_content_page(0)
        else:
            self.set_content_page(1)

    def on_casecheckbutton_toggled(self, widget): # on_file changed to case (file searching idea tossed)
        self.case_sensitive = widget.get_active()

    def on_titlescheckbutton_toggled(self, widget):
        self.titles_only = widget.get_active()

    def on_notebook1_switch_page(self, widget, page, page_num):
        # note: do not use page (not usable in pygtk)
        if page_num == 0:                       # set current_cv
            self.current_cv = self.wTree.get_widget("obj_contentview")
        elif page_num == 1:
            self.current_cv = self.wTree.get_widget("contentview")

        if page_num < 2:
            p = self.get_current_pack()
            if p and p.toc_valid:
                self.do_title(p.filename)       # set window title

    def on_notebook2_switch_page(self, widget, page, page_num):
        # note: do not use page (not usable in pygtk)
        sw = widget.get_nth_page(page_num)      # current scrollwindow
        b = sw.get_child().get_buffer()         # textbuffer
        fc = self.get_current_pack().filename
        if b:
            f = b.get_data("file_name")
            p1 = b.get_data("current_pos")

            if f and f != fc:   # switching to a different file?
                if f == self.prefs.objects_name:
                    self.set_content_page(0)       # switch to global
                else:
                    self.set_content_page(1)       # switch to other

            if f:               # goto this buffer's path
                tv = self.current_cv
                self.cv_event = -1                  # quiet mode (notebook isn't ready for update)
                tv.set_cursor(p1)

        dz_nav.set_nav_buttons(self, page_num)
        dz_edit.do_select_edit_page(self, page_num)

    def on_close_editor_button_clicked(self, widget):
        self.wTree.get_widget("view_editor1").set_active(False)

    def on_hex_remove_button_clicked(self, widget):
        pass

    def on_hex_insert_button_clicked(self, widget):
        pass

    def on_hex_goto_button_clicked(self, widget):
        pass

    def on_hex_find_button_clicked(self, widget):
        pass

# preferences

class Window2:
    def __init__(self, main_window):
        logging.info("prefs window init")

        self.w1 = main_window
        self.prefs = main_window.prefs

        # load widgets
        self.wTree = gtk.glade.XML("dizziness.glade", "window2")

        # connect signals to my callbacks
        s = {}
        for a in dir(self):
            s[a] = getattr(self, a)
        self.wTree.signal_autoconnect(s)

        # set icon
        self.icon = gtk.gdk.pixbuf_new_from_file("dizziness.png")
        self.wTree.get_widget("window2").set_icon(self.icon)

        # show
        self.wTree.get_widget("window2").show_all()

        # remember settings
        b = self.prefs.dump_one
        if b:
            self.wTree.get_widget("onefilebutton").set_active(True)
        else:
            self.wTree.get_widget("splitbutton").set_active(True)
        self.wTree.get_widget("dumpentry").set_sensitive(b)
        self.wTree.get_widget("getfilebutton1").set_sensitive(b)

        self.wTree.get_widget("dumpentry").set_text(self.prefs.dump_name)

        if self.prefs.dump_mode == 0:
            self.wTree.get_widget("dumptextbutton").set_active(True)
        else:
            self.wTree.get_widget("dumphtmlbutton").set_active(True)

        self.wTree.get_widget("unknowncheck").set_active(self.prefs.view_unknown)

        self.wTree.get_widget("objectsentry").set_text(self.prefs.objects_name)
        self.wTree.get_widget("wantsentry").set_text(self.prefs.wants_name)

        if self.prefs.holiday1:
            self.wTree.get_widget("holiday1entry").set_text(self.prefs.holiday1)
        if self.prefs.holiday2:
            self.wTree.get_widget("holiday2entry").set_text(self.prefs.holiday2)

        self.wTree.get_widget("reconfigbutton").set_sensitive(self.w1.can_reconfig())

        ccb = self.wTree.get_widget("ccodecombobox")
        if ccb.get_active() < 0:
            m = gtk.ListStore(str)
            for i in range(1, 45):
                m.append([dz_util.get_lang(i, False)])
            cell = gtk.CellRendererText()
            ccb.pack_start(cell, True)
            ccb.add_attribute(cell, 'text', 0)
            ccb.set_model(m)
            ccb.set_active(self.prefs.ccode - 1)

    def on_window2_destroy(self, widget):
        # reenable the prefs widget on the main window
        self.w1.wTree.get_widget("prefsbutton").set_sensitive(True)
        self.w1.wTree.get_widget("preferences1").set_sensitive(True)
        self.w1.w2 = None

    def on_onefilebutton_toggled(self, widget):
        b = widget.get_active()
        # disable text entry and button if "one file" is not selected
        self.wTree.get_widget("dumpentry").set_sensitive(b)
        self.wTree.get_widget("getfilebutton1").set_sensitive(b)
        self.prefs.dump_one = b
    def on_splitbutton_toggled(self, widget):   # todo: decide whether this is redundant
        pass

    def on_dumpentry_changed(self, widget):
        self.prefs.dump_name = widget.get_chars(0, -1)
    def on_getfilebutton1_clicked(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(action=gtk.FILE_CHOOSER_ACTION_SAVE, buttons=b)
        d.set_title("Dump File As...")              # todo: move this string (reword?)
        d.set_default_response(gtk.RESPONSE_ACCEPT) # type new name, hit return = done
        d.set_filename(dz_util.get_good_filename(self.prefs.dump_name))
        d.set_do_overwrite_confirmation(True)       # don't overwrite by default
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.prefs.dump_name = d.get_filename()
            self.wTree.get_widget("dumpentry").set_text(self.prefs.dump_name)
        d.destroy()

    # change default (i.e. the toolbar button) dump mode
    def on_dumptextbutton_toggled(self, widget):
        if widget.get_active():
            self.prefs.dump_mode = 0
    def on_dumphtmlbutton_toggled(self, widget):
        if widget.get_active():
            self.prefs.dump_mode = 1
    # todo: add more modes (maybe)

    def on_unknowncheck_toggled(self, widget):
        self.prefs.view_unknown = widget.get_active()

    def on_objectsentry_changed(self, widget):
        self.prefs.objects_name = widget.get_chars(0, -1)
        self.wTree.get_widget("reconfigbutton").set_sensitive(self.w1.can_reconfig())

    def on_getfilebutton2_clicked(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(buttons=b)
        d.set_title("Select objects.package...")          # todo: move this string
        f = gtk.FileFilter()
        f.add_pattern("objects.package")    # only one filename is acceptable here
        d.set_filter(f)
        d.set_filename(dz_util.get_good_filename(self.prefs.objects_name))
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.prefs.objects_name = d.get_filename()
            self.wTree.get_widget("objectsentry").set_text(self.prefs.objects_name)
            self.wTree.get_widget("reconfigbutton").set_sensitive(self.w1.can_reconfig())
        d.destroy()

    def on_wantsentry_changed(self, widget):
        self.prefs.wants_name = widget.get_chars(0, -1)
        self.wTree.get_widget("reconfigbutton").set_sensitive(self.w1.can_reconfig())
    def on_getfilebutton3_clicked(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(buttons=b)
        d.set_title("Select Wants.package...")          # todo: move this string
        f = gtk.FileFilter()
        f.add_pattern("Wants.package")    # only one filename is acceptable here
        d.set_filter(f)
        d.set_filename(dz_util.get_good_filename(self.prefs.wants_name))
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.prefs.wants_name = d.get_filename()
            self.wTree.get_widget("wantsentry").set_text(self.prefs.wants_name)
            self.wTree.get_widget("reconfigbutton").set_sensitive(self.w1.can_reconfig())
        d.destroy()

    def on_holiday1entry_changed(self, widget):
        self.prefs.holiday1 = widget.get_chars(0, -1)
    def on_getfilebutton4_clicked(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(buttons=b)
        d.set_title("Select Holiday package...")          # todo: move this string
        f = gtk.FileFilter()
        f.add_pattern("*.package")
        d.set_filter(f)
        if self.prefs.holiday1:
            d.set_filename(dz_util.get_good_filename(self.prefs.holiday1))
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.prefs.holiday1 = d.get_filename()
            self.wTree.get_widget("holiday1entry").set_text(self.prefs.holiday1)
        d.destroy()

    def on_holiday2entry_changed(self, widget):
        self.prefs.holiday2 = widget.get_chars(0, -1)
    def on_getfilebutton5_clicked(self, widget):
        b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        d = gtk.FileChooserDialog(buttons=b)
        d.set_title("Select Holiday package...")          # todo: move this string
        f = gtk.FileFilter()
        f.add_pattern("*.package")
        d.set_filter(f)
        if self.prefs.holiday2:
            d.set_filename(dz_util.get_good_filename(self.prefs.holiday2))
        r = d.run()
        if r == gtk.RESPONSE_ACCEPT:
            self.prefs.holiday2 = d.get_filename()
            self.wTree.get_widget("holiday2entry").set_text(self.prefs.holiday2)
        d.destroy()

    def on_reconfigbutton_clicked(self, widget):
        self.w1.reconfig()

    def on_ccodecombobox_changed(self, widget):
        self.prefs.ccode = widget.get_active() + 1
        self.w1.do_cv_refresh()

# about

class Window3:
    def __init__(self, main_window):
        logging.info("about window init")

        self.w1 = main_window

        # load widgets
        self.wTree = gtk.glade.XML("dizziness.glade", "window3")

        # connect signals to my callbacks
        s = {}
        for a in dir(self):
            s[a] = getattr(self, a)
        self.wTree.signal_autoconnect(s)

        # set icon
        self.icon = gtk.gdk.pixbuf_new_from_file("dizziness.png")
        self.wTree.get_widget("window3").set_icon(self.icon)

        # show
        self.wTree.get_widget("window3").show_all()

    def on_window3_destroy(self, widget):
        # reenable about widgets in main window
        self.w1.wTree.get_widget("about1").set_sensitive(True)
        self.w1.wTree.get_widget("aboutbutton").set_sensitive(True)
        self.w1.w3 = None

    def on_okbutton_clicked(self, widget):
        self.wTree.get_widget("window3").destroy()

# font

class Window4:
    def __init__(self, main_window, target):
        logging.info("font window init")

        self.w1 = main_window
        self.prefs = main_window.prefs
        self.target = target                # 1 = content views, 2 = tree views

        # create font dialog
        self.dialog = gtk.FontSelectionDialog("Choose Font")
        if self.target == 1:
            self.dialog.set_font_name(pango.FontDescription(self.prefs.font_name).to_string())
        else:
            self.dialog.set_font_name(pango.FontDescription(self.prefs.tree_font).to_string())

        # connect signals to my callbacks
        self.dialog.connect("destroy", self.on_window4_destroy)
        self.dialog.ok_button.connect("clicked", self.on_okbutton_clicked)
        self.dialog.apply_button.connect("clicked", self.on_applybutton_clicked)
        self.dialog.cancel_button.connect("clicked", self.on_cancelbutton_clicked)

        # show
        self.dialog.show_all()

    def on_window4_destroy(self, widget):
        self.w1.wTree.get_widget("font1").set_sensitive(True)
        self.w1.wTree.get_widget("tree_font1").set_sensitive(True)
        self.w1.w4 = None

    def on_okbutton_clicked(self, widget):
        self.on_applybutton_clicked(widget)
        self.dialog.destroy()

    def on_applybutton_clicked(self, widget):
        if self.target == 1:
            self.prefs.font_name = self.dialog.get_font_name()
            self.w1.do_cv_refresh()
        else:
            self.prefs.tree_font = self.dialog.get_font_name()
            self.w1.do_tv_refresh()

    def on_cancelbutton_clicked(self, widget):
        self.dialog.destroy()
