# view handler
#
# - functions
#
# get_view(package, group, type, subtype, instance, mode): get info on specified resource
# get_view2(package, group, type, subtype, instance, meta): get info on specified meta
# get_group_name(package, group): get the name of a group
# get_tips(package, group, type, subtype, instance, meta): get tips on specified meta
# do_dump(package, mode, group_names, group_ids, group_depth, main_window)
# do_dump_one(package, group, name, file, mode): output trees of one particular group to file
# do_dump_one_b(package, group, instance, file): output one tree to file

import gtk
import logging
import os

import dz_basic
import dz_bcon
import dz_lua
import dz_pack
import dz_str
import dz_util
from dz_defs import *

def get_view2(p, g, t, s, i, x):
    """Return a string representing this meta resource."""
    s1 = "Group = 0x%X\n" % g
    if x != 1:  # not group header?
        s1 += "Type = 0x%X\n" % t

    g2 = dz_util.get_semi(p, g)
    if g2 != 0 and g2 != g:
        n = p.prefs.sg_ids.index(g2)
        s1 += "Semi-global = %s (0x%X)\n" % (p.prefs.sg_names[n], g2)
    return s1

def get_view(p, g, t, s, i, mode):
    """Return a string and markup list representing this resource."""
    p.group = g
    p.instance = i
    p.subtype = s
    p.semi = dz_util.get_semi(p, g)
    dm2 = p.prefs.dump_mode  # save dump mode
    p.prefs.dump_mode = 0

    if t == TYPE_STR or t == TYPE_CATS or t == TYPE_CTSS or t == TYPE_TTAs:
        r = p.find({'type': t, 'group': g, 'instance': i})
        if not r:                   # not found
            p.prefs.dump_mode = dm2  # restore dump mode
            return "", []
        b = p.get(r['offset'], r['size'])
        if not b:                   # missing (!)
            p.prefs.dump_mode = dm2  # restore dump mode
            return "", []
        if mode == 0:
            s1 = dz_str.get_str(b, {'list': 1, 'instance': i, 'ccode': p.prefs.ccode})
            s2 = "# Group = 0x%X, " % g
            s2 += "Type = 0x%X" % t
            for x in range(len(s1)):
                s2 += s1[x]
        else:
            s1, s3 = dz_str.get_str(b, {'list': 1, 'instance': i, 'ccode': p.prefs.ccode, 'desc': 1})
            s2 = "# Group = 0x%X, " % g
            s2 += "Type = 0x%X" % t
            for x in range(len(s1)):
                s2 += s1[x]
                if x < len(s3): s2 += s3[x]
        p.prefs.dump_mode = dm2  # restore dump mode
        return s2, []

    elif t == TYPE_BCON:
        s1 = "# Group = 0x%X, " % g
        r = p.find({'type': TYPE_BCON, 'group': g, 'instance': i})
        if not r:
            p.prefs.dump_mode = dm2  # restore dump mode
            return s1, []

        b = p.get(r['offset'], r['size'])
        if not b:
            p.prefs.dump_mode = dm2  # restore dump mode
            return "", []

        j = 0x40                        # get title
        if '\x00' in b[:0x40]: j = b.index('\x00')
        s1 += 'BCON 0x%X(%s):\n\n' % (i, b[:j])

        s1 += dz_bcon.get_const(p, i, 0, b)
        p.prefs.dump_mode = dm2  # restore dump mode
        return s1, []

    elif t == TYPE_BHAV:
        s2 = "# Group = 0x%X, Instance = 0x%X\n" % (g, i)
        sm2 = []
        mr, mofs, smx = -1, 0, 1
        if mode == 0:
            s1, sm = dz_basic.get_tree(p, False)
            if len(sm) > 0:
                mt, mr, ma, mb = sm[0]      # get first header index
            for x in range(len(s1)):
                s2 += s1[x]
                if mr == x and smx < len(sm):
                    # adjust line-relative markup to current position
                    mt, mr, ma, mb = sm[smx]
                    while mt != 0 and smx < len(sm):
                        sm2.append( (mt, mr, ma+mofs, mb+mofs) )
                        smx += 1
                        if smx < len(sm):
                            mt, mr, ma, mb = sm[smx]
                    smx += 1
                mofs = len(s2)
        else:
            s1, s3, sm = dz_basic.get_tree(p, True)
            if len(sm) > 0:
                mt, mr, ma, mb = sm[0]      # get first header index
            for x in range(len(s1)):
                if x < len(s3):
                    s2 += s1[x] + s3[x]
                else:
                    s2 += s1[x]
                if mr == x and smx < len(sm):
                    # adjust line-relative markup to current position
                    mt, mr, ma, mb = sm[smx]
                    while mt != 0 and smx < len(sm):
                        sm2.append( (mt, mr, ma+mofs, mb+mofs) )
                        smx += 1
                        if smx < len(sm):
                            mt, mr, ma, mb = sm[smx]
                    smx += 1
                mofs = len(s2)
        p.prefs.dump_mode = dm2  # restore dump mode
        return s2, sm2

    elif t == TYPE_GLUA or t == TYPE_OLUA:
        s1 = "# Type = 0x%X, " % t
        if s != 0:  # subtype = high instance
            s1 += "Instance = 0x%X 0x%X\n" % (i, s)
        else:
            s1 += "Instance = 0x%X\n" % i
        s2 = dz_lua.get_chunk(p, t)
        for x in range(len(s2)):
            s1 += s2[x]
        p.prefs.dump_mode = dm2  # restore dump mode
        return s1, []

    elif t == TYPE_GLOB:
        s1 = "Group = 0x%X\n" % g
        s1 += "Type = 0x%X\n" % t
        s1 += "Instance = 0x%X\n\n" % i

        r = p.find({'type': t, 'group': g, 'instance': i})
        if r:
            b = p.get(r['offset'], r['size'])
            if b:
                k = ord(b[0x40])            # length (of pstring)
                if ord(b[len(b)-1]) == 0:
                    s1 += b[0x40:len(b)-1]   # cstring, actually
                else:
                    s1 += b[0x41:0x41+k]
        p.prefs.dump_mode = dm2  # restore dump mode
        return s1, []

    elif t == TYPE_NREF:
        s1 = "Group = 0x%X\n" % g
        s1 += "Type = 0x%X\n" % t
        s1 += "Instance = 0x%X\n\n" % i
        r = p.find({'type': t, 'group': g, 'instance': i})
        if r:
            b = p.get(r['offset'], r['size'])
            s1 += b
        p.prefs.dump_mode = dm2  # restore dump mode
        return s1, []

    elif t == TYPE_FWAV:
        s1 = "Group = 0x%X\n" % g
        s1 += "Type = 0x%X\n" % t
        s1 += "Instance = 0x%X\n\n" % i
        r = p.find({'type': t, 'group': g, 'instance': i})
        if r:
            b = p.get(r['offset'], r['size'])
            s1 += b[0x40:len(b)-1]
        p.prefs.dump_mode = dm2  # restore dump mode
        return s1, []

    else:
        s1 = "Group = 0x%X\n" % g
        s1 += "Type = 0x%X\n" % t
        if s != 0:  # subtype = high instance
            s1 += "Instance = 0x%X 0x%X\n" % (i, s)
        else:
            s1 += "Instance = 0x%X\n" % i
        s1 += "\n"

        r = p.find({'type': t, 'group': g, 'subtype':s, 'instance': i})
        if r:
            b = p.get(r['offset'], r['size'])
            if b:
                x = 0
                s2 = ""
                while x < len(b):
                    if (x & 0x0F) == 0:
                        if x != 0:
                            s1 += " | " + s2 + "\n"
                            s2 = ""
                        s1 += "%6X: " % x
                    c = ord(b[x])
                    s1 += "%02X" % c
                    if c >= 0x20 and c <= 0x7E:
                        s2 += b[x]
                    else:
                        s2 += "."
                    if (x & 1) == 1:
                        s1 += " "
                    x += 1
                if len(b):
                    while (x & 0x0F) != 0:
                        s1 += "  "
                        if (x & 1) == 1:
                            s1 += " "
                        x += 1
                    s1 += " | " + s2 + "\n"

    p.prefs.dump_mode = dm2  # restore dump mode
    return s1, []

def get_group_name(p, g):
    """Return the name of this group."""
    if g == GROUP_GLOBAL:
        s = "Global"
    elif g == 0xFFFFFFFF:
        s = "Custom"
    else:
        if g != 0 and g in p.prefs.sg_ids:
            j = p.prefs.sg_ids.index(g)
            s = p.prefs.sg_names[j]
        elif g in GROUP_FIXUP1:
            j = GROUP_FIXUP1.index(g)
            s = "FUBAR"     # note: FUBAR = name is uncertain
            r = p.find({'group': g, 'type': TYPE_NREF})
            if r:
                i2 = r['instance']
                if i2 in GROUP_FIXUP2[j]:
                    g2 = GROUP_FIXUP2[j].index(i2)
                    s = GROUP_FIXUP4[j][g2]
        elif g in p.prefs.object_group:
            j = p.prefs.object_group.index(g)
            s = p.prefs.object_names[j]
        elif g in p.prefs.holiday_group:
            j = p.prefs.holiday_group.index(g)
            s = p.prefs.holiday_names[j]
        else:
            s = "0x%X" % g
    return s

def get_tips(p, g, t, s, i, x):
    """Return a tooltip string for specified label."""
    s1 = ""

    g2 = dz_util.get_semi(p, g)
    if g2 != 0 and g2 != g:
        n = p.prefs.sg_ids.index(g2)
        s1 += "Semi-global: %s\n" % p.prefs.sg_names[n]

    s1 += "Group: " + get_group_name(p, g) + "\n"

    hx, tx = os.path.split(p.filename)
    s1 += "File: %s" % tx
    return s1

def do_dump_one_b(p, g, i, f):
    """Dump one specific behavior tree."""
    # note: p.semi is set externally
    p.group = g
    p.instance = i
    s, sm = dz_basic.get_tree(p, False)
    try:
        f.write(dz_util.get_html(p, TYPE_COMMENT,"\n###\n# Group = "))
        f.write(dz_util.get_html(p, TYPE_ANCHOR_FUNCTION, "0x%X, Instance = 0x%X\n" % (g, i), (g, i) ))
        for x in range(len(s)):
            f.write(s[x])
    except:
        logging.error("dump: failed to write")

def do_dump_one_d(p, g, i, f):
    """Dump one specific behavior tree in data mode."""
    # note: p.semi is set externally
    p.group = g
    p.instance = i
    s1, s2, sm = dz_basic.get_tree(p, True)
    try:
        f.write(dz_util.get_html(p, TYPE_COMMENT,"\n###\n# Group = "))
        f.write(dz_util.get_html(p, TYPE_ANCHOR_FUNCTION, "0x%X, Instance = 0x%X\n" % (g, i), (g, i) ))
        for x in range(len(s1)):
            if x < len(s2):
                f.write(s1[x] + s2[x])
            else:
                f.write(s1[x])
    except:
        logging.error("dump: failed to write")

def do_dump_one_l(p, g, t, n, h, i, f):
    """Dump one specific Lua chunk."""
    p.instance = i
    p.subtype = h
    s = dz_lua.get_chunk(p, t)
    try:
        f.write( "\n# Type = 0x%X, " % t )
        f.write( "Instance = 0x%X 0x%X\n" % (h, i) )
        for x in range(len(s)):
            f.write(s[x])
    except:
        logging.error("dump: failed to write")

def do_dump_one(p, g, nm, f, m):
    """Dump one specific group of behavior trees."""
    # collect instances
    s = []
    r = p.find({'type': TYPE_BHAV, 'group': g})
    while r:
        s.append(r['instance'])
        r = p.find({'type': TYPE_BHAV, 'group': g, 'index': r['index']})

    # collect Lua instances
    sl = []
    slh = []
    sln = []
    slt = []
    r = p.find({'type': TYPE_GLUA, 'group': g})
    while r:
        sl.append(r['instance'])
        slh.append(r['subtype'])
        sln.append(p.get_name_lua(r['offset'], r['size']))
        slt.append(TYPE_GLUA)
        r = p.find({'type': TYPE_GLUA, 'group': g, 'index': r['index']})

    r = p.find({'type': TYPE_OLUA, 'group': g})
    while r:
        sl.append(r['instance'])
        slh.append(r['subtype'])
        sln.append(p.get_name_lua(r['offset'], r['size']))
        slt.append(TYPE_OLUA)
        r = p.find({'type': TYPE_OLUA, 'group': g, 'index': r['index']})

    # sort list
    s2 = []
    for x in range(len(s)):
        ix = False
        for y in range(len(s2)):
            if s[x] < s2[y]:
                s2.insert(y, s[x])
                ix = True
                break   # terminate y loop
        if not ix:
            s2.append(s[x])

    # sort Lua list
    sl2 = []
    slh2 = []
    sln2 = []
    slt2 = []
    for x in range(len(sl)):
        ix = False
        for y in range(len(sl2)):
            if sln[x] < sln2[y]:
                sl2.insert(y, sl[x])
                slh2.insert(y, slh[x])
                sln2.insert(y, sln[x])
                slt2.insert(y, slt[x])
                ix = True
                break   # terminate y loop
        if not ix:
            sl2.append(sl[x])
            slh2.append(slh[x])
            sln2.append(sln[x])
            slt2.append(slt[x])

    # set semi-global
    p.semi = dz_util.get_semi(p, g)

    # go go gadget dump this group
    try:
        f.write(dz_util.get_html(p, TYPE_COMMENT,"\n########\n# Group = "))
        s = "0x%X (%s)\n" % (g, nm)
        if p.prefs.dump_one and m != 4:
            s = dz_util.get_html(p, TYPE_ANCHOR_GROUP, s, g)
        f.write(s)
    except:
        logging.error("dump: failed to write")

    for k in range(len(s2)):
        i = s2[k]
        if m == 2:
            do_dump_one_d(p, g, i, f)
        else:
            do_dump_one_b(p, g, i, f)

    for k in range(len(sl2)):
        do_dump_one_l(p, g, slt2[k], sln2[k], slh2[k], sl2[k], f)

def do_header(f, t):
    f.write("<HTML><HEAD>\n<TITLE>%s</TITLE>\n<LINK REL='stylesheet' HREF='default.css' TYPE='text/css'>\n</HEAD>\n<BODY><PRE>\n" % t)

def do_footer(f):
    f.write("\n</PRE></BODY></HTML>")

def do_dump2(main, p, m, gn, gi, gd):
    """Dump the contents of a package."""
    main.progress_show = True
    main.progress_amount = 0
    main.progress_divisor = 100
    main.status_text = "Dumping..."
    main.wait_gtk()

    # param dump mode: 0/1 = dump file, 2 = data, 3 = group tree only, 4 = selected group

    # prefs dump mode: 0 = text, 1 = html
    m2 = p.prefs.dump_mode

    x, tx = os.path.split(p.filename)           # tx = package name
    hx, x = os.path.split(p.prefs.dump_name)    # hx = dump path header
    hx += os.sep    # + slash

    if m < 3 and not p.prefs.dump_one:  # dump file + split mode
        hx = p.prefs.dump_name + os.sep

    sx = ".txt"
    if m2 == 1: sx = ".html"

    p.ext_flag = 0
    if m > 2 or not p.prefs.dump_one:
        p.ext_flag = 1
    try:
        if p.prefs.dump_one or m > 2:
            f = open(p.prefs.dump_name, 'w')
        else:
            f = open(hx+"group-tree"+sx, 'w')
    except:
        logging.error("dump: failed to open for writing")
        f = None
    if f:
        if m2 == 1: do_header(f, tx + " - Group Tree")
        # dump the group tree
        if m != 4:
            try:
                f.write(dz_util.get_html(p, TYPE_COMMENT,"########\n# Group Tree:\n#\n"))
                for j in range(len(gn)):
                    if gd[j] == 0:
                        f.write(dz_util.get_html(p, TYPE_COMMENT, "# "))
                        s = "%s (0x%X)" % (gn[j], gi[j])
                        f.write(dz_util.get_html(p, TYPE_LINK_GROUP, s, gi[j]))
                    else:
                        f.write(dz_util.get_html(p, TYPE_COMMENT, "# + "))
                        s = "%s (0x%X)" % (gn[j], gi[j])
                        f.write(dz_util.get_html(p, TYPE_LINK_GROUP, s, gi[j]))
                    f.write(dz_util.get_html(p, TYPE_COMMENT, "\n"))
            except:
                logging.error("dump: failed to write group tree")

        if m < 3:
            # dump the behavior trees in this package
            main.progress_divisor = len(gn)
            for j in range(len(gn)):
                if len(gn) > 1:
                    main.status_text = "Dumping %s (%d of %d)..." % (gn[j], j, len(gn))
                    main.progress_amount = j
                    main.wait_gtk()

                if not p.prefs.dump_one:    # split mode
                    if m2 == 1: do_footer(f)
                    f.close()
                    try:
                        sf = "group-%X%s" % (gi[j], sx)
                        f = open(hx+sf, 'w')
                        if m2 == 1: do_header(f, tx + " - " + gn[j])
                    except:
                        logging.error("dump: failed to open for writing")
                        break               # terminate j loop
                do_dump_one(p, gi[j], gn[j], f, m)

        elif m == 4:
            g = p.group
            do_dump_one(p, g, get_group_name(p, g), f, 0)

        # finished
        if m2 == 1: do_footer(f)
        f.close()

    main.progress_show = False
    main.status_text = ""
    main.wait_gtk()

def do_dump(main, mode):
    logging.info("dump, mode = %d" % mode)

    # collect groups from the treeview
    s1, s2, s3 = [], [], []
    tv = main.current_cv
    m = tv.get_model()
    bd = True
    p = main.get_current_pack()

    if m:
        if mode != 4:
            j = m.get_iter_root()
            while j:
                s1.append(m.get_value(j, 0))    # name
                s2.append(m.get_value(j, 1))    # group
                s3.append(0)                    # depth
                k = m.iter_children(j)
                while k:
                    if m.get_value(k, 5) == 1:  # meta = header
                        s1.append(m.get_value(k, 0))    # name
                        s2.append(m.get_value(k, 1))    # group
                        s3.append(1)            # depth
                    k = m.iter_next(k)
                j = m.iter_next(j)
        else:
            nb = main.wTree.get_widget("notebook2")
            sw = nb.get_nth_page(nb.get_current_page())     # current scrollwindow
            b = sw.get_child().get_buffer()                 # textbuffer
            p.group = b.get_data("current_group")

        dn = main.prefs.dump_name
        if mode < 3 and not p.prefs.dump_one:       # dump file + split mode
            b = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT, gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
            d = gtk.FileChooserDialog(action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER, buttons=b)
            d.set_title("Dump To...")
            d.set_filename(dz_util.get_good_filename(dn))
            r = d.run()
            if r == gtk.RESPONSE_ACCEPT:
                main.prefs.dump_name = d.get_filename()
            else:
                bd = False
            d.destroy()

        # dumpy dumpy
        if bd:
            main.wTree.get_widget("menubar1").set_sensitive(False)
            main.wTree.get_widget("toolbar1").set_sensitive(False)
            main.wTree.get_widget("notebook1").set_sensitive(False)
            do_dump2(main, p, mode, s1, s2, s3)
            main.wTree.get_widget("menubar1").set_sensitive(True)
            main.wTree.get_widget("toolbar1").set_sensitive(True)
            main.wTree.get_widget("notebook1").set_sensitive(True)

        main.prefs.dump_name = dn   # restore old dump name
